/*interface of NICSLU*/
/*last modified: apr 14, 2013*/
/*author: Chen, Xiaoming*/

#ifndef __NICSLU__
#define __NICSLU__

#include "nics_config.h"

/*return code*/
#define NICSLU_GENERAL_FAIL							(-1)
#define NICSLU_ARGUMENT_ERROR						(-2)
#define NICSLU_MEMORY_OVERFLOW						(-3)
#define NICSLU_FILE_CANNOT_OPEN						(-4)
#define NICSLU_MATRIX_STRUCTURAL_SINGULAR			(-5)
#define NICSLU_MATRIX_NUMERIC_SINGULAR				(-6)
#define NICSLU_MATRIX_INVALID						(-7)
#define NICSLU_MATRIX_ENTRY_DUPLICATED				(-8)
#define NICSLU_THREADS_NOT_INITIALIZED				(-9)
#define NICSLU_MATRIX_NOT_INITIALIZED				(-10)
#define NICSLU_SCHEDULER_NOT_INITIALIZED			(-11)
#define NICSLU_SINGLE_THREAD						(-12)
#define NICSLU_THREADS_INIT_FAIL					(-13)
#define NICSLU_MATRIX_NOT_ANALYZED					(-14)
#define NICSLU_MATRIX_NOT_FACTORIZED				(-15)
#define NICSLU_NUMERIC_OVERFLOW						(-16)


/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

/*the main data structure*/
typedef struct tagSNicsLU
{
	/*flags*/
	bool__t *flag;

	/*statistics*/
	real__t *stat;

	/*configurations*/
	uint__t *cfgi;
	real__t *cfgf;

	/*matrix data, 6 items*/
	uint__t n;				/*dimension*/
	uint__t nnz;			/*nonzeros of A*/
	real__t *ax;			/*value*/
	uint__t *ai;			/*column/row index*/
	uint__t *ap;			/*row/column header*/
	real__t *rhs;			/*for solve and tsolve*/

	/*other matrix data, 9 items*/
	uint__t *row_perm;		/*row_perm[i]=j-->row i in the permuted matrix is row j in the original matrix*/
	uint__t *row_perm_inv;	/*row_perm_inv[i]=j-->row i in the original matrix is row j in the permuted matrix*/
	uint__t *col_perm;
	uint__t *col_perm_inv;
	real__t *col_scale_perm;/*permuted*/
	real__t *row_scale;		/*not permuted*/
	real__t *cscale;
	int__t *pivot;			/*pivot[i]=j-->column j is the ith pivot column*/
	int__t *pivot_inv;		/*pivot_inv[i]=j-->column i is the jth pivot column*/

	/*lu matrix, 13 items*/
	size_t lu_nnz_est;		/*estimated total nnz by AMD*/
	size_t lu_nnz;			/*nonzeros of factorized matrix L+U-I*/
	size_t l_nnz;			/*inclu diag*/
	size_t u_nnz;			/*inclu diag*/
	real__t *ldiag;			/*udiag=1.0*/
	void *lu_array;			/*lu index and data*/
	size_t *up;				/*u row header, the header of each row*/
	uint__t *llen;			/*exclu diag*/
	uint__t *ulen;			/*exclu diag*/
	size_t *len_est;		/*estimated len, for parallelism, in bytes*/
	size_t *wkld_est;		/*estimated workload, for parallelism*/
	byte__t *row_state;		/*row state, finished or un-finished*/
	void **lu_array2;		/*for parallelism*/

	/*work space, 3 items*/
	void *workspace;
	void **workspace_mt1;
	void **workspace_mt2;

	/*for parallelism, 10 items*/
	int__t thread_work;
	void *thread_id;		/*thread id, internal structure, not open*/
	void *thread_arg;		/*thread argument, internal structure, not open*/
	bool__t *thread_active;
	bool__t *thread_finish;
	uint__t *cluster_start;
	uint__t *cluster_end;
	uint__t pipeline_start;
	uint__t pipeline_end;
	uint__t last_busy;

	/*aegraph, 6 items*/
	uint__t aeg_level;
	uint__t *aeg_data;
	uint__t *aeg_header;
	uint__t aeg_refact_level;
	uint__t *aeg_refact_data;
	uint__t *aeg_refact_header;

	/*timer*/
	void *timer;

} SNicsLU;


/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/
/*******************************************************************************/

/*argument flags*/
#define IN__
#define OUT__
#define INOUT__

#ifdef __cplusplus
extern "C" {
#endif

/*******************************************************************************/
/*******************************************************************************/
/*create and destroy the main structure*/
int__t \
	NicsLU_Initialize( \
	INOUT__ SNicsLU *nicslu);

int__t \
	NicsLU_Destroy( \
	INOUT__ SNicsLU *nicslu);


/*******************************************************************************/
/*******************************************************************************/
/*create and destroy the matrix*/
int__t \
	NicsLU_CreateMatrix( \
	INOUT__ SNicsLU *nicslu, \
	IN__ uint__t n, \
	IN__ uint__t nnz, \
	IN__ real__t *ax, \
	IN__ uint__t *ai, \
	IN__ uint__t *ap);

int__t \
	NicsLU_DestroyMatrix( \
	INOUT__ SNicsLU *nicslu);


/*******************************************************************************/
/*******************************************************************************/
/*create the scheduler, it must be called after NicsLU_Analyze*/
/*return 0: suggest using parallel version*/
/*return >0: suggest using sequential version*/
/*the suggestion is only for NICSLU_Factorize_MT*/
/*NICSLU_ReFactorize_MT can always get effective speedups*/
int__t \
	NicsLU_CreateScheduler( \
	INOUT__ SNicsLU *nicslu);

/*destroy the scheduler*/
int__t \
	NicsLU_DestroyScheduler( \
	INOUT__ SNicsLU *nicslu);


/*******************************************************************************/
/*******************************************************************************/
/*create and destroy the threads*/
int__t \
	NicsLU_CreateThreads( \
	INOUT__ SNicsLU *nicslu, \
	IN__ uint__t thread, \
	IN__ bool__t check);

int__t \
	NicsLU_DestroyThreads( \
	INOUT__ SNicsLU *nicslu);


/*******************************************************************************/
/*******************************************************************************/
/*kernel functions*/

/*preprocess, static pivoting and ordering*/
int__t \
	NicsLU_Analyze( \
	INOUT__ SNicsLU *nicslu);

/*numeric factorization, with partial pivoting*/
int__t \
	NicsLU_Factorize( \
	INOUT__ SNicsLU *nicslu);

/*re-factorization, without partial pivoting*/
int__t \
	NicsLU_ReFactorize( \
	INOUT__ SNicsLU *nicslu, \
	IN__ real__t *ax);

/*multi threads*/
int__t \
	NicsLU_Factorize_MT( \
	INOUT__ SNicsLU *nicslu);

int__t \
	NicsLU_ReFactorize_MT( \
	INOUT__ SNicsLU *nicslu, \
	IN__ real__t *ax);

/*solve*/
int__t \
	NicsLU_Solve( \
	INOUT__ SNicsLU *nicslu, \
	INOUT__ real__t *rhs);/*for inputs, it's b, for outputs, it's overwritten by x*/

/*solve fast*/
int__t \
	NicsLU_SolveFast( \
	INOUT__ SNicsLU *nicslu, \
	INOUT__ real__t *rhs);/*for inputs, it's b, for outputs, it's overwritten by x*/

/*reset matrix values, for factorization*/
int__t \
	NicsLU_ResetMatrixValues( \
	INOUT__ SNicsLU *nicslu, \
	IN__ real__t *ax);

/*refine, when a matrix is ill-conditioned, it may be used, but it's not always ok*/
int__t \
	NicsLU_Refine( \
	INOUT__ SNicsLU *nicslu, \
	INOUT__ real__t *x, \
	IN__ real__t *b, \
	IN__ real__t eps, \
	IN__ uint__t maxiter);/*if set to 0, then no iteration count constraint*/


/*******************************************************************************/
/*******************************************************************************/
/*residual error, error = |Ax-b|, norm: 1-norm(1), 2-norm(2), -norm(other)*/
int__t \
	NicsLU_Residual( \
	IN__ uint__t n, \
	IN__ real__t *ax, \
	IN__ uint__t *ai, \
	IN__ uint__t *ap, \
	IN__ real__t *x, \
	IN__ real__t *b, \
	OUT__ real__t *error, \
	IN__ int__t norm, \
	IN__ uint__t mode);/*mode=0:row mode, mode=1:column mode*/

/*memory throughput, in bytes*/
int__t \
	NicsLU_Throughput( \
	IN__ SNicsLU *nicslu, \
	OUT__ real__t *thr);

/*flops, the total count of floating-point operations*/
int__t \
	NicsLU_Flops( \
	INOUT__ SNicsLU *nicslu, \
	OUT__ real__t *flops);

/*flops of each thread*/
int__t \
	NicsLU_ThreadLoad( \
	IN__ SNicsLU *nicslu, \
	IN__ uint__t thread, \
	OUT__ real__t **thread_flops);

/*transpose the matrix*/
int__t \
	NicsLU_Transpose( \
	IN__ uint__t n, \
	IN__ uint__t nnz, \
	INOUT__ real__t *ax, \
	INOUT__ uint__t *ai, \
	INOUT__ uint__t *ap);

/*dump matrix A after factorization, so that the dumped data can be used for a self-written re-factorization function*/
int__t \
	NicsLU_DumpA( \
	IN__ SNicsLU *nicslu, \
	OUT__ real__t **ax, \
	OUT__ uint__t **ai, \
	OUT__ uint__t **ap);

/*dump lu. inclu diag, udiag=1.0*/
int__t \
	NicsLU_DumpLU( \
	IN__ SNicsLU *nicslu, \
	OUT__ real__t **lx, \
	OUT__ uint__t **li, \
	OUT__ size_t **lp, \
	OUT__ real__t **ux, \
	OUT__ uint__t **ui, \
	OUT__ size_t **up);

/*condition number*/
int__t \
	NicsLU_ConditionNumber( \
	INOUT__ SNicsLU *nicslu, \
	OUT__ real__t *cond);

/*performance evaluation*/
int__t \
	NicsLU_Performance( \
	IN__ SNicsLU *nicslu, \
	OUT__ real__t *perf, \
	OUT__ real__t *tperf, \
	IN__ uint__t thread);


#ifdef __cplusplus
}
#endif

#endif
