
/*!
 * \file kat.c
 * \brief Main routine and banner routine for finesse
 */

/*! \mainpage Finesse - Frequency domain INterferomEter Simulation SoftwarE
 *
 * Version: 0.99.x  (documentation built from $Revision$)
 *
 * \author Andreas Freise
 *
 * FINESSE is a fast interferometer simulation program.  For a given optical
 * setup, it computes the light field amplitudes at every point in the
 * interferometer assuming a steady state.  To do so, the interferometer
 * description is translated into a set of linear equations that are solved
 * numerically. For convenience, a number of standard analyses can be
 * performed automatically by the program, namely computing
 * modulation-demodulation error signals, transfer functions, shot-noise
 * limited sensitivities, and beam shapes.  FINESSE can perform the analysis
 * using the plane-wave approximation or Hermite-Gauss modes. The latter
 * allows computation of the properties of optical systems like telescopes
 * and the effects of mode matching and mirror angular positions. 
 *
 *  @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */

#include "kat.h"
#include "kat_config.h"
#include "kat_mem.h"
#include "kat_read.h"
#include "kat_io.h"
#include "kat_aux.h"
#include "kat_init.h"
#include "kat_calc.h"
#include "kat_spa.h"
#include "klu.h" 
#include "kat_klu.h"
#include "kat_inline.c"
#include "kat_matrix_ccs.h"
#include "kat_quant.h"
#include "utarray.h"

#ifdef SERVERMODE
#include "kat_server.h"
#endif

#ifdef OSUNIX
#include <fpu_control.h>
#endif

init_variables_t init; //!< Structure to hold initialisation information
interferometer_t inter; //!< Interferometer and simulation variables
options_t options; //!< Options passed to finesse
memory_t mem; //!< Structure to hold memory allocation information

local_var_t vlocal; //!< Unchanged variables for faster function calls

int *t_s; //!< Type_of_Signal list [component type]
double *f_s; //!< Frequency value list [frequency]
complex_t ***a_s; //!< Amplitude list [field] [output] [frequeny]

global_var_t vglobal; //!< Structure to hold in, out and gnu filenames
klu_sparse_var_t klu; //!< Variables for KLU sparse matrix routines

klu_objs_t klu_old = {0}; // storage of old klu solver structs

ifo_matrix_vars_t M_ifo_car = {{0}};
ifo_matrix_vars_t M_ifo_sig = {{0}};

matrix_ccs_t M_qni = {0};

UT_icd motion_link_icd = {sizeof(motion_link_t), NULL, NULL, NULL};

int carrier_frequencies_tuned = 0;
int signal_frequencies_tuned = 0;

time_t now; //!< Current time
char mydate[12] = {0}; //!< Date as at time of compilation

char functokens[27] = {0}; //!< Array of function tokens

FILE *fp_log; //!< Log file output file pointer
bool bCygserverRunning;

extern bool enableTiming;
transfer_func_t mtf_free_mass;
transfer_func_t mtf_unity;

//! The main routine

/*!
 * \param argc The number of input parameters
 * \param argv The array of input parameters
 */
int main(int argc, char *argv[]) {
#ifdef OSUNIX
    /*
      This puts the X86 FPU in 64-bit precision mode.  The default
      under Linux is to use 80-bit mode, which produces subtle
      differences from FreeBSD and other systems, eg,
      (int)(1000*atof("0.3")) is 300 in 64-bit mode, 299 in 80-bit
      mode.

      To be tested! Especially the function 'pow' might cause problems.
     */
    fpu_control_t cw;

    _FPU_GETCW(cw);
    cw &= ~_FPU_EXTENDED;
    cw |= _FPU_DOUBLE;
    _FPU_SETCW(cw);
#endif

    FILE *ifp;
    FILE *fp = NULL;
    (void) fp;
    
    // setup default free mass for optics
    mtf_free_mass.gain = 1.0;
    mtf_free_mass.phase = 0.0;
    mtf_free_mass.num_poles = 2;
    mtf_free_mass.num_zeros = 0;
    mtf_free_mass.type = GENERAL_TF;
    mtf_free_mass.index = -1;
    strcpy(mtf_free_mass.name, "free_mass");
    // set 2 complex poles with a value of 0
    mtf_free_mass.poles[0] = complex_0;
    mtf_free_mass.poles[1] = complex_0;
    
    mtf_unity.gain = 1.0;
    mtf_unity.phase = 0.0;
    mtf_unity.num_poles = 0;
    mtf_unity.num_zeros = 0;
    mtf_unity.type = Q_F_TF;
    mtf_unity.index = -2;
    strcpy(mtf_free_mass.name, "unity");
    
    bCygserverRunning = isCygserverRunning();
            
    char *s, *s2;
    int basenamelength;
    char errmsg[ERR_MSG_LEN] = {0};

    inter.setup = false;

    strcpy(mydate, MYTIME);
    strcpy(functokens, "abcdfghijklmnopqrstuvwxyz");

    /* **** get options **** */

    // reset options for ...
    options.quiet = 0; // switch off all output
    options.help = 0; // print help
    options.sparse_solver = KLU_FULL; // use KLU as default
    options.sparse_solver_manual_selection = 0; // overwrite KLU-SPARSE setting manually
    options.check = 0; // do a consistency check of the interferometer matrix
    options.perl1 = 0; // switch off some output
    options.maxmin = 0; // print max/min values
    options.servermode = 0; // switch between standalone and server mode
    options.portnumber = 0; // port number for client server connections (11000 to 110010)
    options.version = 0; // print version number and quit?
    options.noheader = 0; // do not print header in output file
    options.no_output_bksp =  0;
    strcpy(options.out_format, DEFAULT_OUT_FORMAT);
    
    bool converting = false;
    
    // process command line arguments
    int num_options = 0;
    while ((argc > 1) && (argv[1 + num_options][0] == '-')) {
        bool found = false;

        if(converting)
            gerror("The --convert option can only be used by itself");
            
        // process -h option (print first help page)
        if (strcasecmp(argv[1 + num_options], "-h") == 0) {
            options.help = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -hh option (print second help page)
        else if (strcasecmp(argv[1 + num_options], "-hh") == 0) {
            options.help = 2;
            ++num_options;
            --argc;
            found = true;
        }
          
        else if (strcasecmp(argv[1 + num_options], "--convert") == 0) {
            // this should be the first argument 
            if(num_options != 0)
                gerror("The --convert option can only be used by itself");
            
            converting = true;
            ++num_options;
            --argc;
            found = true;
        }
            // process -v option 
        else if (strcasecmp(argv[1 + num_options], "-v") == 0) {
            options.version = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -max option (print max/min values)
        else if (strcasecmp(argv[1 + num_options], "-max") == 0) {
            options.maxmin = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -klu option (switch to KLU package)
        else if (strcasecmp(argv[1 + num_options], "-klu") == 0) {
            options.sparse_solver = KLU;
            options.sparse_solver_manual_selection = 1;
            ++num_options;
            --argc;
            found = true;
        }           
        else if (strcasecmp(argv[1 + num_options], "-klu-full") == 0) {
            options.sparse_solver = KLU_FULL;
            options.sparse_solver_manual_selection = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -sparse option (switch to SPARSE package)
        else if (strcasecmp(argv[1 + num_options], "-sparse") == 0) {
            warn("SPARSE has been removed, defaulting to KLU");
            options.sparse_solver = KLU;
            options.sparse_solver_manual_selection = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process --perl1 option (switch off some output)
        else if (strcasecmp(argv[1 + num_options], "--perl1") == 0) {
            options.perl1 = 1;
            ++num_options;
            --argc;
            found = true;
        }
        else if (strcasecmp(argv[1 + num_options], "--no-backspace") == 0) {
            options.no_output_bksp = 1;
            ++num_options;
            --argc;
            found = true;
        }
        
            // process --quiet option (switch off all output)
        else if (strcasecmp(argv[1 + num_options], "--quiet") == 0) {
            options.quiet = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process --noheader option (do not print header in output file)
        else if (strcasecmp(argv[1 + num_options], "--noheader") == 0) {
            options.noheader = 1;
            ++num_options;
            --argc;
            found = true;
        }
#ifdef TEST
            // process --test option (runs unit tests, NOT for user distro)
        else if (strcasecmp(argv[1 + num_options], "--test") == 0) {
            printf("\n##### Running tests #####\n");
            RunAllTests();
            printf("##### Finished tests #####\n");

            // clean up nicely
            fflush(stdout);
            fflush(stderr);
            exit(0);
        }
#endif

            // process --new option (switch on new features)
        else if (strcasecmp(argv[1 + num_options], "--new") == 0) {
            if (options.new == -1) {
                gerror("options --new and --old mutually exclusive\n");
            }
            options.new = 1;
            ++num_options;
            --argc;
            found = true;
        }
            // process --old option (try to switch back to last version)
        else if (strcasecmp(argv[1 + num_options], "--old") == 0) {
            if (options.new == 1) {
                gerror("options --new and --old mutually exclusive\n");
            }
            options.new = -1;
            ++num_options;
            --argc;
            found = true;
        }
            // process -c option (do consistency check of interferometer matrix)
        else if (strcasecmp(argv[1 + num_options], "-c") == 0) {
            options.check = 1;
            ++num_options;
            --argc;
            found = true;
        }
           // process format option
        else if (strncmp (argv[1 + num_options], "-format=", 8) == 0) {
            char* format = strchr(argv[1 + num_options], '=');
            
            if(format != NULL)
                format++;
            else
                gerror("There was no equal sign found for format string, e.g -format=%#.15g");
            
            if(strlen(format) == 0)
                gerror("-format output must specify a string");
            
            strcpy(options.out_format, format);
            strcat(options.out_format, " "); // add on an extra space
            
            ++num_options;
            --argc;
            found = true;
        }
           // process -f option (print data to file with more precision)
        else if (strcasecmp(argv[1 + num_options], "-f") == 0) {
            warn("-f option depreciated, default is now 15dp output.");
            ++num_options;
            --argc;
            found = true;
        }
        else if (strcasecmp(argv[1 + num_options], "--perf-timing") == 0) {
            enableTiming = true;
            ++num_options;
            --argc;
            found = true;
        }
            // process --server option (switch into server mode)
        else if (strcasecmp(argv[1 + num_options], "--server") == 0) {
            options.servermode = 1;
            ++num_options;
            --argc;
            // try to read port number 
            if ((num_options < argc) && (argv[1 + num_options][0] != '-')) {
                if (!(my_atoi(argv[1 + num_options], &options.portnumber)) &&
                        options.portnumber >= 11000 &&
                        options.portnumber <= 11010) {
                    found = true;
                    ++num_options;
                    --argc;
                }
            }

            if (!found) {
                banner(stdout, "*", "*", "*");
                gerror("usage: kat --server <portnumber (11000 to 11010)> "
                        "[options] inputfile ...\n");
            }
        }

        // warn that option is not found
        if (!found) {
            banner(stdout, "*", "*", "*");
            gerror("unknown option '%s'\n", argv[1 + num_options]);
        }        
    }

    // print version number
    if (options.version == 1) {
        print_version(stdout);
        my_finish(0);
    }

    // print the first help page
    if (options.help == 1) {
        print_help();
        my_finish(0);
    }

    // print the second help page
    if (options.help == 2) {
        print_help2();
        my_finish(0);
    }

    // switch of max if in servermode
    if (options.maxmin && options.servermode) {
        warn("the option '-max' cannot be used in server mode.\n");
        options.maxmin = 0;
    }
    
    if(!converting){
        // check that the correct number of args are passed in
        if (argc < 2 || argc > 4) {
            banner(stdout, "*", "*", "*");
            strcpy(errmsg, "usage: kat [options] infile [outfile [gnufile]]\n");
            strcat(errmsg, "       or usage: kat [options] basename\n");
            strcat(errmsg, "       (kat -h displays help message)\n");
            gerror(errmsg);
        }
    }else{
        // should only be the .knm file name after the -convert command
        if(argc != 3 && argc != 4)
            gerror("usage: kat --convert knm_file_prefix 1/2 (1:mirror, 2:beamsplitter) [new_knm_file_prefix]\n"
                   "       or usage: kat [options] basename\n"
                   "       (kat -h displays help message)\n");
        
        KNM_COMPONENT_t knmcmp;
        char *str_knmcmp = argv[2 + num_options];

        knmcmp  = atoi(str_knmcmp);
        
        if(knmcmp != MIRROR_CMP && knmcmp != BEAMSPLITTER_CMP)
            gerror("The argument for the component '%s' was not valid, enter 1 or 2 for either a mirror or beamsplitter.", str_knmcmp);

        if(argc == 3) {                        
            convert_knm_file(argv[1 + num_options], "", knmcmp);
        } else if(argc == 4) {
            convert_knm_file(argv[1 + num_options], argv[3 + num_options], knmcmp);
        } 
        
        // don't do anything after the convert so exit
        exit(0);
    }


    // process the arguments
    if (argc == 2) { // only the basename is specified...
        inter.basename = duplicate_string(argv[num_options + 1]);
        basenamelength = strlen(inter.basename);
        if (basenamelength > 4) {
            s2 = inter.basename + strlen(inter.basename) - 4;
            if (strcasecmp(s2, ".kat") == 0) {
                inter.basename[strlen(inter.basename) - 4] = '\0';
            }
        }

        s2 = inter.basename + strlen(inter.basename) - 1;
        if (*s2 == '.') {
            *s2 = '\0';
        }

        // adf 31072012 removed as it was breaking absolute and relative pathes on OSX/Unix
        // stripping leading dots and slashes
        //while (*inter.basename == '.' || *inter.basename == '/' || *inter.basename == '\\') {
        //    inter.basename++;
        //}
        
        // ddb - use the cygpath routine to change path to unix like one
#ifdef __CYGWIN__
        strcpy(inter.basename, get_path_i(inter.basename,true));
#endif

        strcpy(vglobal.input_fname, inter.basename);
        strcat(vglobal.input_fname, ".kat");
        strcpy(vglobal.output_fname, inter.basename);
        strcat(vglobal.output_fname, ".out");
        strcpy(vglobal.gnuplot_fname, inter.basename);
        strcat(vglobal.gnuplot_fname, ".gnu");
        strcpy(vglobal.log_fname, inter.basename);
        strcat(vglobal.log_fname, ".log");
    } else if (argc == 3) { // the input and output file names are specified...
        inter.basename = duplicate_string(argv[num_options + 1]);
        if ((s = strchr(inter.basename, '.')) != NULL) {
            *s = '\0';
        }
        strcpy(vglobal.input_fname, argv[num_options + 1]);
        strcpy(vglobal.output_fname, argv[num_options + 2]);
        strcpy(vglobal.gnuplot_fname, inter.basename);
        strcat(vglobal.gnuplot_fname, ".gnu");
        strcpy(vglobal.log_fname, inter.basename);
        strcat(vglobal.log_fname, ".log");
    } else if (argc == 4) { // input, output and gnuplot file names are given...
        inter.basename = duplicate_string(argv[num_options + 1]);
        if ((s = strchr(inter.basename, '.')) != NULL) {
            *s = '\0';
        }
        strcpy(vglobal.input_fname, argv[num_options + 1]);
        strcpy(vglobal.output_fname, argv[num_options + 2]);
        strcpy(vglobal.gnuplot_fname, argv[num_options + 3]);
        strcpy(vglobal.log_fname, inter.basename);
        strcat(vglobal.log_fname, ".log");
    }

    int tid_main = startTimer("MAIN");
    int tid_startup = startTimer("STARTUP");
    
    // Generating name for Matlab output files
    // (some of this code is repeated in kat_gnu.c void matlabfile(FILE *fp, char *fname)
    // TODO define new string ...)
    char matlab_func_str[LINE_LEN];
    strcpy(matlab_func_str, inter.basename);
    int replace_error = replace_string(matlab_func_str, "-", "_", 1, 1);
    if (replace_error == -1) {
        bug_error("could not replace '-' in matlab string");
    }
    strcpy(vglobal.matlab_fname, matlab_func_str);
    strcat(vglobal.matlab_fname, ".m");

    // Generating name for Python output files
    strcpy(vglobal.python_fname, inter.basename);
    strcat(vglobal.python_fname, ".py");

    // open the log file
    open_file_to_write_ascii(vglobal.log_fname, &fp_log);

    /*
    if ((fp_log = fopen(vglobal.log_fname, "w")) == NULL) {
      gerror("cannot open '%s' for log output\n", vglobal.log_fname);
    }
     */

    // set up message reporting if perl1/quiet option isn't set
    if (!options.perl1 && !options.quiet) {
        banner(stdout, vglobal.input_fname, vglobal.output_fname,
                vglobal.gnuplot_fname);
    }
    // output the banner to the log file in any case
    if (fp_log != NULL) {
        banner(fp_log, vglobal.input_fname, vglobal.output_fname,
                vglobal.gnuplot_fname);
    }


    // check to see if we can open the input file
    if ((ifp = fopen(vglobal.input_fname, "r")) == NULL) {
        gerror("cannot open '%s'\n", vglobal.input_fname);
    }

    // read in environmental variables
    read_env();
    endTimer(tid_startup);
    
    int tid_init = startTimer("INIT");
    // read the initialisation file
    read_init();
    initialise_simulation_variables();
    endTimer(tid_init);
    
    // prescan and reading of input file
    int tid_prescan = startTimer("PRESCAN");
    pre_scan(ifp);
    endTimer(tid_prescan);
    
    int tid_read = startTimer("READ_FILE");
    
    read_file(ifp);
    fclose(ifp);
    fflush(stdout);
    
    // ensure that all light inputs are actually connected to something
    {
        int i;
    
        for(i=0; i<inter.num_light_inputs; i++){
            light_in_t *in = &inter.light_in_list[i];
            assert(in != NULL);
            
            node_connections_t *nc = &inter.node_conn_list[in->node_index];
            assert(nc != NULL);
            
            if((nc->comp_1 == NULL && nc->comp_2 == NULL)) {
                gerror("Input %s must be connected to another optical component, e.g. a space (detectors do not count)\n", in->name);
            }
        }
    }
    
    endTimer(tid_read);
        
    int i = 0;
        
    switch(options.sparse_solver){
        case KLU:
            break;
            
        case KLU_FULL:
            // set type of matrices
            M_ifo_car.type = STANDARD;
            M_ifo_sig.type = SIGNAL_QCORR;
            
            // first we need to reduce the carrier frequencies if multiple input
            // light sources are present at the same frequency. Also need to take
            // into account if one of those laser frequencies are being tuned or put
            reduce_frequency_linked_list();
            
            // The number of carrier frequencies have now been reduced so that
            // only unique and tuned frequencies exist. Now we generate all the 
            // modulation and sideband frequencies around them, as well as the
            // user frequency rules
            fill_frequency_linked_list();
            
            // now do yet another reduction to remove more duplicates that are not tuned...
            reduce_frequency_linked_list();
            
            inter.num_carrier_freqs = 0;
            inter.num_signal_freqs = 0;
            
            // convert the linked list into an array for index access
            inter.carrier_f_list = (frequency_t**) calloc(inter.frequency_linked_list.count, sizeof(frequency_t*));
            inter.signal_f_list  = (frequency_t**) calloc(inter.frequency_linked_list.count, sizeof(frequency_t*));
            
            if(inter.carrier_f_list == NULL || inter.signal_f_list == NULL)
                gerror("Memory allocation for used frequency list failed");
            
            // Now loop over the reduced frequencies and add them to a static array
            // which will be used for accessing frequencies for eitehr carriers
            // or signals
            frequency_node_t *curr = inter.frequency_linked_list.head;
            
            while(curr != NULL){
                if(inter.num_signal_freqs + inter.num_carrier_freqs >= inter.frequency_linked_list.count)
                    bug_error("Trying to add more frequencies than memory allocated for");
                
                if(curr->this->isTuned){
                    if(curr->this->type == SIGNAL_FIELD)
                        signal_frequencies_tuned = 1;
                    else 
                        carrier_frequencies_tuned = 1;
                }
                
                if(curr->this->type == CARRIER_FIELD 
                        || curr->this->type == MODULATOR_FIELD
                        || curr->this->type == USER_FIELD){
                    
                    curr->this->index = inter.num_carrier_freqs;
                    inter.carrier_f_list[inter.num_carrier_freqs] = curr->this; 

                    curr = curr->next;

                    inter.num_carrier_freqs++;
                    
                } else if(curr->this->type == SIGNAL_FIELD) {
                    curr->this->index = inter.num_signal_freqs;
                    inter.signal_f_list[inter.num_signal_freqs] = curr->this; 

                    curr = curr->next;

                    inter.num_signal_freqs++;
                }
            }
            
            // ensure we have everything
            assert(inter.num_carrier_freqs + inter.num_signal_freqs == inter.frequency_linked_list.count);
            
            // inform matrix which list of frequencies they are associated with
            M_ifo_car.frequencies = inter.carrier_f_list;
            M_ifo_car.num_frequencies = inter.num_carrier_freqs;
            
            M_ifo_sig.frequencies = inter.signal_f_list;
            M_ifo_sig.num_frequencies = inter.num_signal_freqs;
            
            // Now allocate memory for storing the coupling between matrices
            alloc_matrix_ccs_modulators(&M_ifo_car);
            
            if(inter.num_signal_freqs > 0)
                alloc_matrix_ccs_modulators(&M_ifo_sig);
            
            if(inter.frequency_flag & PRINT_FREQUENCIES)
                print_used_frequency_list();
               
            // Now we need to fill in the f_couple_allocd array for all the elements
            // that might couple the various frequencies. This is used to determine
            // whether to add the elements needed for coupling between the frequencies.
            // The worst case scenario here is when a put command is setting the value
            // of the frequency, if this is the case then there is no way of telling what
            // value it will be at some point. So it may end up coupling or it may not
            // so we have to allocate the memory and matrix elements for it just incase
            update_frequency_coupling(1);
            
            if(inter.frequency_flag & PRINT_MODULATOR_COUPLINGS){
                for(i=0; i<inter.num_modulators; i++){
                    printf(" * Carrier Frequencies\n");
                    print_frequency_couplings(inter.modulator_list[i].name, inter.num_carrier_freqs,
                                                M_ifo_car.mod_does_f_couple[i],
                                                M_ifo_car.mod_f_couple_order[i]);

                    if(inter.num_signal_freqs > 0){
                        printf("* Signal Frequencies\n");
                        print_frequency_couplings(inter.modulator_list[i].name, inter.num_signal_freqs,
                                                    M_ifo_sig.mod_does_f_couple[i],
                                                    M_ifo_sig.mod_f_couple_order[i]);
                    }
                }
            }
            
            // We also need to compute which index each frequency will be placed in the result arrays: f_s,a_s and t_s
                   
            // Important!! To get more comparable results with the old Finesse code the order
            // in which the frequencies are stored and later used for computations is important.
            // This can be the difference between things equally exactly 0 or not for example.

            int f,g;
            frequency_t *f1,*f2;
            inter.num_frequencies = 0;
            
            for(f=0; f<inter.num_carrier_freqs; f++){
                f1 = inter.carrier_f_list[f];
                f1->result_index = inter.num_frequencies++;
                t_s[f1->result_index] = f1->type;
                f_s[f1->result_index] = f1->f;
                
                // now add signal frequencies if they exist
                for(g=0; g<inter.num_signal_freqs; g++){
                    f2 = inter.signal_f_list[g];
                    
                    if(f2->carrier == f1){
                        f2->result_index = inter.num_frequencies++;
                        // store the signal frequencies after the carriers
                        t_s[f2->result_index] = f2->type;
                        f_s[f2->result_index] = f2->f;
                    }
                }
            }

            assert(inter.num_frequencies == inter.num_signal_freqs + inter.num_carrier_freqs);
            
            // Now we have computed and parsed how many frequencies and various degrees of freedom
            // each component has we can allocate more sensibly the the memory and compute the total
            // number of motion equations.
            int i;
            for(i=0; i<inter.num_mirrors; i++){
                mirror_t *m = &inter.mirror_list[i];
                inter.num_motion_eqns += alloc_mirror_motions(m, inter.num_signal_freqs);
            }

            for(i=0; i<inter.num_beamsplitters; i++){
                beamsplitter_t *bs = &inter.bs_list[i];
                inter.num_motion_eqns += alloc_bs_motions(bs, inter.num_signal_freqs);
            }
        
            break;
        default:
            bug_error("sparse solver not handled");
    }
    
    // check whether we should automatically retrace
    if (inter.rebuild & 2 && !inter.retrace_manual) {
        inter.retrace = 1;
        message("Using automatic retrace\n");
    }

    // print helpful info if required
    if ((inter.tem > 4 && options.check) || inter.tem > 10) {
        message("  ...initialising system, please wait!\n");
    }

    fflush(stdout);

    // set everything up
    int tid_setup = startTimer("SETUP_SYSTEM");
    setup_system();
    endTimer(tid_setup);
    
    fflush(stdout);

    // do sparse matrix stuff
    unsigned tid = startTimer("SPA_CREATE");
    
    if (options.sparse_solver == KLU) {
        klu_create();
        klu_get();
    } else if (options.sparse_solver == KLU_FULL) {
        // solve everything in one matrix
        long bytes = 0;
        
        // number of carriers are just the number of carrier frequencies we have with how many HG fields
        // then the number of reduced nodes, carriers are not computed at dump nodes
        M_ifo_car.M.num_eqns = M_ifo_car.num_frequencies * inter.num_fields * 2 * inter.num_reduced_nodes;
        
        assert(M_ifo_car.M.num_eqns > 0);
        
        alloc_ifo_matrix_ccs(&bytes, &M_ifo_car);
        
        check_node_direction();
        
        // position each of the components in the matrix
        // false here to not allocate dump nodes
        build_ccs_matrix(&M_ifo_car);
        
        get_ccs_matrix(&M_ifo_car);
        
        set_light_input_ports();
        
        klu_objs_t *kobj = (klu_objs_t*)M_ifo_car.solver_opts;
        
        klu_l_defaults(&(kobj->Common));

        kobj->Common.ordering = 0;
        kobj->Common.scale = 2;
        kobj->Common.btf = 1;
        kobj->Common.maxwork = 0;
        
        if(M_ifo_sig.num_frequencies > 0){            
            // The signal matrix also depends on the number of signal frequencies and also the number
            // of nodes including dump nodes as these will later be used for computing quantum noise fluctuations
            // Dump nodes only have an input port and no output hence twice the reduced nodes.
            M_ifo_sig.M.num_eqns = (M_ifo_sig.num_frequencies
                                                * inter.num_fields
                                                * (2 * inter.num_reduced_nodes))
                                                + inter.num_motion_eqns + inter.num_slinks;
            
            alloc_ifo_matrix_ccs(&bytes, &M_ifo_sig);
            
            // only allocate dump nodes if we are computing the quantum noise fluctuations
            build_ccs_matrix(&M_ifo_sig);
            
            get_ccs_matrix(&M_ifo_sig);
            
            // set defaults for signal matrix computation 
            kobj = (klu_objs_t*)M_ifo_sig.solver_opts;
            
            klu_l_defaults(&(kobj->Common));

            kobj->Common.ordering = 0;
            kobj->Common.scale = 2;
            kobj->Common.btf = 1;
            kobj->Common.maxwork = 0;
            
        } else
            M_ifo_sig.M.num_eqns = -1;    
    } else 
        bug_error("spare_solver %i was not expected", options.sparse_solver);
    
    endTimer(tid);
    
    tid = startTimer("INIT_QUANT");
    init_quantum_components();
    endTimer(tid);
    
    // check the connections in the interferometer
    tid = startTimer("CHECK_CONN");
    check_connections();
    endTimer(tid);
    
    if (!options.servermode) {
        make_space();
    }
    
    // fill in the matrix
    tid = startTimer("SPARSE_FILL");
    
    if (options.sparse_solver == KLU) {
        fill_matrix(inter.light_in_list[0].f->f, STANDARD);
    } else if(options.sparse_solver == KLU_FULL) {
        
        fill_ccs_matrix(&M_ifo_car);
        
        if(M_ifo_sig.M.num_eqns > 0)
            fill_ccs_matrix(&M_ifo_sig);
        
    } else 
        bug_error("spare_solver %i was not expected", options.sparse_solver);
    
    endTimer(tid);
        
    // print matrix to file for looking at the sparsity for example 
    if (inter.printmatrix) {
        if (options.sparse_solver == KLU) {
            dump_matrix_klu(&klu, "klu_matrix.dat");
        } else if (options.sparse_solver == KLU_FULL){
            dump_matrix_ccs(&M_ifo_car.M, "klu_full_matrix_car.dat");
            dump_matrix_ccs(&M_ifo_sig.M, "klu_full_matrix_sig.dat");
        }else 
            bug_error("spare_solver %i was not expected", options.sparse_solver);
    }
    
    // Preorder KLU Matrix
    tid = startTimer("MATRIX_PREORDER");
    if (options.sparse_solver == KLU) {
        
        klu_old.Symbolic = klu_l_analyze(inter.num_eqns, klu.col_ptr, klu.row_index, &(klu_old.Common));
        klu_old.Numeric = klu_zl_factor(klu.col_ptr, klu.row_index, klu.value, klu_old.Symbolic, &(klu_old.Common));
        klu_zl_sort(klu_old.Symbolic, klu_old.Numeric, &(klu_old.Common));
        
    } else if(options.sparse_solver == KLU_FULL) {
        
        klu_objs_t *klu_obj = (klu_objs_t*) M_ifo_car.solver_opts;
        
        klu_obj->Symbolic = klu_l_analyze(M_ifo_car.M.num_eqns, M_ifo_car.M.col_ptr, M_ifo_car.M.row_idx, &(klu_obj->Common));
        
        if(klu_obj->Common.status != KLU_OK)
            gerror("An error occurred in KLU during analysis: STATUS=%i\n", klu_obj->Common.status);
        
        klu_obj->Numeric = klu_zl_factor(M_ifo_car.M.col_ptr, 
                                         M_ifo_car.M.row_idx,
                                         M_ifo_car.M.values,
                                         klu_obj->Symbolic, &(klu_obj->Common));
        
        if(klu_obj->Common.status == KLU_SINGULAR)
            gerror("Carrier matrix is singular!!\n", klu_obj->Common.status);
        else if(klu_obj->Common.status != KLU_OK)
            gerror("An error occurred in carrier matrix KLU during factorisation: STATUS=%i\n", klu_obj->Common.status);
        
        if(klu_obj->Common.status == KLU_OK)
            klu_zl_sort(klu_obj->Symbolic, klu_obj->Numeric, &(klu_obj->Common));
        
        if(M_ifo_sig.num_frequencies > 0){
            klu_obj = (klu_objs_t*)M_ifo_sig.solver_opts;
            
            klu_obj->Symbolic = klu_l_analyze(M_ifo_sig.M.num_eqns,
                                              M_ifo_sig.M.col_ptr,
                                              M_ifo_sig.M.row_idx,
                                              &(klu_obj->Common));
        
            if(klu_obj->Common.status != KLU_OK){
                gerror("An error occurred in signal matrix KLU during analysis: STATUS=%i\n", klu_obj->Common.status);
            }
            
            klu_obj->Numeric = klu_zl_factor(M_ifo_sig.M.col_ptr,
                                             M_ifo_sig.M.row_idx,
                                             M_ifo_sig.M.values, 
                                             klu_obj->Symbolic, &(klu_obj->Common));
            
            if(klu_obj->Common.status == KLU_SINGULAR)
                warn("Signal matrix is singular!!\n", klu_obj->Common.status);
            else if(klu_obj->Common.status != KLU_OK)
                gerror("An error occurred in KLU during factorisation: STATUS=%i\n", klu_obj->Common.status);

            if(klu_obj->Common.status == KLU_OK)
                klu_zl_sort(klu_obj->Symbolic, klu_obj->Numeric, &(klu_obj->Common));
        }
            
    } else 
        bug_error("spare_solver %i was not expected", options.sparse_solver);
    endTimer(tid);
    
    //assert(mem.restab != NULL);

    if (inter.x3_axis_is_set) {
        x3_out();
    } else if (options.servermode) {
#ifdef SERVERMODE
        initserver();
#else
        printf("Servermode not implemented in Windows binary\n");
#endif
    } else {
        x12_out();
    }


    //  if (OS == WIN)
    //    spa_destroy(); // Memory leak ???????????????????????????

    tid = startTimer("MATRIX_DESTROY");
    
    if (options.sparse_solver == KLU) {
        klu_destroy();
    } else if(options.sparse_solver == KLU_FULL){
        free_ccs_matrix(&M_ifo_car);
    } else 
        bug_error("spare_solver %i was not expected", options.sparse_solver);

    endTimer(tid);
    
    // TODO need to add more memory freeing but for now here is the mirror stuff
    //free_memory_for_mirror_list();

    // be nice and clean up after ourselves
    my_finish(0);
    
    endTimer(tid_main);
    
    closeTimerFile();
    
    return 0;
}

//! Prints the banner for Finesse

/*!
 * \param fp The file pointer for direction of the output
 * \param input_fname The input file name
 * \param output_fname The output file name
 * \param gnuplot_fname The name of the gnuplot file
 */
void banner(FILE *fp, char *input_fname, char *output_fname,
        char *gnuplot_fname) {
    now = time(NULL);

    char extra_space[80]="                                                     ";
    int num_spaces=33-strlen(VERSION)-strlen(GIT_REVISION);
    if (num_spaces<0 || num_spaces>20){
        num_spaces=0;
    }
    
    char debug_build[LINE_LEN] = {0};
    
#ifdef DEBUG
    sprintf(debug_build,"WARNING: USING DEBUG BUILD!!");
#else
    //sprintf(debug_build,"");
#endif
    
    // print the banner if not quiet
    fprintf(fp, "\n"
            "------------------------------------------------------------------------\n"
            "                     FINESSE %s %.*s (build %s)\n"
            "       o_.-=.        Frequency domain INterferomEter Simulation SoftwarE\n"
            "      (\\'\".\\|        %s         http://www.gwoptics.org/finesse/\n"
            "      .>' (_--.      %s\n"
            "   _=/d   ,^\\        Input file %s,\n"
            "  ~~ \\)-'   '        Output file %s,\n"
            "     / |             Gnuplot file %s \n"
            "    '  '                                        %s"
            "------------------------------------------------------------------------\n",
            VERSION, num_spaces, extra_space, GIT_REVISION, mydate,debug_build,
            get_path_i(input_fname,false), get_path_i(output_fname,false), get_path_i(gnuplot_fname,false),
            asctime(localtime(&now)));
}

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
