// $Id$

/*!
 * \file kat_calc.c
 * \brief The routines of the core calculations in Finesse
 * 
 * @section Copyright notice
 *
 *  This file is part of the interferometer simulation Finesse
 *  http://www.gwoptics.org/finesse
 *
 *  Copyright (C) 1999 onwards Andreas Freise
 *  with parts of the code written by Daniel Brown, Paul Cochrane
 *  and Gerhard Heinzel.
 *
 *  This program is free software; you can redistribute it and/or modify it under
 *  the terms of the GNU General Public License version 3 as published
 *  by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *  See the GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along with
 *  this library; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
 *  Suite 330, Boston, MA 02111-1307 USA
 */

#include "kat.h"
#include "kat_inline.c"
#include "kat_calc.h"
#include "kat_io.h"
#include "kat_gnu.h"
#include "kat_optics.h"
#include "kat_quant.h"
#include "kat_spa.h"
#include "kat_aux.h"
#include "kat_init.h"
#include "kat_aa.h"
#include "kat_matrix_ccs.h"
#include <errno.h>
#include <gsl/gsl_cblas.h>

extern init_variables_t init;
extern interferometer_t inter;
extern options_t options;
extern memory_t mem;
extern local_var_t vlocal;

extern global_var_t vglobal;
extern time_t now;

extern ifo_matrix_vars_t M_ifo_car; // carrier and modulator field matrix
extern ifo_matrix_vars_t M_ifo_sig;  // signal fields and optic degrees of freedom

extern klu_sparse_var_t klu;

extern int carrier_frequencies_tuned;
extern int signal_frequencies_tuned;

extern const complex_t complex_i;
extern const complex_t complex_1;
extern const complex_t complex_m1;
extern const complex_t complex_0;

static double dx; //!< A step distance in the x direction
static double dy; //!< A step distance in the y direction

static double x_axis_point; //!< Current distance along x-axis
static double y_axis_point; //!< Current distance along y-axis

int current_point; //!< The current point used for progress output
int current_point_tot; //!< Overall current point used for progress output

//! A flag indicating that result 'looks like' a transfer function to Finesse
extern int *transfer;

extern double *f_s; //!< Frequency value list [frequency]
extern complex_t ***a_s; //!< Amplitude list [field] [output] [frequency]
extern int *t_s; //!< Type_of_Signal list [frequency]

bool update_fields; //!< Flag to say whether or not to update the fields

FILE *ofp; //!< The output file pointer
FILE *gfp; //!< The gnuplot output file pointer
FILE *mfp; //!< The matlab output file pointer
FILE *pfp; //!< The python output file pointer

int compute_data_msg_flags = 0;

/**
 * Use this function to update the xaxis values. This also does other internal
 * checks to ensure that the negative xaxis values are fsigs are updated correctly
 * 
 * @param axis
 * @param value
 */
void update_xaxis_values(axis_t *axis, double value){
    
    *(axis->xaxis) = value;
    axis->minus_xaxis = -1.0 * value;
    
    if(axis->xaxis == &inter.fsig)
        inter.mfsig = axis->minus_xaxis;
    else if(axis->xaxis == &inter.mfsig)
        inter.fsig = axis->minus_xaxis;
}

/**
 * Given an xaxis object this returns the min/max values and the step size.
 * round argument can be used to either round or not round the step size
 * using round_to_dp.
 * 
 * @param axis
 * @param dx
 * @param x_min
 * @param x_max
 * @param round
 */
void get_xaxis_values(axis_t *axis, double *dx, double *x_min, double *x_max, int round){
    assert(axis != NULL);
    assert(dx != NULL);
    assert(x_min != NULL);
    assert(x_max != NULL);
    
    if (axis->xtype == FLIN) {
        *x_min = axis->xmin;
        *x_max = axis->xmax;
    } else if (axis->xtype == FLOG) {
        *x_min = log(axis->xmin);
        *x_max = log(axis->xmax);
    } else {
        bug_error("axis min max error %s", axis->xname);
    }

    if (axis->xsteps) {
        *dx = (*x_max - *x_min) / (axis->xsteps);
        if(round) *dx = round_to_dp(*dx);
    } else {
        *dx = 0.0;
    }
}

/**
 * Computes the Laplace transformed (s-domain) transfer function H(s).
 * Transfer function is defined by poles and zeros as stored
 * in the mech_transfer_func_t object.
 * 
 * @param mtf pointer to mechanical transfer function
 * @param s complex s-domain value
 * @return H(s)
 */
complex_t calc_transfer_func(transfer_func_t *mtf, complex_t s){
    assert(mtf != NULL);
    
    complex_t rtn = z_by_xph(complex_1, mtf->gain, mtf->phase);
    double omega;
    
    int i;
    if(mtf->num_zeros > 0){
        
        switch(mtf->type){
            case GENERAL_TF:
                for(i=0; i<mtf->num_zeros; i++)
                    rtn = z_by_z(rtn, z_m_z(s, mtf->zeros[i]));
                
                break;
                
            case Q_F_TF:
                
                for(i=0; i<mtf->num_zeros; i++){
                    omega = TWOPI*mtf->zero_Q_f[i].f;
                    assert(mtf->zero_Q_f[i].Q !=0 );
                    
                    rtn = z_by_z(rtn, z_pl_z(z_pl_z(z_by_x(s, omega/mtf->zero_Q_f[i].Q), z_by_z(s,s)), co(omega*omega,0))); 
                }
                
                break;
            default:
                bug_error("unexpected transfer function type");
        }
    }
    
    if(mtf->num_poles > 0){
        complex_t tmp = complex_1;
        
        switch(mtf->type){
            case GENERAL_TF:
                for(i=0; i<mtf->num_poles; i++)
                    tmp = z_by_z(tmp, z_m_z(s, mtf->poles[i]));
                
                break;
                
            case Q_F_TF:
                for(i=0; i<mtf->num_poles; i++){
                    omega = TWOPI*mtf->pole_Q_f[i].f;
                    assert(mtf->pole_Q_f[i].Q !=0 );
                    
                    tmp = z_by_z(tmp, z_pl_z(z_pl_z(z_by_x(s, omega/mtf->pole_Q_f[i].Q), z_by_z(s,s)), co(omega*omega,0))); 
                }
                
                break;
            default:
                bug_error("unexpected transfer function type");
        }
        
        rtn = div_complex(rtn, tmp);
    }
    
    return rtn;
}

//! Compute all data, function sweeps x-axis calling all computation functions

/*!
 *
 * \see Test_compute_data()
 */
void compute_data(void) {
    // initialise variables local to this file
    // when x3axis is set this compute_data is called multiple times
    if(!inter.x3_axis_is_set){
        current_point = 0;
        current_point_tot = 0;
    }
    
    dx = 0.0;
    dy = 0.0;

    // initialise transfer function flag array
    int light_out_index;
    for (light_out_index = 0;
            light_out_index < inter.num_light_outputs;
            light_out_index++) {
        transfer[light_out_index] = 0;
    }

    double x_min = 0.0;
    double x_max = 0.0;
    double y_min = 0.0;
    double y_max = 0.0;

    // initialising x-axis ------------------------------------------------
    get_xaxis_values(&inter.x1, &dx, &x_min, &x_max, 1);

    // initialising second x-axis for 3D plot --------------------------
    int oneline = 0;
    if (inter.splot) {
        get_xaxis_values(&inter.x2, &dy, &y_min, &y_max, 1);
        
        oneline = inter.num_output_cols * (inter.x2.xsteps + 1);
    }

    // trying to compute one datapoint with original setup
    // this is useful for locks .... 
    // maybe this should be below if(inter.splot) below???

    // setting the first value for x and y axes so that
    // the data point actually corresponds to the first value in xaxis and
    // not to the static setup.
    update_xaxis_values(&inter.x1, inter.x1.xmin);
    
    if (inter.splot) {
        update_xaxis_values(&inter.x2, inter.x2.xmin);
    }

    update_fields = true;
    tune_parameters();
    
    if(options.sparse_solver == KLU)
        data_point_old();
    else if(options.sparse_solver == KLU_FULL)
        data_point_new();
    else
        bug_error("sparse_sovler not handled");
    
    if (inter.debug & 8192) {
        //dump_matrix(STANDARD, "stdout");
        dump_matrix_mirror_elements(stdout);
        dump_matrix_space_elements(stdout);
    }

    // initialise detector outputs, this is needed for functions and locks that read detector signals
    compute_functions();
    compute_puts();
    fill_detectors();

    if (inter.beam.x && inter.beam.y && !(compute_data_msg_flags & 1) ) {
        compute_data_msg_flags |= 1;
        message("Tuning only beam detector x1/x2 -> computing interferometer only once.\n");
    }

    if (!inter.beam.x && inter.beam.y && !(compute_data_msg_flags & 2)) {
        compute_data_msg_flags |= 2;
        message("Tuning only beam detector along x2 -> computing interferometer only along x1.\n");
    }

    if (inter.beam.x && !inter.beam.y && inter.splot && !(compute_data_msg_flags & 4)) {
        compute_data_msg_flags |= 4;
        message("Consider switching x1 and x2 axis for speed up!\n");
    }

    // start running along x-axis -------------------------------------
    int x_step_index;
    for (x_step_index = 0; x_step_index <= inter.x1.xsteps; x_step_index++) {
        if (inter.x1.xtype == FLOG) {
            x_axis_point = exp(x_min + x_step_index * dx);
            // in order to avoid rounding error we explictly use the max value as last step
            if (x_step_index == inter.x1.xsteps) {
                x_axis_point = exp(x_max);
            }
        } else {
            x_axis_point = x_min + x_step_index*dx;
            
            // in order to avoid rounding error we explictly use the max value as last step
            if (x_step_index == inter.x1.xsteps) {
                x_axis_point = x_max;
            }
        }


        if (inter.beam.x) {
            update_fields = false;
        } else {
            update_fields = true;
        }

        // setting next x value ---------------------------------------
        update_xaxis_values(&inter.x1, x_axis_point);
        
        if (inter.debug & 4) {
            message("xx = %.15g\n", x_axis_point);
        }

        // 3D plot ? --------------------------------------------------
        if (inter.splot) {

            // compute data point now if x2-axis is just beam analyser axis 
            // (but x1 is not)
            if (inter.beam.y) {
                if (!inter.beam.x) {
                    if (update_fields) {
                        tune_parameters();
                    } else {
                        compute_functions();
                        compute_puts();
                        fill_detectors();
                        //                      compute_functions();                                
                    }
                }
                update_fields = false;
            } else {
                update_fields = true;
            }

            // start running along second x-axis ----------------------
            int y_step_index;
            for (y_step_index = 0; y_step_index <= inter.x2.xsteps; y_step_index++) {
                set_progress_action_text("Calculating");
                //set_progress_message_text("");
                
                if (inter.x2.xtype == FLOG) {
                    y_axis_point = exp(y_min + y_step_index * dy);
                    // in order to avoid rounding error we explictly use the max value as last step
                    if (y_step_index == inter.x2.xsteps) {
                        y_axis_point = exp(y_max);
                    }
                } else {
                    y_axis_point = y_min + y_step_index*dy;
                    // in order to avoid rounding error we explictly use the max value as last step
                    if (y_step_index == inter.x2.xsteps) {
                        y_axis_point = y_max;
                    }
                }

                // setting next x2 value -------------------------------
                update_xaxis_values(&inter.x2, y_axis_point);
                
                if (inter.debug & 4) {
                    message("yy = %.15g\n", y_axis_point);
                }

                vlocal.respt = mem.restab + x_step_index * oneline + y_step_index * inter.num_output_cols;

                // writing x2 values to output array -------------------
                if (inter.x1.xtype == FLOG) {
                    *(vlocal.respt++) = x_axis_point / inter.x1.op;
                } else {
                    *(vlocal.respt++) = x_axis_point - inter.x1.op;
                }

                if (inter.x2.xtype == FLOG) {
                    *(vlocal.respt++) = y_axis_point / inter.x2.op;
                } else {
                    *(vlocal.respt++) = y_axis_point - inter.x2.op;
                }
                
                // *** DO THE CALCULATION for this x2 value ----------
                //     (anwrite_resultd write output)
                write_result();
            }
        } else {
            set_progress_action_text("Calculating");
            //set_progress_message_text("");
            
            vlocal.respt = mem.restab + x_step_index * inter.num_output_cols;

            // writing x values to output array -----------------------
            if (inter.x1.xtype == FLOG) {
                *(vlocal.respt++) = x_axis_point / inter.x1.op;
            } else {
                *(vlocal.respt++) = x_axis_point - inter.x1.op;
            }

            // *** DO THE CALCULATION for this x1 value ----------
            //     (and write output)
            write_result();
        }
        
        set_progress_action_text("Calculating");

        if ((inter.trace & 64 || inter.debug & 32) && inter.rebuild == 2) {
            message("x=%s:\n", double_form(x_axis_point));
        }
    }
}


//! Write the result of a calculation to the memory array

/*!
 *
 * \see Test_write_result()
 */
void write_result(void) {
    if (inter.num_diff_cmds) {
        compute_derivative(vlocal.zsolution, 0); // compute a derivative of a y value 
    } else {
        compute_result(vlocal.zsolution); // compute the y value 
    }

    current_point_tot++;
    current_point++;

    if (current_point > init.percentage_step / 100.0 * init.num_points_total) {
        if (!options.quiet) {
            print_progress(init.num_points_total, current_point_tot); // print progress percents 
        }
        current_point = 0;
    }

    create_data_array(vlocal.zsolution); // write result to memory array  
}

//! Compute the derivative

/*!
 * \param zsol complex array of which to take the derivative
 * \param this_derivative_index index of the current differentation
 *
 *
 * \see Test_compute_derivative()
 */
void compute_derivative(complex_t *zsol, int this_derivative_index) {
    // sanity check on input
    assert(this_derivative_index >= 0);

    // compute new data points at which the function
    // value is to be computed 
    double back_step =         *inter.deriv_list[this_derivative_index].xderiv - init.deriv_h;
    double back_half_step =    *inter.deriv_list[this_derivative_index].xderiv - init.deriv_h / 2.0;
    double forward_step =      *inter.deriv_list[this_derivative_index].xderiv + init.deriv_h;
    double forward_half_step = *inter.deriv_list[this_derivative_index].xderiv + init.deriv_h / 2.0;
    double current_point =     *inter.deriv_list[this_derivative_index].xderiv;

    int next_derivative_index = this_derivative_index + 1;

    // check if data points are within the allowed range
    if ((back_half_step >= inter.deriv_list[this_derivative_index].min ||
            !inter.deriv_list[this_derivative_index].lborder) &&
            (forward_half_step <= inter.deriv_list[this_derivative_index].min ||
            !inter.deriv_list[this_derivative_index].uborder)) {
        *inter.deriv_list[this_derivative_index].xderiv = back_half_step;
        // do more differentiations? 
        if (next_derivative_index < inter.num_diff_cmds) {
            compute_derivative(vlocal.zderiv1, next_derivative_index);
        } else {
            compute_result(vlocal.zderiv1);
        }

        *inter.deriv_list[this_derivative_index].xderiv = forward_half_step;
        if (next_derivative_index < inter.num_diff_cmds) {
            compute_derivative(vlocal.zderiv2, next_derivative_index);
        } else {
            compute_result(vlocal.zderiv2);
        }
    } else {
        if (back_half_step < inter.deriv_list[this_derivative_index].min &&
                inter.deriv_list[this_derivative_index].lborder) {
            *inter.deriv_list[this_derivative_index].xderiv = current_point;
            if (next_derivative_index < inter.num_diff_cmds) {
                compute_derivative(vlocal.zderiv1, next_derivative_index);
            } else {
                compute_result(vlocal.zderiv1);
            }

            *inter.deriv_list[this_derivative_index].xderiv = forward_step;
            if (next_derivative_index < inter.num_diff_cmds) {
                compute_derivative(vlocal.zderiv2, next_derivative_index);
            } else {
                compute_result(vlocal.zderiv2);
            }
        } else {
            *inter.deriv_list[this_derivative_index].xderiv = back_step;
            if (next_derivative_index < inter.num_diff_cmds) {
                compute_derivative(vlocal.zderiv1, next_derivative_index);
            } else {
                compute_result(vlocal.zderiv1);
            }

            *inter.deriv_list[this_derivative_index].xderiv = current_point;
            if (next_derivative_index < inter.num_diff_cmds) {
                compute_derivative(vlocal.zderiv2, next_derivative_index);
            } else {
                compute_result(vlocal.zderiv2);
            }
        }
    }

    *inter.deriv_list[this_derivative_index].xderiv = current_point;
    if (inter.debug & 2048) {
        message("x +-h= %g %g\n", back_step, forward_step);
    }

    int output_index;
    for (output_index = 0; output_index < inter.num_outputs; output_index++) {
        zsol[output_index] = z_by_x(z_m_z(vlocal.zderiv2[output_index],
                vlocal.zderiv1[output_index]),
                1.0 / init.deriv_h);
        if (inter.debug & 2048) {
            message("zsol=%s, zabs=%g\n",
                    complex_form15(zsol[output_index]), zabs(zsol[output_index]));
        }
    }
}

//! Compute the result of one given data set (one point along the axis)

/*!
 * Modifies the argument in place.
 *
 * \param z_tmp input complex array (result returned here also)
 *
 * \param z_tmp array of output signals
 *
 * \see Test_compute_result()
 */
void compute_result(complex_t *z_tmp) {

    if (update_fields) {
        // perform all locks, funcs, puts etc and then call solve the matrix and
        // compute the detector signals
        tune_parameters();
    } else {
        // use the previous solution and just re-compute the detector signals
        compute_functions();
        compute_puts();
        fill_detectors();
    }

    // for each defined output set the signal calculated via fill_detectors()
    int i;
    for (i = 0; i < inter.num_outputs; i++) {
        z_tmp[i] = inter.output_data_list[i].signal;
    }
}

//! Compute fsig amplitude for SIG_X SIG_Y signals

/*!
 * \param sig_node signal node
 * \param q Gaussian beam q parameter
 *
 * \todo sig_node -> signal_node
 */
complex_t xyamp(node_t *sig_node, complex_t q) {
    complex_t txs;
    double nr, w0;

    nr = *(sig_node->n); 
    // TODO fix the amplitude below (i.e. add equation to manual)
    // the factor of two here is empirical but NOT UNDERSTOOD
    // should actually be * (-1) 
    w0 = 2.0 / w0_size(q, nr);
    txs = z_by_x(q, w0);
    //printf("node: %s, q: %s, amp: %s (abs: %g angle %g)\n",sig_node.name,complex_form(q),complex_form(txs),zabs(txs), zdeg(txs)); 
    return (txs);
}

//! Compute the RHS vector elements for signal modulation 

/*!
 * \param f actual frequency (carrier + signal)
 * \param f1 carrier frequency
 * \param carrier flag indicating what type f1 is
 * \param lower2 flag: +1 (-1) upper (lower) signal sideband
 * \param modulator_index index of modulator component
 * \param modulation_order order of modulation sideband of the modulator
 * \param lower1 flag: +1 (-1) upper (lower) modulator sideband
 * \param z field amplitude at modulator
 *
 * \see Test_signals_rhs()
 */
void signals_rhs(double f, double f1, int carrier, int lower2,
        int modulator_index, int modulation_order, int lower1, complex_t z) {
    // sanity checks on inputs
    assert(carrier == CARRIER_FIELD || carrier == MODULATOR_FIELD);
    assert(lower2 == 1 || lower2 == -1);

    assert(modulator_index >= 0);
    // inter.num_modulators can be zero in which case the upper bound check
    // will fail incorrectly, so if it is nonzero, run the assert, otherwise
    // continue
#ifndef NDEBUG
    if (inter.num_modulators > 0) {
        assert(modulator_index < inter.num_modulators);
    }
#endif
    
    // loop over all signals
    int signal_index;
    for (signal_index = 0; signal_index < inter.num_signals; signal_index++) {
        signal_t *signal;
        signal = &(inter.signal_list[signal_index]);

        double factor = signal->amplitude * EPSILON * 0.5; //Epsilon/2 for fsig
        double phase = (90.0 + lower2 * signal->phase);

        switch (signal->component_type) {
            case BEAMSPLITTER:
                set_beamsplitter_signal_rhs(signal, factor, phase, f, f1, lower2);
                break;
            case MIRROR:
                set_mirror_signal_rhs(signal, factor, phase, f, f1, lower2);
                break;
            case SPACE:
                set_space_signal_rhs(signal, factor, phase, f1, lower2);
                break;
            case LIGHT_INPUT:
                set_light_input_signal_rhs(signal, factor, phase, carrier);
                break;
            case MODULATOR:
                set_modulator_signal_rhs(signal, factor, phase, lower1, carrier, modulator_index, modulation_order, z);
                break;
            default:
                bug_error("rhs_sig");
        }
    }
}

void calc_beamsplitter_signal_vars( beamsplitter_t *beamsplitter, double *factor, double *ph1, double f, double f_c){
    // TODO check refractive index
    *factor *= 2.0 * (1.0 + f_c / inter.f0)
                * sqrt(beamsplitter->R)
                * cos(beamsplitter->alpha * RAD);
    
    *ph1 = beamsplitter->phi
                * ((1.0 + f / inter.f0) + (1.0 + f_c / inter.f0))
                * cos(beamsplitter->alpha * RAD);
}

complex_t calc_reflect_tilt_signal(signal_t *signal, complex_t **knm, int in, int out, complex_t X_bar){
    complex_t mult = complex_0;
                
    int l, tn=0, tm=0;
    
    get_tem_modes_from_field_index(&tn, &tm, in);
    
    // do the matrix multiplication, alignment knm matrix is very sparse though
    // as each mode will only couple into at most 2 other modes. This is due to
    // the assumption that the fsig amplitude is small (i.e. small angle) and the
    // coupling is linear
    if(signal->type == SIG_X) {
        l = get_field_index_from_tem(tn + 1, tm);

        if(l < inter.num_fields) { // ensure not coupling to mode higher than we are simulating
             mult = z_pl_z(mult, z_by_z( z_by_x(X_bar, sqrt(tn)), knm[l][out]));
             printf("A [%i %i] X [%i %i]\n", in, tn+1, l, out);
        }
        
        if(tn > 0){
             l = get_field_index_from_tem(tn - 1, tm);
             mult = z_pl_z(mult, z_by_z( z_by_x(X_bar, sqrt(tn)), knm[l][out]));
             printf("B [%i %i] X [%i %i]\n", in, tn, l, out);
        }

    } else if(signal->type == SIG_Y) {
        l = get_field_index_from_tem(tn, tm+1);

        if(l < inter.num_fields) // ensure not coupling to mode higher than we are simulating
             mult = z_pl_z(mult, z_by_z( z_by_x(X_bar, -1.0*sqrt(tm+1)), knm[l][out]));

        if(tn > 0){
             l = get_field_index_from_tem(tn, tm - 1);
             mult = z_pl_z(mult, z_by_z( z_by_x(X_bar, sqrt(tm)), knm[l][out]));
        }
    }
    
    return mult;
}

void fill_bs_signal_rhs(signal_t *signal, ifo_matrix_vars_t* M_car, ifo_matrix_vars_t* M_sig){
    assert(M_sig != NULL);
    assert(M_car != NULL);
    assert(signal != NULL);
    
    int f,i, tn,tm;
    beamsplitter_t *bs = &inter.bs_list[signal->component_index];
    
    node_t *node1 = &inter.node_list[bs->node1_index];
    node_t *node2 = &inter.node_list[bs->node2_index];
    node_t *node3 = &inter.node_list[bs->node3_index];
    node_t *node4 = &inter.node_list[bs->node4_index];
    
    for(i=0; i<bs->num_motions; i++){
        int xidx = bs->x_rhs_idx + i;
        
        switch(bs->motion_type[i]) {
            case Z:
                if(signal->type == SIG_PHS) {
                    assert(bs->long_tf != NULL);
                    double k = TWOPI/init.lambda;
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/(init.x_scale*k), signal->phase);
                } else if(signal->type == SIG_Z) {
                    assert(bs->long_tf != NULL);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/init.x_scale, signal->phase);
                } else if(signal->type == SIG_FZ) {
                    assert(bs->long_tf != NULL);
                    // compute transfer function for positive signal frequency
                    complex_t G_Omega = z_by_x(calc_transfer_func(bs->long_tf, co(0, TWOPI *inter.fsig)), 1.0/bs->mass);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(G_Omega, signal->amplitude/init.x_scale, signal->phase);
                }
                break;
            case ROTX:
                if(signal->type == SIG_X){
                    assert(bs->rot_x_tf != NULL);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/init.x_scale, signal->phase);
                } else if(signal->type == SIG_FRX) {
                    assert(bs->rot_x_tf != NULL);
                    // compute transfer function for positive signal frequency
                    complex_t G_Omega = z_by_x(calc_transfer_func(bs->rot_x_tf, co(0, TWOPI *inter.fsig)), 1.0/bs->Ix);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(G_Omega, signal->amplitude/init.x_scale, signal->phase);
                }
                break;
                
            case ROTY:
                if(signal->type == SIG_Y) {
                    assert(bs->rot_y_tf != NULL);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/init.x_scale, signal->phase); 
                } else if(signal->type == SIG_FRY) {
                    assert(bs->rot_y_tf != NULL);
                    // compute transfer function for positive signal frequency
                    complex_t G_Omega = z_by_x(calc_transfer_func(bs->rot_y_tf, co(0, TWOPI *inter.fsig)), 1.0/bs->Iy);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(G_Omega, signal->amplitude/init.x_scale, signal->phase);
                }
                break;
            default:
                bug_error("not handled");
        }
    }
    
    if((signal->type == SIG_AMP || signal->type == SIG_PHS) && bs->mass == 0.0){
        for(f=0; f<M_sig->num_frequencies; f++){
            frequency_t *f_sig = M_sig->frequencies[f];
            double phase = (90.0 + f_sig->order * signal->phase);
            double factor = signal->amplitude * EPSILON * 0.5; //Epsilon/2 for fsig
            double turn, ph1;

            if(signal->type == SIG_AMP)
                phase -= 90;

            calc_beamsplitter_signal_vars(bs, &factor, &ph1, f_sig->f, f_sig->carrier->f);

            if(!node1->gnd_node && !node2->gnd_node){
                double phase1 = phase + ph1;

                // get the index for the signal field of frequency f and mode i and 
                // the carrier field which will be its source
                int sidx12 = M_sig->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(2, f, 0, M_sig->num_frequencies);
                int cidx12 = M_car->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, 0, M_car->num_frequencies);

                int sidx21 = M_sig->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(2, f, 0, M_sig->num_frequencies);
                int cidx21 = M_car->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, 0, M_car->num_frequencies);

                // idea is we compute: signal fields = factor*exp(i*phase)) * (knm * carrier fields)
                complex_t *sig_vec = &(((complex_t*)M_sig->rhs_values)[sidx12]);
                complex_t *car_vec = &(((complex_t*)M_car->rhs_values)[cidx12]);

                cblas_zgemv(CblasRowMajor, CblasNoTrans, inter.num_fields, inter.num_fields,
                        &complex_1, &(bs->knm.k12[0][0]), inter.num_fields, car_vec, 1, &complex_0, sig_vec, 1);

                sig_vec = &(((complex_t*)M_sig->rhs_values)[sidx21]);
                car_vec = &(((complex_t*)M_car->rhs_values)[cidx21]);

                cblas_zgemv(CblasRowMajor, CblasNoTrans, inter.num_fields, inter.num_fields,
                        &complex_1, &(bs->knm.k21[0][0]), inter.num_fields, car_vec, 1, &complex_0, sig_vec, 1);

                // now apply all the phase factors etc.
                for(i=0; i<inter.num_fields;i++){
                    get_tem_modes_from_field_index(&tn, &tm, i);
                    turn = turnit(tn);

                    sidx12 =  M_sig->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(2, f, i, M_sig->num_frequencies);
                    ((complex_t*)M_sig->rhs_values)[sidx12] = z_by_xph(((complex_t*)M_sig->rhs_values)[sidx12], turn * factor, phase1);

                    sidx21 =  M_sig->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(2, f, i, M_sig->num_frequencies);
                    ((complex_t*)M_sig->rhs_values)[sidx21] = z_by_xph(((complex_t*)M_sig->rhs_values)[sidx21], turn * factor, phase1);

                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) { 
                        ((complex_t*)M_sig->rhs_values)[sidx21].im *= -1;
                        ((complex_t*)M_sig->rhs_values)[sidx12].im *= -1;
                    }
                }
            }

            if(!node3->gnd_node && !node4->gnd_node){
                double phase2 = phase - ph1 + 180.0 * f_sig->order;
                // check if we need to conjugate the value for the lower sideband
                if(f_sig->order == -1)
                    phase2 *= -1;

                // get the index for the signal field of frequency f and mode i and 
                // the carrier field which will be its source
                int sidx34 = M_sig->node_rhs_idx_1[node4->list_index] + get_node_rhs_idx(2, f, 0, M_sig->num_frequencies);
                int cidx34 = M_car->node_rhs_idx_1[node3->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, 0, M_car->num_frequencies);

                int sidx43 = M_sig->node_rhs_idx_1[node3->list_index] + get_node_rhs_idx(2, f, 0, M_sig->num_frequencies);
                int cidx43 = M_car->node_rhs_idx_1[node4->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, 0, M_car->num_frequencies);

                // idea is we compute: signal fields = factor*exp(i*phase)) * (knm * carrier fields)
                complex_t *sig_vec = &(((complex_t*)M_sig->rhs_values)[sidx34]);
                complex_t *car_vec = &(((complex_t*)M_car->rhs_values)[cidx34]);

                cblas_zgemv(CblasRowMajor, CblasNoTrans, inter.num_fields, inter.num_fields,
                        &complex_1, &(bs->knm.k34[0][0]), inter.num_fields, car_vec, 1, &complex_0, sig_vec, 1);

                sig_vec = &(((complex_t*)M_sig->rhs_values)[sidx43]);
                car_vec = &(((complex_t*)M_car->rhs_values)[cidx43]);

                cblas_zgemv(CblasRowMajor, CblasNoTrans, inter.num_fields, inter.num_fields,
                        &complex_1, &(bs->knm.k43[0][0]), inter.num_fields, car_vec, 1, &complex_0, sig_vec, 1);

                // now apply all the phase factors etc.
                for(i=0; i<inter.num_fields;i++){
                    get_tem_modes_from_field_index(&tn, &tm, i);
                    turn = turnit(tn);

                    sidx34 =  M_sig->node_rhs_idx_1[node4->list_index] + get_node_rhs_idx(2, f, i, M_sig->num_frequencies);
                    ((complex_t*)M_sig->rhs_values)[sidx34] = z_by_xph(((complex_t*)M_sig->rhs_values)[sidx34], turn * factor, phase2);

                    sidx43 =  M_sig->node_rhs_idx_1[node3->list_index] + get_node_rhs_idx(2, f, i, M_sig->num_frequencies);
                    ((complex_t*)M_sig->rhs_values)[sidx43] = z_by_xph(((complex_t*)M_sig->rhs_values)[sidx43], turn * factor, phase2);

                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1){
                        ((complex_t*)M_sig->rhs_values)[sidx43].im *= -1;
                        ((complex_t*)M_sig->rhs_values)[sidx34].im *= -1;
                    }
                }
            }
        }
    
    } else if((signal->type == SIG_X && bs->Ix == 0) || (signal->type == SIG_Y && bs->Iy == 0)) { 

        double ph2;
        if (signal->type == SIG_Y) {
            ph2 = 180.0;
        } else {
            ph2 = 0.0;
        }

        for(f=0; f<M_sig->num_frequencies; f++){
            frequency_t *f_sig = M_sig->frequencies[f];
            double phase = f_sig->order * signal->phase;
            double factor = signal->amplitude * EPSILON * 0.5; //Epsilon/2 for fsig
            double turn = 1.0, ph1 = 0.0;

            complex_t txs1 = complex_0;
            complex_t txs2 = complex_0;
            complex_t txs3 = complex_0;
            complex_t txs4 = complex_0;

            complex_t q = complex_0;

            if ((!node1->gnd_node) && (!node2->gnd_node)) {
                if (signal->type == SIG_X) {
                    q = node1->qx;
                } else {
                    q = node1->qy;
                }

                if (node1->component_index == signal->component_index) {
                    q = cminus(cconj(q));
                }
                txs1 = xyamp(node1, q);

                if (signal->type == SIG_X) {
                    q = node2->qx;
                } else {
                    q = node2->qy;
                }
                if (node2->component_index == signal->component_index) {
                    q = cminus(cconj(q));
                }
                txs2 = xyamp(node2, q);
            }

            if (!node3->gnd_node && !node4->gnd_node) {
                if (signal->type == SIG_X) {
                    q = node3->qx;
                } else {
                    q = node3->qy;
                }
                if (node3->component_index == signal->component_index) {
                    q = cminus(cconj(q));
                }
                txs3 = xyamp(node3, q);

                if (signal->type == SIG_X) {
                    q = node4->qx;
                } else {
                    q = node4->qy;
                }
                if (node4->component_index == signal->component_index) {
                    q = cminus(cconj(q));
                }
                txs4 = xyamp(node4, q);
            }

            // the old Gouy phase problem again: I should now do a proper
            // rev_gouy with txs1 and txs2, but I think it will just 
            // produce a constant phase of 90 deg. So this is what I do instead.

            txs1 = co(0, zmod(txs1));
            txs2 = co(0, zmod(txs2));
            txs3 = co(0, zmod(txs3));
            txs4 = co(0, zmod(txs4));

            int n, i, j;
            int cidx, sidx;

            complex_t *crhs = (complex_t*)M_car->rhs_values;
            complex_t *srhs = (complex_t*)M_sig->rhs_values;

            // compute purely reflected carrier field, do this by taking the incoming beam and computing matrix multiplication
            // with the reflection elements of the matrix. Store reflected field in memory already allocated for this in the
            // signal structure
            for(i=0; i < inter.num_fields; i++){
                signal->ao1[i] = complex_0;
                signal->ao2[i] = complex_0;
                signal->ao3[i] = complex_0;
                signal->ao4[i] = complex_0;

                for(j=0; j < inter.num_fields; j++){
                    if(!node1->gnd_node && !node2->gnd_node) {
                        cidx = M_car->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, j, M_car->num_frequencies);
                        signal->ao1[i] = z_pl_z(signal->ao1[i], z_by_z(*bs->a21f[f_sig->carrier->index][j][i], crhs[cidx]));

                        cidx = M_car->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, j, M_car->num_frequencies);
                        signal->ao2[i] = z_pl_z(signal->ao2[i], z_by_z(*bs->a12f[f_sig->carrier->index][j][i], crhs[cidx]));
                    }

                    if(!node3->gnd_node && !node4->gnd_node) {
                        cidx = M_car->node_rhs_idx_1[node4->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, j, M_car->num_frequencies);
                        signal->ao3[i] = z_pl_z(signal->ao3[i], z_by_z(*bs->a43f[f_sig->carrier->index][j][i], crhs[cidx]));

                        cidx = M_car->node_rhs_idx_1[node3->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, j, M_car->num_frequencies);
                        signal->ao4[i] = z_pl_z(signal->ao4[i], z_by_z(*bs->a34f[f_sig->carrier->index][j][i], crhs[cidx]));
                    }
                }
            }

            for (n = 0; n < inter.num_fields; n++) {
                int tn, tm;
                int field;
                //double turn;
                get_tem_modes_from_field_index(&tn, &tm, n);
                // TODO fix this! The `turn' below should be fixed once this is better understood!
                //turn = turnit(tn);
                //turn = 1.0;

                complex_t a12 = complex_0;
                complex_t a21 = complex_0;
                complex_t a34 = complex_0;
                complex_t a43 = complex_0;

                if (tn && signal->type == SIG_X) {
                    field = get_field_index_from_tem(tn - 1, tm);
                    if(!node1->gnd_node && !node2->gnd_node){
                        a12 = z_by_x(z_by_z(signal->ao2[field], cconj(txs2)), sqrt(tn));
                        a21 = z_by_x(z_by_z(signal->ao1[field], cconj(txs1)), sqrt(tn));
                    }

                    if(!node3->gnd_node && !node4->gnd_node){
                        a34 = z_by_x(z_by_z(signal->ao4[field], cconj(txs4)), sqrt(tn));
                        a43 = z_by_x(z_by_z(signal->ao3[field], cconj(txs3)), sqrt(tn));
                    }
                } else if (tm && signal->type == SIG_Y) {
                    field = get_field_index_from_tem(tn, tm - 1);

                    if(!node1->gnd_node && !node2->gnd_node){
                        a12 = z_by_x(z_by_z(signal->ao2[field], cconj(txs2)), sqrt(tm));
                        a21 = z_by_x(z_by_z(signal->ao1[field], cconj(txs1)), sqrt(tm));
                    }

                    if(!node3->gnd_node && !node4->gnd_node){
                        a34 = z_by_x(z_by_z(signal->ao4[field], cconj(txs4)), sqrt(tm));
                        a43 = z_by_x(z_by_z(signal->ao3[field], cconj(txs3)), sqrt(tm));
                    }
                }

                int ttn, ttm;
                if (tn + tm < inter.tem) {
                    if (signal->type == SIG_X) {
                        ttn = tn + 1;
                        ttm = tm;
                    } else {
                        ttn = tn;
                        ttm = tm + 1;
                    }

                    field = get_field_index_from_tem(ttn, ttm);

                    if(!node1->gnd_node && !node2->gnd_node){
                        a12 = z_pl_z(a12,z_by_x(z_by_z(signal->ao2[field], txs2), -1.0 * sqrt(ttn)));
                        a21 = z_pl_z(a21,z_by_x(z_by_z(signal->ao1[field], txs1), -1.0 * sqrt(ttn)));
                    }

                    if(!node3->gnd_node && !node4->gnd_node){
                        a34 = z_pl_z(a34,z_by_x(z_by_z(signal->ao4[field], txs4), -1.0 * sqrt(ttn)));
                        a43 = z_pl_z(a43,z_by_x(z_by_z(signal->ao3[field], txs3), -1.0 * sqrt(ttn)));
                    }
                }

                if (!node1->gnd_node) {
                    sidx = M_sig->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(2, f, n, M_sig->num_frequencies);
                    srhs[sidx] = z_by_xph(a21, turn * factor, phase + ph1 + ph2);
                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) srhs[sidx].im *= -1;
                }

                if (!node2->gnd_node) {
                    sidx = M_sig->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(2, f, n, M_sig->num_frequencies);
                    srhs[sidx] = z_by_xph(a12, turn * factor, phase + ph1 + ph2);
                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) srhs[sidx].im *= -1;
                }

                if (!node3->gnd_node) {
                    sidx = M_sig->node_rhs_idx_1[node3->list_index] + get_node_rhs_idx(2, f, n, M_sig->num_frequencies);
                    srhs[sidx] = z_by_xph(a43, turn * factor, phase + ph1 + ph2);
                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) srhs[sidx].im *= -1;
                }

                if (!node4->gnd_node) {
                    sidx = M_sig->node_rhs_idx_1[node4->list_index] + get_node_rhs_idx(2, f, n, M_sig->num_frequencies);
                    srhs[sidx] = z_by_xph(a34, turn * factor, phase + ph1 + ph2);
                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) srhs[sidx].im *= -1;
                }
            }
        }
    }
}

/** Set the RHS vector signal elements for a beamsplitter
 * \param signal the signal object to set the elements for
 * \param factor an amplification factor
 * \param phase base phase of the signal ???
 * \param f a base frequency ???
 * \param f1 another base frequency ???
 * \param sideband_index index indicating which sideband is being processed
 *
 * \todo large chunks of this untested
 *
 * \see Test_set_beamsplitter_signal_rhs()
 */
void set_beamsplitter_signal_rhs(
        signal_t *signal,
        double factor,
        double phase,
        double f,
        double f1,
        int sideband_index) {
    assert(sideband_index == 1 || sideband_index == -1);

    node_t signal_node1 = inter.node_list[*(signal->node1_index)];
    node_t signal_node2 = inter.node_list[*(signal->node2_index)];
    node_t signal_node3 = inter.node_list[*(signal->node3_index)];
    node_t signal_node4 = inter.node_list[*(signal->node4_index)];

    beamsplitter_t *beamsplitter;
    beamsplitter = &inter.bs_list[signal->component_index];
        
    if (signal->type == SIG_AMP || signal->type == SIG_PHS) {

        // amplitude modulation (test)
        if (signal->type == SIG_AMP) {
            phase = phase - 90.0;
        }

        double ph1;

        calc_beamsplitter_signal_vars(beamsplitter, &factor, &ph1, f, f1);
        
        int field_index;
        int tn, tm;
        double turn;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            get_tem_modes_from_field_index(&tn, &tm, field_index);
            turn = turnit(tn);

            if (NOT signal_node1.gnd_node) {
                set_rhs(2 * inter.num_fields * signal->node1_reduced_index +
                        signal->node1_amp_dir->fro + 2 * field_index,
                        z_by_ph(z_by_x(signal->a1[field_index], turn * factor), phase + ph1),
                        STANDARD);
            }

            if (NOT signal_node2.gnd_node) {
                set_rhs(2 * inter.num_fields * signal->node2_reduced_index +
                        signal->node2_amp_dir->fro + 2 * field_index,
                        z_by_ph(z_by_x(signal->a2[field_index], turn * factor), phase + ph1),
                        STANDARD);
            }

            if (NOT signal_node3.gnd_node) {
                set_rhs(2 * inter.num_fields * signal->node3_reduced_index +
                        signal->node3_amp_dir->fro + 2 * field_index,
                        z_by_ph(z_by_x(signal->a3[field_index], turn * factor), phase - ph1 + 180.0 * sideband_index),
                        STANDARD);
            }

            if (NOT signal_node4.gnd_node) {
                set_rhs(2 * inter.num_fields * signal->node4_reduced_index +
                        signal->node4_amp_dir->fro + 2 * field_index,
                        z_by_ph(z_by_x(signal->a4[field_index], turn * factor), phase - ph1 + 180.0 * sideband_index),
                        STANDARD);
            }
        }
    } else if (signal->type == SIG_X || signal->type == SIG_Y) {
        double ph1 = 0.0;

        // switch back to amplitude modulation
        phase = phase - 90.0;

        complex_t txs1 = complex_0;
        complex_t txs2 = complex_0;
        complex_t txs3 = complex_0;
        complex_t txs4 = complex_0;

        complex_t q = complex_0;

        if ((NOT signal_node1.gnd_node) && (NOT signal_node2.gnd_node)) {
            if (signal->type == SIG_X) {
                q = signal_node1.qx;
            } else {
                q = signal_node1.qy;
            }

            if (signal_node1.component_index == signal->component_index) {
                q = cminus(cconj(q));
            }
            txs1 = xyamp(&signal_node1, q);

            if (signal->type == SIG_X) {
                q = signal_node2.qx;
            } else {
                q = signal_node2.qy;
            }
            if (signal_node2.component_index == signal->component_index) {
                q = cminus(cconj(q));
            }
            txs2 = xyamp(&signal_node2, q);
        }

        if ((NOT signal_node3.gnd_node) && (NOT signal_node4.gnd_node)) {
            if (signal->type == SIG_X) {
                q = signal_node3.qx;
            } else {
                q = signal_node3.qy;
            }
            if (signal_node3.component_index == signal->component_index) {
                q = cminus(cconj(q));
            }
            txs3 = xyamp(&signal_node3, q);

            if (signal->type == SIG_X) {
                q = signal_node4.qx;
            } else {
                q = signal_node4.qy;
            }
            if (signal_node4.component_index == signal->component_index) {
                q = cminus(cconj(q));
            }
            txs4 = xyamp(&signal_node4, q);
        }

        // the old Gouy phase problem again: I should now do a proper
        // rev_gouy with txs1 and txs2, but I think it will just 
        // produce a constant phase of 90 deg. So this is what I do instead.

        txs1 = co(0, zmod(txs1));
        txs2 = co(0, zmod(txs2));
        txs3 = co(0, zmod(txs3));
        txs4 = co(0, zmod(txs4));


        int n;
        for (n = 0; n < inter.num_fields; n++) {
            int tn, tm;
            double turn;
            get_tem_modes_from_field_index(&tn, &tm, n);
            // TODO fix this! The `turn' below should be fixed once this is better understood!
            //turn = turnit(tn);
            turn = 1.0;

            complex_t a12 = complex_0;
            complex_t a21 = complex_0;
            complex_t a34 = complex_0;
            complex_t a43 = complex_0;

            if (tn && signal->type == SIG_X) {
                a12 = z_by_x(z_by_z(signal->ao1[get_field_index_from_tem(tn - 1, tm)],
                        cconj(txs2)), sqrt(tn));
                a21 = z_by_x(z_by_z(signal->ao2[get_field_index_from_tem(tn - 1, tm)],
                        cconj(txs1)), sqrt(tn));
                a34 = z_by_x(z_by_z(signal->ao3[get_field_index_from_tem(tn - 1, tm)],
                        cconj(txs4)), sqrt(tn));
                a43 = z_by_x(z_by_z(signal->ao4[get_field_index_from_tem(tn - 1, tm)],
                        cconj(txs3)), sqrt(tn));
            } else if (tm && signal->type == SIG_Y) {
                a12 = z_by_x(z_by_z
                        (signal->ao1[get_field_index_from_tem(tn, tm - 1)],
                        cconj(txs2)), sqrt(tm));
                a21 = z_by_x(z_by_z
                        (signal->ao2[get_field_index_from_tem(tn, tm - 1)],
                        cconj(txs1)), sqrt(tm));
                a34 = z_by_x(z_by_z
                        (signal->ao3[get_field_index_from_tem(tn, tm - 1)],
                        cconj(txs4)), sqrt(tm));
                a43 = z_by_x(z_by_z
                        (signal->ao4[get_field_index_from_tem(tn, tm - 1)],
                        cconj(txs3)), sqrt(tm));
            }

            int ttn, ttm;
            if (tn + tm < inter.tem) {
                if (signal->type == SIG_X) {
                    ttn = tn + 1;
                    ttm = tm;
                } else {
                    ttn = tn;
                    ttm = tm + 1;
                }
                a12 = z_pl_z(a12,
                        z_by_x(z_by_z
                        (signal->ao1[get_field_index_from_tem(ttn, ttm)],
                        txs2), -1.0 * sqrt(ttn)));
                a21 = z_pl_z(a21,
                        z_by_x(z_by_z
                        (signal->ao2[get_field_index_from_tem(ttn, ttm)],
                        txs1), -1.0 * sqrt(ttn)));
                a34 = z_pl_z(a34,
                        z_by_x(z_by_z
                        (signal->ao3[get_field_index_from_tem(ttn, ttm)],
                        txs4), -1.0 * sqrt(ttn)));
                a43 = z_pl_z(a43,
                        z_by_x(z_by_z
                        (signal->ao4[get_field_index_from_tem(ttn, ttm)],
                        txs3), -1.0 * sqrt(ttn)));
            }

            double ph2;
            if (signal->type == SIG_Y) {
                ph2 = 180.0;
            } else {
                ph2 = 0.0;
            }

            if (NOT signal_node2.gnd_node) {
                // set rhs vector element for node1
                set_rhs(2 * inter.num_fields * signal->node2_reduced_index
                        + signal->node2_amp_dir->fro + 2 * n,
                        z_by_ph(z_by_x(a12, turn * factor), phase + ph1 + ph2),
                        STANDARD);
            }

            if (NOT signal_node1.gnd_node) {
                // set rhs vector element for node2
                set_rhs(2 * inter.num_fields * signal->node1_reduced_index
                        + signal->node1_amp_dir->fro + 2 * n,
                        z_by_ph(z_by_x(a21, turn * factor), phase + ph1 + ph2),
                        STANDARD);

                if (NOT signal_node4.gnd_node) {
                    // set rhs vector element for node1
                    set_rhs(2 * inter.num_fields * signal->node4_reduced_index
                            + signal->node4_amp_dir->fro + 2 * n,
                            z_by_ph(z_by_x(a34, turn * factor), phase + ph1 + ph2),
                            STANDARD);
                }

                if (NOT signal_node3.gnd_node) {
                    // set rhs vector element for node2
                    set_rhs(2 * inter.num_fields * signal->node3_reduced_index
                            + signal->node3_amp_dir->fro + 2 * n,
                            z_by_ph(z_by_x(a43, turn * factor), phase + ph1 + ph2),
                            STANDARD);
                }
            }
        }
    } else {
        bug_error("wrong signal type for bs");
    }

    //! \todo return from here one day!!!

}

void calc_mirror_signal_vars(signal_t *signal, mirror_t *mirror, double *factor, double *phase, double *ph1, double f, double f_c){
    assert(signal != NULL);
    assert(mirror != NULL);
    assert(factor != NULL);
    assert(phase != NULL);
    assert(ph1 != NULL);
    
    // amplitude modulation (test)
    if (signal->type == SIG_AMP) {
        *phase = *phase - 90.0;
        // energy conserving amplitude modulation uses EPSILON/4
        *factor *= 0.5 * sqrt(mirror->R);
    } else if(signal->type == SIG_PHS) {
        *factor *= 2.0 * (1 + f_c / inter.f0) * sqrt(mirror->R);
    } else if(signal->type == SIG_Z) {
        double k = TWOPI / init.lambda;
        *factor *= 2.0 * k *(1 + f_c / inter.f0) * sqrt(mirror->R);
    }

    *ph1 = mirror->phi * ((1 + f / inter.f0) + (1 + f_c / inter.f0));
}

void fill_mirror_signal_rhs(signal_t *signal, ifo_matrix_vars_t* M_car, ifo_matrix_vars_t* M_sig){
    assert(M_sig != NULL);
    assert(M_car != NULL);
    assert(signal != NULL);
    
    int f,i, tn,tm;
    mirror_t *mirror = &inter.mirror_list[signal->component_index];
                
    node_t *node1 = &inter.node_list[mirror->node1_index];
    node_t *node2 = &inter.node_list[mirror->node2_index];
    
    for(i=0; i<mirror->num_motions; i++){
        int xidx = mirror->x_rhs_idx + i;
        
        switch(mirror->motion_type[i]) {
            case Z:
                if(signal->type == SIG_PHS) {
                    assert(mirror->long_tf != NULL);
                    double k = TWOPI/init.lambda;
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/(init.x_scale*k), signal->phase);
                } else if(signal->type == SIG_Z) {
                    assert(mirror->long_tf != NULL);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/init.x_scale, signal->phase);
                } else if(signal->type == SIG_FZ) {
                    assert(mirror->long_tf != NULL);
                    // compute transfer function for positive signal frequency
                    complex_t G_Omega = z_by_x(calc_transfer_func(mirror->long_tf, co(0, TWOPI *inter.fsig)), 1.0/mirror->mass);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(G_Omega, signal->amplitude/init.x_scale, signal->phase);
                }
                break;
                
            case ROTX:
                if(signal->type == SIG_X){
                    assert(mirror->rot_x_tf != NULL);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/init.x_scale, signal->phase);
                } else if(signal->type == SIG_FRX) {
                    assert(mirror->rot_x_tf != NULL);
                    // compute transfer function for positive signal frequency
                    complex_t G_Omega = z_by_x(calc_transfer_func(mirror->rot_x_tf, co(0, TWOPI *inter.fsig)), 1.0/mirror->Ix);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(G_Omega, signal->amplitude/init.x_scale, signal->phase);
                }
                break;
                
            case ROTY:
                if(signal->type == SIG_Y) {
                    assert(mirror->rot_y_tf != NULL);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/init.x_scale, signal->phase); 
                } else if(signal->type == SIG_FRY) {
                    assert(mirror->rot_y_tf != NULL);
                    // compute transfer function for positive signal frequency
                    complex_t G_Omega = z_by_x(calc_transfer_func(mirror->rot_y_tf, co(0, TWOPI *inter.fsig)), 1.0/mirror->Iy);
                    ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(G_Omega, signal->amplitude/init.x_scale, signal->phase);
                }
                break;
                
            case SURFACE:
                
                warn("%i %i\n", i, mirror->num_motions);
                if(i == mirror->num_motions + signal->surface_mode_index){
                    if(signal->type == SIG_SURF) {
                        assert(mirror->surface_motions_tf != NULL);
                        assert(mirror->surface_motions_tf[signal->surface_mode_index] != NULL);
                        ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(complex_1, signal->amplitude/init.x_scale, signal->phase);
                    } else if(signal->type == SIG_FSURF) {
                        assert(mirror->surface_motions_tf != NULL);
                        assert(mirror->surface_motions_tf[signal->surface_mode_index] != NULL);
                        // compute transfer function for positive signal frequency
                        complex_t G_Omega = calc_transfer_func(mirror->surface_motions_tf[signal->surface_mode_index], co(0, TWOPI *inter.fsig));
                        ((complex_t*)M_sig->rhs_values)[xidx] = z_by_xph(G_Omega, signal->amplitude/init.x_scale, signal->phase);
                    }
                }
                break;
            default:
                bug_error("not handled");
        }
    }
    
    if((signal->type == SIG_AMP || signal->type == SIG_PHS || signal->type == SIG_Z) && mirror->mass == 0){
        for(f=0; f<M_sig->num_frequencies; f++){
            frequency_t *f_sig = M_sig->frequencies[f];
            double phase = (90.0 + f_sig->order * signal->phase);
            double factor = signal->amplitude * EPSILON * 0.5; //Epsilon/2 for fsig
            double turn, ph1;

            calc_mirror_signal_vars(signal, mirror, &factor, &phase, &ph1, f_sig->f, f_sig->carrier->f);

            if(!node1->gnd_node){

                // get the index for the signal field of frequency f and mode i and 
                // the carrier field which will be its source
                int sidx = M_sig->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(2, f, 0, M_sig->num_frequencies);
                // get the incoming field at the mirror as we then have to compute the purely
                // reflected part of the field by multiplying with the mirror knm matrix. We can't use
                // the reflected carrier field as this will contain the non-reflected tranmitted field too
                int cidx = M_car->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, 0, M_car->num_frequencies);

                // idea is we compute: signal fields = factor*exp(i*phase)) * (knm * carrier fields)
                complex_t *sig_vec = &(((complex_t*)M_sig->rhs_values)[sidx]);
                complex_t *car_vec = &(((complex_t*)M_car->rhs_values)[cidx]);

                // multiply carrier vector by knm matrix and store result in signal vector
                cblas_zgemv(CblasRowMajor, CblasNoTrans, inter.num_fields, inter.num_fields,
                        &complex_1, &(mirror->knm.k11[0][0]), inter.num_fields, car_vec, 1, &complex_0, sig_vec, 1);

                // now apply all the phase factors etc.
                for(i=0; i<inter.num_fields;i++){
                    get_tem_modes_from_field_index(&tn, &tm, i);
                    turn = turnit(tn);
                    sidx =  M_sig->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(2, f, i, M_sig->num_frequencies);

                    ((complex_t*)M_sig->rhs_values)[sidx] = z_by_xph(((complex_t*)M_sig->rhs_values)[sidx], turn * factor, phase + ph1);

                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) ((complex_t*)M_sig->rhs_values)[sidx].im *= -1;
                }
            }

            if(!node2->gnd_node){

                // get the index for the signal field of frequency f and mode i and 
                // the carrier field which will be its source
                int sidx = M_sig->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(1, f, 0, M_sig->num_frequencies);
                // get the incoming field at the mirror as we then have to compute the purely
                // reflected part of the field by multiplying with the mirror knm matrix. We can't use
                // the reflected carrier field as this will contain the non-reflected tranmitted field too
                int cidx = M_car->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(2, f_sig->carrier->index, 0, M_car->num_frequencies);

                complex_t *sig_vec = &(((complex_t*)M_sig->rhs_values)[sidx]);
                complex_t *car_vec = &(((complex_t*)M_car->rhs_values)[cidx]);

                cblas_zgemv (CblasRowMajor, CblasNoTrans, inter.num_fields, inter.num_fields,
                        &complex_1, &(mirror->knm.k22[0][0]), inter.num_fields, car_vec, 1, &complex_0, sig_vec, 1);

                // now apply all the phase factors etc.
                for(i=0; i<inter.num_fields;i++){
                    get_tem_modes_from_field_index(&tn, &tm, i);
                    turn = turnit(tn);
                    sidx =  M_sig->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(1, f, i, M_sig->num_frequencies);

                    ((complex_t*)M_sig->rhs_values)[sidx] = z_by_xph(((complex_t*)M_sig->rhs_values)[sidx], turn * factor, phase - ph1 + 180 * f_sig->order);

                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) ((complex_t*)M_sig->rhs_values)[sidx].im *= -1;
                }
            }
        }

    } else if((signal->type == SIG_X && mirror->Ix == 0) || (signal->type == SIG_Y && mirror->Iy == 0)) { 

        double ph2;
        // when the fsig is vertical the apparent sign at the back surface is reversed
        if (signal->type == SIG_Y)
            ph2 = 180.0;
        else
            ph2 = 0.0;

        for(f=0; f<M_sig->num_frequencies; f++){
            frequency_t *f_sig = M_sig->frequencies[f];
            double phase = f_sig->order * signal->phase; // amplitude modulated for alignment signals
            double factor = signal->amplitude * EPSILON * 0.5; //Epsilon/2 for fsig

            complex_t q1, q2;
            complex_t txs1 = complex_0, txs2 = complex_0;

            if(!node1->gnd_node){
                if (signal->type == SIG_X) {
                    q1 = node1->qx;
                } else {
                    q1 = node1->qy;
                }

                if (node1->component_index == signal->component_index) {
                    q1 = cminus(cconj(q1));
                }

                // compute q/w_0
                txs1 = xyamp(node1, q1);
            }

            if(!node2->gnd_node){
                if (signal->type == SIG_X) {
                    q2 = node2->qx;
                } else {
                    q2 = node2->qy;
                }

                // new testing: reversed
                if (node2->component_index == signal->component_index) {
                    q2 = cminus(cconj(q2));
                }

                // compute q/w_0
                txs2 = xyamp(node2, q2);
            }

            // the old Gouy phase problem again: I should now do a proper
            // rev_gouy with txs1 and txs2, but I think it will just 
            // produce a constant phase of 90 deg. So this is what I do instead.

            txs1 = co(0, zmod(txs1));
            txs2 = co(0, zmod(txs2));

            // testing :  ################################################################################

            int cidx, sidx, car_field;
            complex_t *crhs = (complex_t*)M_car->rhs_values;
            complex_t *srhs = (complex_t*)M_sig->rhs_values;

            int i,j,field_index;

            // compute purely reflected carrier field, do this by taking the incoming beam and computing matrix multiplication
            // with the reflection elements of the matrix. Store reflected field in memory already allocated for this in the
            // signal structure
            for(i=0; i < inter.num_fields; i++){
                signal->ao1[i] = complex_0;
                signal->ao2[i] = complex_0;

                for(j=0; j < inter.num_fields; j++){
                    if(!node1->gnd_node) {
                        cidx = M_car->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(1, f_sig->carrier->index, j, M_car->num_frequencies);
                        signal->ao1[i] = z_pl_z(signal->ao1[i], z_by_z(*mirror->a11f[f_sig->carrier->index][j][i], crhs[cidx]));
                    }

                    if(!node2->gnd_node) {
                        cidx = M_car->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(2, f_sig->carrier->index, j, M_car->num_frequencies);
                        signal->ao2[i] = z_pl_z(signal->ao2[i], z_by_z(*mirror->a22f[f_sig->carrier->index][j][i], crhs[cidx]));
                    }
                }
            }

            for (field_index = 0; field_index < inter.num_fields; field_index++) {
                complex_t a11 = complex_0;
                complex_t a22 = complex_0;
                int tn, tm;
                double turn;

                get_tem_modes_from_field_index(&tn, &tm, field_index);
                //turn = turnit(tn);
                // ddb - I don't think we are using turn it here initially as we take the
                // outgoing field, which is already turned.
                turn = 1.0;
                //TODO: test and the finalize the above !

                // the aim below is to select the carrier field index and from the carrier
                // matrix extract the value to use in the signal field computation
                if (tn && signal->type == SIG_X) {
                    car_field = get_field_index_from_tem(tn - 1, tm);

                    a11 = z_by_x(z_by_z(signal->ao1[car_field], cconj(txs1)), 1.0 * sqrt(tn));
                    a22 = z_by_x(z_by_z(signal->ao2[car_field], cconj(txs2)), 1.0 * sqrt(tn));
                } else if (tm && signal->type == SIG_Y) {
                    car_field = get_field_index_from_tem(tn, tm - 1);

                    a11 = z_by_x(z_by_z(signal->ao1[car_field], cconj(txs1)), 1.0 * sqrt(tm));
                    a22 = z_by_x(z_by_z(signal->ao2[car_field], cconj(txs2)), 1.0 * sqrt(tm));
                }

                if (tn + tm < inter.tem) {
                    if (signal->type == SIG_X) {
                        car_field = get_field_index_from_tem(tn + 1, tm);

                        a11 = z_pl_z(a11, z_by_x(z_by_z(signal->ao1[car_field], txs1), -1.0 * sqrt(tn + 1)));
                        a22 = z_pl_z(a22, z_by_x(z_by_z(signal->ao2[car_field], txs2), -1.0 * sqrt(tn + 1)));
                    } else {
                        car_field = get_field_index_from_tem(tn, tm + 1);
                        a11 = z_pl_z(a11, z_by_x( z_by_z(signal->ao1[car_field], txs1), -1.0 * sqrt(tm + 1)));
                        a22 = z_pl_z(a22, z_by_x( z_by_z(signal->ao2[car_field], txs2), -1.0 * sqrt(tm + 1)));
                    }
                }

                if (!node1->gnd_node) {
                    sidx = M_sig->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(2, f, field_index, M_sig->num_frequencies);
                    srhs[sidx] = z_by_xph(a11, turn * factor, phase);
                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) srhs[sidx].im *= -1;
                }

                if (!node2->gnd_node) {
                    sidx = M_sig->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(1, f, field_index, M_sig->num_frequencies);
                    srhs[sidx] = z_by_xph(a22, turn * factor, phase + ph2);
                    // check if we need to conjugate the value for the lower sideband
                    if(f_sig->order == -1) srhs[sidx].im *= -1;
                }
            }
        }
    }    
}

/** Set the RHS vector signal elements for a mirror
 * \param signal the signal object to set the elements for
 * \param factor an amplification factor
 * \param phase base phase of the signal ???
 * \param f actual frequency (carrier + signal)
 * \param f1 carrier frequency
 * \param lower2 index indicating which sideband is being processed
 *
 * \todo lower2 -> sideband_index
 *
 * \see Test_set_mirror_signal_rhs()
 */ 
void set_mirror_signal_rhs(
        signal_t *signal,
        double factor,
        double phase,
        double f,
        double f_c,
        int lower2) {
    
    node_t signal_node1 = inter.node_list[*(signal->node1_index)];
    node_t signal_node2 = inter.node_list[*(signal->node2_index)];
    
    mirror_t *mirror = &inter.mirror_list[signal->component_index];
    
    if(options.sparse_solver != KLU)
        bug_error("sparse solver not compatiable with RHS vector filling");
    
    if (signal->type == SIG_AMP || signal->type == SIG_PHS) {
     
        int tn, tm, field_index;
        double turn, ph1;
        
        calc_mirror_signal_vars(signal, mirror, &factor, &phase, &ph1, f, f_c);
        
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            get_tem_modes_from_field_index(&tn, &tm, field_index);
            turn = turnit(tn);

            int rhs_idx;
            
            // set rhs vector element for node1
            if (NOT signal_node1.gnd_node) {
                rhs_idx = 2 * inter.num_fields * signal->node1_reduced_index + signal->node1_amp_dir->fro + 2 * field_index;
                set_rhs(rhs_idx, z_by_ph(z_by_x(signal->a1[field_index], turn * factor), phase + ph1), STANDARD);
            }

            // set rhs vector element for node2
            if (NOT signal_node2.gnd_node) {
                rhs_idx = 2 * inter.num_fields * signal->node2_reduced_index + signal->node2_amp_dir->fro + 2 * field_index;
                set_rhs(rhs_idx, z_by_ph(z_by_x(signal->a2[field_index], turn * factor), phase - ph1 + 180 * lower2), STANDARD); // +ph1 to -ph1 060302
            }
        }
    }// alignment transfer function 
    else if (signal->type == SIG_Y || signal->type == SIG_X) {

        double ph1 = 0.0; // tuning not needed as this is computed in the computation of sig->ao

        // switch back to amplitude modulation
        phase = phase - 90.0;

        complex_t q1, q2;
        complex_t txs1 = complex_0, txs2 = complex_0;
        
        if(!signal_node1.gnd_node){
            if (signal->type == SIG_X) {
                q1 = signal_node1.qx;
            } else {
                q1 = signal_node1.qy;
            }

            if (signal_node1.component_index == signal->component_index) {
                q1 = cminus(cconj(q1));
            }
            
            // compute q/w_0
            txs1 = xyamp(&signal_node1, q1);
        }
        
        if(!signal_node2.gnd_node){
            if (signal->type == SIG_X) {
                q2 = signal_node2.qx;
            } else {
                q2 = signal_node2.qy;
            }
            // maybe on the back this should be reversed as the rest of the equation assumes the same
            // amplitude. this would yield a minus in xyamp
            /*
            if (signal_node2.component_index == signal->component_index) {
              q2 = cminus(cconj(q2));
            }
             */
            // new testing: reversed
            if (signal_node2.component_index == signal->component_index) {
                q2 = cminus(cconj(q2));
            }

            // compute q/w_0
            txs2 = xyamp(&signal_node2, q2);
        }
        
        // the old Gouy phase problem again: I should now do a proper
        // rev_gouy with txs1 and txs2, but I think it will just 
        // produce a constant phase of 90 deg. So this is what I do instead.

        txs1 = co(0, zmod(txs1));
        txs2 = co(0, zmod(txs2));
        
        // testing :  ################################################################################

        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            complex_t a11 = complex_0;
            complex_t a22 = complex_0;
            int tn, tm;
            double turn;

            get_tem_modes_from_field_index(&tn, &tm, field_index);
            //turn = turnit(tn);
            turn = 1.0;
            //TODO: test and the finalize the above !

            if (tn && signal->type == SIG_X) {
                a11 = z_by_x(z_by_z(signal->ao1[get_field_index_from_tem(tn - 1, tm)], cconj(txs1)), 1.0 * sqrt(tn));
                a22 = z_by_x(z_by_z(signal->ao2[get_field_index_from_tem(tn - 1, tm)], cconj(txs2)), 1.0 * sqrt(tn));
            } else if (tm && signal->type == SIG_Y) {
                a11 = z_by_x(z_by_z(signal->ao1[get_field_index_from_tem(tn, tm - 1)], cconj(txs1)), 1.0 * sqrt(tm));
                a22 = z_by_x(z_by_z(signal->ao2[get_field_index_from_tem(tn, tm - 1)], cconj(txs2)), 1.0 * sqrt(tm));
            }

            if (tn + tm < inter.tem) {
                if (signal->type == SIG_X) {
                    a11 = z_pl_z(a11, z_by_x(z_by_z(signal->ao1[get_field_index_from_tem(tn + 1, tm)], txs1), -1.0 * sqrt(tn + 1)));
                    a22 = z_pl_z(a22, z_by_x(z_by_z(signal->ao2[get_field_index_from_tem(tn + 1, tm)], txs2), -1.0 * sqrt(tn + 1)));
                } else {
                    a11 = z_pl_z(a11, z_by_x( z_by_z(signal->ao1[get_field_index_from_tem(tn, tm + 1)], txs1), -1.0 * sqrt(tm + 1)));
                    a22 = z_pl_z(a22, z_by_x( z_by_z(signal->ao2[get_field_index_from_tem(tn, tm + 1)], txs2), -1.0 * sqrt(tm + 1)));
                }
            }

            if (NOT signal_node1.gnd_node) {
                // set rhs vector element for node1, removed 
                set_rhs(2 * inter.num_fields * signal->node1_reduced_index +
                        signal->node1_amp_dir->fro + 2 * field_index,
                        z_by_ph(z_by_x(a11, turn * factor), phase + ph1),
                        STANDARD);
            }

            if (NOT signal_node2.gnd_node) {
                // when the fsig is vertical the apparent sign at the back surface is reversed
                double ph2;
                if (signal->type == SIG_Y) {
                    ph2 = 180.0;
                } else {
                    ph2 = 0.0;
                }

                set_rhs(2 * inter.num_fields * signal->node2_reduced_index +
                        signal->node2_amp_dir->fro + 2 * field_index,
                        z_by_ph(z_by_x(a22, turn * factor), phase - ph1 + ph2),
                        STANDARD);
            }
        }
    } else {
        bug_error("wrong signal type for mirror");
    }
}

void calc_space_signal_vars(space_t *space, double *factor, double *ph1, double f_c, int order){
    *factor *= sin(space->n * space->L * 2 * PI * inter.fsig / (2*init.clight)) * (f_c + inter.f0) / inter.fsig;
    *ph1 = space->n * space->L / init.clight * (-1.0 * f_c - order * inter.fsig * 0.5) * 360.0;
}

void fill_space_signal_rhs(signal_t *signal, ifo_matrix_vars_t* M_car, ifo_matrix_vars_t* M_sig) {
    
    node_t signal_node1 = inter.node_list[*(signal->node1_index)];
    node_t signal_node2 = inter.node_list[*(signal->node2_index)];
    int f;
    space_t *space = &inter.space_list[signal->component_index];
    
    if (signal->type == SIG_PHS) {
        double ph1;
        
        for(f=0; f < M_sig->num_frequencies; f++){
            frequency_t *f_sig = M_sig->frequencies[f];
            
            double phase = (90.0 + f_sig->order * signal->phase);
            double factor = signal->amplitude * EPSILON * 0.5;
            
            calc_space_signal_vars(space, &factor, &ph1, f_sig->carrier->f, f_sig->order);
            
            int i, sidx=0, cidx=0, iport1, oport1, iport2, oport2;
            iport1 = 0;//silence compiler warning
            iport2 = 0;
            oport1 = 0;
            oport2 = 0;
                
            // select the outgoing field and use the carrier there to generate the sidebands
            // !! NOTE !!
            // The space IN and OUT ports are actually opposite, as IN and OUT refer to 
            // what is going in and out of a COMPONENT not a space.
            if(signal_node1.direction == IN_OUT){
                iport1 = 2;
                oport1 = 1;
            }else if(signal_node1.direction == OUT_IN){
                iport1 = 1;
                oport1 = 2;
            }else
                bug_error("not expected");
            
            if(signal_node2.direction == IN_OUT){
                iport2 = 2;
                oport2 = 1;
            }else if(signal_node2.direction == OUT_IN){
                iport2 = 1;
                oport2 = 2;
            }else
                bug_error("not expected");

            complex_t *sig_vec = NULL, *car_vec = NULL;
            
            if (!signal_node1.gnd_node && !signal_node2.gnd_node) {
                sidx = M_sig->node_rhs_idx_1[signal_node2.list_index] + get_node_rhs_idx(oport2, f, 0, M_sig->num_frequencies);
                cidx = M_car->node_rhs_idx_1[signal_node1.list_index] + get_node_rhs_idx(iport1, f_sig->carrier->index, 0, M_car->num_frequencies);

                sig_vec = &(((complex_t*)M_sig->rhs_values)[sidx]);
                car_vec = &(((complex_t*)M_car->rhs_values)[cidx]);
                
                cblas_zgemv (CblasRowMajor, CblasNoTrans, inter.num_fields, inter.num_fields,
                    &complex_1, &(space->k12[0][0]), inter.num_fields, car_vec, 1, &complex_0, sig_vec, 1);
                
                for (i = 0; i < inter.num_fields; i++) {
                    sidx = M_sig->node_rhs_idx_1[signal_node2.list_index] + get_node_rhs_idx(oport2, f, i, M_sig->num_frequencies);
                    ((complex_t*)M_sig->rhs_values)[sidx] = z_by_xph(((complex_t*)M_sig->rhs_values)[sidx], factor, phase+ph1);
                    if(f_sig->order == -1) ((complex_t*)M_sig->rhs_values)[sidx].im *= -1;
                }
                
                sidx = M_sig->node_rhs_idx_1[signal_node1.list_index] + get_node_rhs_idx(oport1, f, 0, M_sig->num_frequencies);
                cidx = M_car->node_rhs_idx_1[signal_node2.list_index] + get_node_rhs_idx(iport2, f_sig->carrier->index, 0, M_car->num_frequencies);
                
                sig_vec = &(((complex_t*)M_sig->rhs_values)[sidx]);
                car_vec = &(((complex_t*)M_car->rhs_values)[cidx]);

                cblas_zgemv (CblasRowMajor, CblasNoTrans, inter.num_fields, inter.num_fields,
                    &complex_1, &(space->k21[0][0]), inter.num_fields, car_vec, 1, &complex_0, sig_vec, 1);
                
                for (i = 0; i < inter.num_fields; i++) {
                    sidx = M_sig->node_rhs_idx_1[signal_node1.list_index] + get_node_rhs_idx(oport1, f, i, M_sig->num_frequencies);
                    
                    ((complex_t*)M_sig->rhs_values)[sidx] = z_by_xph(((complex_t*)M_sig->rhs_values)[sidx], factor, phase +ph1);
                    if(f_sig->order == -1) ((complex_t*)M_sig->rhs_values)[sidx].im *= -1;
                }
            }
        }
    } else 
        bug_error("wrong signal type for space");   
}

/*! Set the RHS vector signal elements for a space
 * \param signal the signal object to set the elements for
 * \param factor an amplification factor
 * \param phase base phase of the signal ???
 * \param f1 another base frequency ???
 * \param lower2 index indicating which sideband is being processed
 *
 * \todo lower2 -> sideband_index
 *
 * \see Test_set_space_signal_rhs()
 */
void set_space_signal_rhs(
        signal_t *signal,
        double factor,
        double phase,
        double f1,
        int lower2) {
    node_t signal_node1 = inter.node_list[*(signal->node1_index)];
    node_t signal_node2 = inter.node_list[*(signal->node2_index)];

    if (signal->type == SIG_AMP || signal->type == SIG_PHS) {
        // amplitude modulation (test)
        if (signal->type == SIG_AMP) {
            gerror("this modulation type is not yet implemented\n");
        }

        double ph1;
        
        calc_space_signal_vars(&inter.space_list[signal->component_index], &factor, &ph1, f1, lower2);
        
        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            if (NOT signal_node1.gnd_node) {
                set_rhs(2 * inter.num_fields * signal->node1_reduced_index +
                        signal->node1_amp_dir->fro + 2 * field_index,
                        z_by_ph(z_by_x(signal->a1[field_index], factor), phase + ph1),
                        STANDARD);
            }

            if (NOT signal_node2.gnd_node) {
                set_rhs(2 * inter.num_fields * signal->node2_reduced_index +
                        signal->node2_amp_dir->fro + 2 * field_index,
                        z_by_ph(z_by_x(signal->a2[field_index], factor), phase + ph1),
                        STANDARD);
            }
        }
    } else {
        bug_error("wrong signal type for space");
    }

    //! \todo return from here one day!!!
}

//! Set the RHS vector signal elements for a laser

/*!
 * \param signal the signal object to set the elements for
 * \param factor an amplification factor
 * \param phase base phase of the signal ???
 * \param carrier type of input field
 *
 * \see Test_set_light_input_signal_rhs()
 */
void set_light_input_signal_rhs(
        signal_t *signal,
        double factor,
        double phase,
        int carrier) {
    node_t signal_node1 = inter.node_list[*(signal->node1_index)];

    if (carrier == CARRIER_FIELD) { // Input modulation can only be around carrier 
        if (signal->type == SIG_AMP ||
                signal->type == SIG_PHS ||
                signal->type == SIG_FRQ) {

            // amplitude/phase mod
            if (signal->type == SIG_AMP ||
                    signal->type == SIG_PHS) {
                bug_error
                        ("Input: amp/phs: this modulation type is not yet implemented");
            }

            if (inter.fsig == 0) {
                factor = 1e20;
                warn("fsig=0 for frequency modulation is not useful!\n");
            } else {
                factor = factor / (2.0 * PI * inter.fsig);
            }

            int field_index;
            for (field_index = 0; field_index < inter.num_fields; field_index++) {
                if (NOT signal_node1.gnd_node) {
                    set_rhs(2 * inter.num_fields * signal->node1_reduced_index + 1 +
                            2 * field_index,
                            z_by_ph(z_by_x(signal->a1[field_index], factor), phase),
                            STANDARD);
                }
            }
        } else {
            bug_error("wrong signal type for input");
        }
    }
    // set to zero??

    //! \todo return from here one day!!!
}

void fill_modulator_signal_rhs(signal_t *signal, ifo_matrix_vars_t* M_car, ifo_matrix_vars_t* M_sig){
    
    node_t *node1 = &inter.node_list[*(signal->node1_index)];
    node_t *node2 = &inter.node_list[*(signal->node2_index)];
    
    // if either are a dump then there are no signal sidebands being generated around
    // any carriers. Dump node could be an input for vacuum quantum noise but signal sidebands
    // of noise would be negligible I would assume...
    if(node1->gnd_node || node2->gnd_node)
        return;
    
    int fin, fout, fs, i, cidx, sidx;
    
    complex_t *crhs = (complex_t*) M_car->rhs_values;
    complex_t *srhs = (complex_t*) M_sig->rhs_values;
    
    if (signal->type == SIG_PHS || signal->type == SIG_FRQ) {
        /* So the idea here is that any carrier being coupled by this modulator
         * will then have phase/frequency sidebands around it. So we first determine
         * which carriers are coupling then add the signal sideband there
         */
        for(fin=0; fin<M_car->num_frequencies; fin++){
            for(fout=0; fout<M_car->num_frequencies; fout++){
                // modulator noise won't affect direct coupling from same frequency
                if(M_car->mod_does_f_couple[signal->component_index][fin][fout] && fin!=fout){
                    frequency_t *fcar = M_car->frequencies[fout];
                    // Now we know this carrier has been created by this modulator
                    // therefore some noise will be present on it. Now we just need to find
                    // which signal frequencies belong to it
                    for(fs=0; fs<M_sig->num_frequencies; fs++){
                        frequency_t *fsig = M_sig->frequencies[fs];
                        
                        if(fsig->carrier == fcar){
                            // This should be equation 3.27 in the manual
                            
                            int car_mod_order = M_car->mod_f_couple_order[signal->component_index][fin][fout]; // The modulation order of this signal sidebands carrier
                            
                            double phase = 90.0 + fsig->order * signal->phase;
                            double factor = car_mod_order * signal->amplitude * EPSILON * 0.5;

                            for(i=0; i<inter.num_fields; i++){
                                // compute output at node 1 of a modulator, take outgoing field here and make sidebands around it
                                cidx = M_car->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(2, fout, i, M_car->num_frequencies);
                                sidx = M_sig->node_rhs_idx_1[node1->list_index] + get_node_rhs_idx(2, fs, i, M_sig->num_frequencies);
                                srhs[sidx] = z_pl_z(srhs[sidx], z_by_xph(crhs[cidx], factor, phase));

                                // now node2...
                                cidx = M_car->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(1, fout, i, M_car->num_frequencies);
                                sidx = M_sig->node_rhs_idx_1[node2->list_index] + get_node_rhs_idx(1, fs, i, M_sig->num_frequencies);
                                srhs[sidx] = z_pl_z(srhs[sidx], z_by_xph(crhs[cidx], factor, phase));
                            }
                        }
                    }
                    
                }
            }
        }
    } else if (signal->type == SIG_AMP) {
        bug_error("mod amp: this modulation type is not implemented yet");
    } else {
        bug_error("wrong signal type for mod");
    }
}

/** Set the RHS vector signal elements for a modulator
 * \param signal the signal object to set the elements for
 * \param factor an amplification factor
 * \param phase base phase of the signal ???
 * \param lower1 index indicating which sideband is being processed
 * \param carrier type of input field
 * \param modulator_index index describing which modulator is being used
 * \param modulation_order order of the modulation
 * \param z input complex amplitude
 *
 * \todo lower1 -> sideband_index
 *
 * \see Test_set_modulator_signal_rhs()
 */
void set_modulator_signal_rhs(
        signal_t *signal,
        double factor,
        double phase,
        int lower1,
        int carrier,
        int modulator_index,
        int modulation_order,
        complex_t z) {
    node_t signal_node1 = inter.node_list[*(signal->node1_index)];
    node_t signal_node2 = inter.node_list[*(signal->node2_index)];

    if (signal->type == SIG_PHS || signal->type == SIG_FRQ) {
        if (carrier != CARRIER_FIELD) { // No Phasenoise sidebands on Carrier 
            assert(lower1 == 1 || lower1 == -1);

            if (signal->component_index == modulator_index) { //11.10.2000
                modulator_t *modulator;
                modulator = &(inter.modulator_list[modulator_index]);
                factor *= lower1 * (modulation_order + 1); // 181004 adf after MM
                //message("k=%d , factor: %g phase %g\n",
                //        lower1 * (modulation_order + 1), factor, phase);

                int field_index;
                for (field_index = 0; field_index < inter.num_fields; field_index++) {
                    if (NOT signal_node1.gnd_node) {
                        set_rhs(2 * inter.num_fields * signal->node1_reduced_index +
                                signal->node1_amp_dir->fro + 2 * field_index,
                                z_by_ph(z_by_x(z_by_z(modulator->a1[field_index], z),
                                factor), phase),
                                STANDARD);
                    }

                    if (NOT signal_node2.gnd_node) {
                        set_rhs(2 * inter.num_fields * signal->node2_reduced_index +
                                signal->node2_amp_dir->fro + 2 * field_index,
                                z_by_ph(z_by_x(z_by_z(modulator->a2[field_index], z),
                                factor), phase),
                                STANDARD);
                    }
                }
            }
        }
    } else if (signal->type == SIG_AMP) {
        bug_error("mod amp: this modulation type is not implemented yet");
    } else {
        bug_error("wrong signal type for mod");
    }

    //! \todo return from here one day!!!
}

void set_light_input_ports(){
    
    int i;
    for(i=0; i<inter.num_light_inputs; i++){
        light_in_t *light_input;
        light_input = &(inter.light_in_list[i]);
        
        node_t* n = &inter.node_list[light_input->node_index];
        node_connections_t *nc = &inter.node_conn_list[light_input->node_index];
        
        // All laser light inputs should be at a loose node of a space or a component
        // that isn't connected to anything else.

        int is_loose = n->connect && (nc->comp_1 == NULL)^(nc->comp_2 == NULL);

        if(!is_loose)
            bug_error("Light input %s should be connected to a loose node", light_input->name);

        if((nc->comp_1 != NULL && nc->type_1 == SPACE) || (nc->comp_2 != NULL && nc->type_2 == SPACE)){
            // if the light input is connected to a space...
            if(n->direction == IN_OUT) { // first port in the input
                light_input->node_port = 2;
            } else if(n->direction == OUT_IN) { // second port is the input
                light_input->node_port = 1;
            } else
                bug_error("Node did not have direction set when filling in RHS vector");
        } else if((nc->comp_1 != NULL && nc->type_1 != SPACE) || (nc->comp_2 != NULL && nc->type_2 != SPACE)){
            // if the light input is connected to something else...
            if(n->direction == IN_OUT) { // first port in the input
                light_input->node_port = 1;
            } else if(n->direction == OUT_IN) { // second port is the input
                light_input->node_port = 2;
            }else
                bug_error("Node did not have direction set when filling in RHS vector");
        }
    }
}

int x_count = 0;

/**
 * Computes a data point for all light amplitudes for one point on the x-axis
 * This differs from the old data_point routine which handles frequencies 
 * sequentially instead of solving all frequencies at the same time.
 */
void data_point_new(){
    // ensure that the fsig variables are always sensible
    assert((inter.fsig==0 && inter.mfsig == 0) || (inter.fsig == -inter.mfsig));
    
    x_count++;
    
    if(options.sparse_solver != KLU_FULL)
        bug_error("data_point_new is not compatiable with the solver used");
    
    // check global flag to see any of the frequencies are getting tuned
    if(carrier_frequencies_tuned || signal_frequencies_tuned){
        // refresh the used frequencies
        update_frequency_list();

        // update which frequencies now couple to which
        update_frequency_coupling(0);
        
        int f;
        frequency_t *f1 = NULL;

        // update result frequency list
        for(f=0; f<inter.num_carrier_freqs; f++){
            f1 = inter.carrier_f_list[f];
            f_s[f1->result_index] = f1->f;
        }
        
        for(f=0; f<inter.num_signal_freqs; f++){
            f1 = inter.signal_f_list[f];
            f_s[f1->result_index] = f1->f;
        }
    }
    
    // re fill the matrix with the latest values for the data point
    fill_ccs_matrix(&M_ifo_car);
    factor_matrix(STANDARD);
    clear_rhs(STANDARD);
    
    int using_sig_qcorr = inter.num_signal_freqs > 0;
    
    if(using_sig_qcorr){
        // clear the rhs vector as we will add to it in the next loop for the laser frequency modulation...
        clear_rhs(SIGNAL_QCORR);
    }
    
    // Now we must insert all the various sources to the RHS vector
    // for the carrier matrix
    int i,j;
    for(i=0; i<inter.num_light_inputs; i++){
        light_in_t *light_input;
        light_input = &(inter.light_in_list[i]);
        
        node_t* n = &inter.node_list[light_input->node_index];

        int rhs_idx_car = 0;
        int rhs_idx_sig = 0;
        
        if(light_input->node_port == 2) { // first port in the input
            rhs_idx_car = M_ifo_car.node_rhs_idx_2[n->list_index];
            rhs_idx_sig = (using_sig_qcorr) ? M_ifo_sig.node_rhs_idx_2[n->list_index] : -1;
        } else if(light_input->node_port == 1) { // second port is the input
            rhs_idx_car = M_ifo_car.node_rhs_idx_1[n->list_index];
            rhs_idx_sig = (using_sig_qcorr) ? M_ifo_sig.node_rhs_idx_1[n->list_index] : -1;
        } else
            bug_error("Node did not have direction set when filling in RHS vector");
        
        double sig_factor = 0.0;
        double sig_mod_phase = 0; // Phase of modulation
        
        signal_t *signal = light_input->modulation_signal;
        
        if(signal){
            assert(inter.num_signal_freqs > 0);
            
            sig_factor = signal->amplitude * EPSILON * 0.5;
            
            switch(signal->type){
                case SIG_AMP:
                    sig_mod_phase = 0; // 0 for amplitude modulation
                    sig_factor *= 0.5; // energy conserving amplitude modulation uses
                                   // m/4 rather than m/2
                    break;
                    
                case SIG_PHS:
                    sig_mod_phase = 90; // for phase/frequency modulation;
                    
                    break;
                case SIG_FRQ:
                    sig_mod_phase = 90; // for phase/frequency modulation
                    
                    if (inter.fsig == 0) {
                        sig_factor = 1e20;
                        warn("fsig=0 for frequency modulation is not useful!\n");
                    } else {
                        sig_factor /= (2.0 * PI * inter.fsig);
                    }
                    
                    break;
                default:
                    bug_error("wrong signal type for input");
            }
        }
        
        for (j = 0; j < inter.num_fields; j++) {
            complex_t input_amplitude = z_by_ph(co(sqrt(light_input->I0 / init.epsilon_0_c), 0.0), light_input->phase);
            input_amplitude = z_by_z(input_amplitude, light_input->power_coeff_list[j]);
            
            set_rhs(rhs_idx_car + get_node_rhs_idx(1, light_input->f->index, j, M_ifo_car.num_frequencies), input_amplitude, STANDARD);
            
            // if there is a signal on this laser then add in the sideband source in RHS
            if(signal){
                int i;
                // find the signal sidebands of the laser carrier
                for(i=0; i<inter.num_signal_freqs; i++){
                    frequency_t *freq_sig = inter.signal_f_list[i];
                    
                    // if this signals carrier is this inputs frequency...
                    if(freq_sig->carrier == light_input->f){
                        int sidx = rhs_idx_sig + get_node_rhs_idx(1, freq_sig->index, j, M_ifo_sig.num_frequencies);
                        
                        ((complex_t*)M_ifo_sig.rhs_values)[sidx] = z_by_xph(input_amplitude, sig_factor, sig_mod_phase + freq_sig->order * signal->phase);
                        
                        // check if we need to conjugate lower sideband
                        if(freq_sig->order == -1)
                            ((complex_t*)M_ifo_sig.rhs_values)[sidx].im *= -1;
                    }
                }
            }
        }
    }
    
    // solve the carrier matrix
    solve_matrix(STANDARD);
    store_carrier_solutions();
    
    if(using_sig_qcorr){
        // now we have computed the carrier fields everywhere we can compute the 
        // sideband sources for the RHS
        // loop over all signals
        int signal_index;
        bool nosignals = true;
        
        for (signal_index = 0; signal_index < inter.num_signals; signal_index++) {
            signal_t *signal;
            signal = &(inter.signal_list[signal_index]);

            if(signal->type != NOSIGNAL){
                nosignals = false;
                switch (signal->component_type) {
                    case MIRROR:
                        fill_mirror_signal_rhs(signal, &M_ifo_car, &M_ifo_sig);
                        break;
                    case BEAMSPLITTER:
                        fill_bs_signal_rhs(signal,&M_ifo_car, &M_ifo_sig);
                        break;
                    case SPACE:
                        fill_space_signal_rhs(signal,&M_ifo_car, &M_ifo_sig);
                        break;
                    case MODULATOR:
                        fill_modulator_signal_rhs(signal, &M_ifo_car, &M_ifo_sig);
                        break;
                    case LIGHT_INPUT:
                        // light inputs are handled above when computing the laser input RHS for the carrier
                        break;
                    default:
                        bug_error("rhs_sig");
                }
            }
        }
        
        if(!nosignals){
            // if there are some signals we need to fill the matrix and solve
            fill_ccs_matrix(&M_ifo_sig);
            factor_matrix(SIGNAL_QCORR);
            solve_matrix(SIGNAL_QCORR);
            store_signal_solutions();
            
            if(inter.printmatrix){
                char fname[100] = {0};
                sprintf(fname, "klu_full_matrix_sig_lbl.dat");
                dump_matrix_ccs_labels(&M_ifo_sig, fname);

                sprintf(fname, "klu_full_matrix_sig%i.dat", x_count);
                dump_matrix_ccs(&(M_ifo_sig.M), fname); 
            }
        }
        
        
        // finally if we are computing some quantum noise we need to set that
        // matrix up, as it differs when using modulators as they mix frequencies
        // coherently in signal matrix          
        if(inter.num_qnoise_inputs > 0 && (nosignals || inter.num_modulators > 0 || inter.num_slinks > 0)){
            if(nosignals) {
                // if there were no signals, we wouldn't have filled the signal
                // matrix yet
                fill_ccs_matrix(&M_ifo_sig);
            }
            
            fill_matrix_modulator_elements_quantum(&M_ifo_sig);
            factor_matrix(SIGNAL_QCORR);
            
            if(inter.printmatrix){
                char fname[100] = {0};
                sprintf(fname, "klu_full_matrix_qn%i.dat", x_count);
                dump_matrix_ccs(&(M_ifo_sig.M), fname); 
            }
        }
        
        // actual quantum noise computations are done on a per detector basis
        // in fill_detector. We must ensure that the signal matrix is not changed
        // from here onwards.
    }  
}

/*!
 * Calculate all light amplitudes for ONE POINT on x-axis
 *
 * \see Test_data_point()
 */
void data_point_old(void) {
    int tid = startTimer("DATA_POINT");
    
    inter.num_frequencies = 0;
    int light_input_index;
    for (light_input_index = 0;
            light_input_index < inter.num_light_inputs;
            light_input_index++) { // all carriers

        light_in_t *light_input;
        light_input = &(inter.light_in_list[light_input_index]);

        double f = light_input->f->f;
        double f1 = f;
        int tid_spa;
        
        tid_spa = startTimer("SPA_FILL");
        fill_matrix(f1, STANDARD);
        endTimer(tid_spa);
        
        tid_spa = startTimer("SPA_RHS_CLEAR");
        clear_rhs(STANDARD);
        endTimer(tid_spa);
        
        // assign the RHS vector elements
        int field_index;
        
        tid_spa = startTimer("SPA_ASSIGN_RHS");
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            complex_t carrier_field_amplitude =
                    z_by_ph(co(sqrt(light_input->I0 / init.epsilon_0_c), 0.0),
                    light_input->phase);
            carrier_field_amplitude =
                    z_by_z(carrier_field_amplitude,
                    light_input->power_coeff_list[field_index]);

            set_rhs(light_input->node_reduced_index * 2 * inter.num_fields + 1 + 2 * field_index, carrier_field_amplitude, STANDARD);
        }
        endTimer(tid_spa);
        
        tid_spa = startTimer("SPA_FACTOR");
        factor_matrix(STANDARD);
        endTimer(tid_spa);
        
        tid_spa = startTimer("SPA_SOLVE");
        solve_matrix(STANDARD);
        endTimer(tid_spa);
        
#if DEBUG
        // initialisation
        //dump_solution_vector(STANDARD, "file_append");
#endif

        // CARRIER_FIELD --------------------------------------------------------
        // store the solution for the carrier
        tid_spa = startTimer("SPA_SOLVE");
        store_solution_old(f, inter.num_frequencies, CARRIER_FIELD);
        endTimer(tid_spa);
        
        t_s[inter.num_frequencies] = CARRIER_FIELD;
        f_s[inter.num_frequencies] = f;
        inter.num_frequencies++;
        if (inter.num_frequencies > mem.num_frequencies) {
            bug_error("too many freqs (carrier)");
        }
        
        tid_spa = startTimer("FILL_MODULATORS");
        // fill modulators with carrier -----------------------------------
        fill_modulators();
        endTimer(tid_spa);

        if (inter.num_signals) {
            solve_signal_sidebands(f1, CARRIER_FIELD, 0, 0, 0, complex_0);
        }

        // MODULATION frequencies ------------------------------------------
        tid_spa = startTimer("PROCESS_MOD_SIDEBANDS");
        process_modulation_sidebands(f1);
        endTimer(tid_spa);
    }
    
    endTimer(tid);
}

//! Process the modulation sidebands

/*!
 * \param base_frequency base frequency for calculating the modulation frequency
 *
 * \see Test_process_modulation_sidebands()
 */
void process_modulation_sidebands(double base_frequency) {
    int modulator_index;
    for (modulator_index = 0;
            modulator_index < inter.num_modulators;
            modulator_index++) {

        modulator_t *modulator;
        modulator = &(inter.modulator_list[modulator_index]);

        int modulation_order_index;
        for (modulation_order_index = 0;
                modulation_order_index < modulator->order;
                modulation_order_index++) {
            int modulation_sideband_flag = 1;

            int modulation_sideband_index;
            for (modulation_sideband_index = 0;
                    modulation_sideband_index < 2;
                    modulation_sideband_index++) {

                double f = base_frequency +
                        modulation_sideband_flag * (modulation_order_index + 1) * modulator->f_mod;
                double f2 = f;
                complex_t modulation_field_amplitude =
                        get_modulation_amplitude(modulator_index, modulation_order_index + 1,
                        modulation_sideband_flag);

                fill_matrix(f, STANDARD);
                clear_rhs(STANDARD);

                int field_index;
                for (field_index = 0; field_index < inter.num_fields; field_index++) {

                    set_rhs(2 * inter.num_fields * modulator->node1_reduced_index +
                            modulator->node1_amp_dir.fro + 2 * field_index,
                            z_by_z(modulator->a1[field_index], modulation_field_amplitude),
                            STANDARD);

                    set_rhs(2 * inter.num_fields * modulator->node2_reduced_index +
                            modulator->node2_amp_dir.fro + 2 * field_index,
                            z_by_z(modulator->a2[field_index], modulation_field_amplitude),
                            STANDARD);

                    if (inter.debug & 8) {
                        message("fill mod: %s , a1=%s a2=%s\n",
                                modulator->name,
                                complex_form(z_by_z(modulator->a1[field_index],
                                modulation_field_amplitude)),
                                complex_form(z_by_z(modulator->a2[field_index],
                                modulation_field_amplitude)));
                        message("fill mod a1: amp=%s ph=%s\n",
                                xdouble_form(zmod(z_by_z(modulator->a1[field_index],
                                modulation_field_amplitude))),
                                xdouble_form(zdeg(z_by_z(modulator->a1[field_index],
                                modulation_field_amplitude))));
                        message("fill mod a2: amp=%s ph=%s\n",
                                xdouble_form(zmod(z_by_z(modulator->a2[field_index],
                                modulation_field_amplitude))),
                                xdouble_form(zdeg(z_by_z(modulator->a2[field_index],
                                modulation_field_amplitude))));
                    }
                }

                factor_matrix(STANDARD);
                solve_matrix(STANDARD);

                // store the solution for the current modulation frequency
                store_solution_old(f, inter.num_frequencies, MODULATOR_FIELD);

                t_s[inter.num_frequencies] = MODULATOR_FIELD;
                f_s[inter.num_frequencies] = f;
                inter.num_frequencies++;

                if (inter.num_frequencies > mem.num_frequencies) {
                    bug_error("too many freqs (mod1)");
                }

                if (modulator->single_sideband_mode == ON) { // skip second sideband in SSB-mode
                    modulation_sideband_index++;
                }

                if (inter.num_signals) {
                    solve_signal_sidebands(f2, MODULATOR_FIELD, modulator_index,
                            modulation_order_index, modulation_sideband_flag,
                            modulation_field_amplitude);
                }
                modulation_sideband_flag = -1; // changed 181004 adf after MM
            }
        }
    }
}

//! Solve the sparse matrix for signal sidebands

/*!
 * Fill signals around carrier or modulation fields, solve the sparse matrix
 * and store its solution
 *
 * \param base_frequency frequency of light filed for which the matrix has just been solved
 * \param field_type the kind of field CARRIER_FIELD or MODULATOR_FIELD
 * \param modulator_index index of current modulator or 0 if CARRIER_FIELD
 * \param modulation_order_index order of modulation or 0 if CARRIER_FIELD
 * \param modulation_sideband_flag flag for current modulation sideband or 0
 * for CARRIER_FIELD
 * \param field_amplitude amplitude of the current field
 *
 * \see Test_solve_signal_sidebands()
 */
void solve_signal_sidebands(double base_frequency, int field_type,
        int modulator_index, int modulation_order_index,
        int modulation_sideband_flag, complex_t field_amplitude) {
    // sanity check inputs
    assert(field_type == CARRIER_FIELD ||
            field_type == MODULATOR_FIELD);

    assert(modulator_index >= 0);
#ifndef NDEBUG
    if (modulator_index != 0) {
        assert(modulator_index < inter.num_modulators);
    }
#endif

    assert(modulation_sideband_flag == -1 ||
            modulation_sideband_flag == 0 ||
            modulation_sideband_flag == 1);

    // signal frequencies around modulator freqs
    fill_signals(); // fill signals with field amplitudes from current solution 
    int signal_sideband_flag = 1;

    int signal_sideband_index;
    for (signal_sideband_index = 0;
            signal_sideband_index < 2;
            signal_sideband_index++) {

        double f = base_frequency + signal_sideband_flag * inter.fsig;
        fill_matrix(f, STANDARD);
        clear_rhs(STANDARD);

        signals_rhs(f, base_frequency, field_type, signal_sideband_flag, modulator_index, modulation_order_index, modulation_sideband_flag, field_amplitude);

        factor_matrix(STANDARD);
        solve_matrix(STANDARD);

        // store the solution for signal sidebands around the
        // current modulation frequency
        if (field_type == CARRIER_FIELD) {
            store_solution_old(f, inter.num_frequencies, CARRIER_SIGNAL);
        } else if (field_type == MODULATOR_FIELD) {
            store_solution_old(f, inter.num_frequencies, MODULATOR_SIGNAL);
        } else {
            bug_error("Incorrect field type\n");
        }

        t_s[inter.num_frequencies] = SIGNAL_FIELD;
        f_s[inter.num_frequencies] = f;
        inter.num_frequencies++;

        if (inter.num_frequencies > mem.num_frequencies) {
            bug_error("too many freqs (sig2)");
        }

        signal_sideband_flag = -1;
    }

}


//! Fill detectors with output signals computed from field amplitudes

/*!
 * \todo can this routine be made clearer?
 *
 * \see Test_fill_detectors()
 */
void fill_detectors(void) {
    complex_t z;
    motion_out_t *mout;
    beampar_out_t bpout;
    cavity_par_out_t cpout;
    complex_t *q;
    int detector_index;

    int output_index;
    output_data_t *output_data;

    for (output_index = 0; output_index < inter.num_outputs; output_index++) {
        output_data = &(inter.output_data_list[output_index]);
        detector_index = output_data->detector_index;
        
        //if (inter.debug)
        //printf("Detector: %d (%s) type:%d\n", 
        //i, output_data->name, output_data->detector_type);

        if (output_data->detector_type == PD0 ||
                output_data->detector_type == PD1 ||
                output_data->detector_type == AD ||
                output_data->detector_type == QNOISE ||
                output_data->detector_type == SHOT ||
                output_data->detector_type == QSHOT ||
                output_data->detector_type == CONV ||
                output_data->detector_type == BEAM) {

            output_data->signal = complex_0;
            const int num_demods = inter.light_out_list[detector_index].num_demods;
            
            // calculate detector sigs from a_s[][]
            switch (output_data->detector_type) {
                case AD:
                    output_data->signal = amplitude_detector(detector_index);
                    break;
                case QNOISE:     
                    if(inter.fsig == 0 && !inter.warned_quantum_no_fsig){
                        warn("Quantum noise computations require a signal frequency to be specified using 'fsig'\n");
                        warn("Use 'fsig name frequency [phase] [amplitude]' if no signal needs to be injected.\n");
                        inter.warned_quantum_no_fsig = true;
                    }
                    
                    if(inter.light_out_list[detector_index].sensitivity == OFF){
                        output_data->signal.re = compute_qnoised_output(detector_index);
                        output_data->signal.im = 0;
                    } else {
                        // must compute signal first to fill in max phase computation
                        double signal_power = zmod(photo_detector2(detector_index, num_demods));
                        double noise_power  = compute_qnoised_output(detector_index);
                        
                        // rescale if needed into proper units of amplitude spectral density
                        if(init.quantum_output_scale == PSD_HF)
                            noise_power = sqrt(noise_power*H_PLANCK * inter.f0);
                        else if(init.quantum_output_scale == ASD_HF)
                            noise_power *= sqrt(H_PLANCK * inter.f0);
                        else if(init.quantum_output_scale == PSD)
                            noise_power = sqrt(noise_power);
                        
                        output_data->signal.im = 0;
                        
                        if(inter.light_out_list[detector_index].sensitivity == ON)
                            output_data->signal.re = noise_power/signal_power;
                        else
                            output_data->signal.re = signal_power/noise_power;
                    }
                    
                    break;
                case PD0:
                    output_data->signal = photo_detector0(detector_index);
                    break;
                case PD1:
                    output_data->signal = photo_detector2(detector_index, num_demods);
                    break;
                case SHOT:
                    z = photo_detector0(detector_index);
                    output_data->signal.re = schottky_shotnoise(z.re);
                    output_data->signal.im = 0.0;
                    break;
                    
                case QSHOT:
                    if(inter.light_out_list[detector_index].sensitivity == OFF){
                        output_data->signal.re = compute_qshot_output(detector_index);
                        output_data->signal.im = 0;
                    } else {
                        // must compute signal first to fill in max phase computation
                        double signal_power = zmod(photo_detector2(detector_index, num_demods));
                        double noise_power  = compute_qshot_output(detector_index);
                        
                        // rescale if needed into proper units of amplitude spectral density
                        if(init.quantum_output_scale == PSD_HF)
                            noise_power = sqrt(noise_power*H_PLANCK * inter.f0);
                        else if(init.quantum_output_scale == ASD_HF)
                            noise_power *= sqrt(H_PLANCK * inter.f0);
                        else if(init.quantum_output_scale == PSD)
                            noise_power = sqrt(noise_power);
                        
                        output_data->signal.im = 0;
                        
                        if(inter.light_out_list[detector_index].sensitivity == ON)
                            output_data->signal.re = noise_power/signal_power;
                        else
                            output_data->signal.re = signal_power/noise_power;
                    }
                    break;
                case BEAM:
                    output_data->signal = beam_shape(detector_index);
                    break;
                case CONV:
                    output_data->signal = beam_convolution(detector_index);
                    break;
                default:
                    message("i=%d, type=%d\n", output_index, output_data->detector_type);
                    bug_error("unknown/incorrect detector type");
            }

            if (inter.debug & 4) {
                message("dp1 z_tmp[%d]=%s\n",
                        output_index, complex_form15(output_data->signal));
            }

            // handle default sensitivity computation
            if(output_data->detector_type != QNOISE && output_data->detector_type != QSHOT ){
                        
                if (inter.light_out_list[detector_index].sensitivity == ON) {
                
                    z = photo_detector0(detector_index); // Bug fix Michaela 200104
                    double noise_power = schottky_shotnoise(z.re);
                    
                    output_data->signal.re = noise_power / zmod(output_data->signal);
                    output_data->signal.im = 0.0;
                    
                } else if (inter.light_out_list[detector_index].sensitivity == NORM) {
                    z = photo_detector0(detector_index);
                    
                    output_data->signal.re =  zmod(output_data->signal) / schottky_shotnoise(z.re);
                    output_data->signal.im = 0.0;
                    
                } else if(inter.light_out_list[detector_index].sensitivity != OFF) {
                    bug_error("unexpected sensitivity value");
                }
            }
            
        } else if(output_data->detector_type == OLTFD){
            if(inter.num_signals > 0){
                // zero the input array
                memset(M_ifo_sig.rhs_values, 0, sizeof(complex_t)*M_ifo_sig.M.num_eqns);

                openloopgain_out_t *olg = &(inter.openloopgain_list[output_data->detector_index]);

                int rhs_idx = -1;

                if(olg->comp_type == MIRROR) {
                    mirror_t *m = &inter.mirror_list[olg->comp_index];
                    rhs_idx = m->x_rhs_idx + olg->motion_index;
                } else if(olg->comp_type == BEAMSPLITTER) {
                    beamsplitter_t *bs = &inter.bs_list[olg->comp_index];
                    rhs_idx = bs->x_rhs_idx + olg->motion_index;
                } else {
                    bug_error("Unhanled component type");
                }

                // fill vector to select out the diagonal transfer function for the motion
                // which should be Delta A/A as in Evans paper
                complex_t *crhs = (complex_t*)M_ifo_sig.rhs_values;
                crhs[rhs_idx] = complex_1;

                solve_ccs_matrix(&M_ifo_sig, M_ifo_sig.rhs_values, false, false);

                // multiple by -1 as need to include phase inversion??
                output_data->signal = z_pl_z(complex_1, z_by_z(complex_m1, crhs[rhs_idx]));
            } else {
                output_data->signal = complex_0;
            }
        } else if(output_data->detector_type == SD){
            if(inter.fsig == 0 && !inter.warned_quantum_no_fsig){
                warn("Quantum noise computations require a signal frequency to be specified using 'fsig'\n");
                warn("Use 'fsig name frequency [phase] [amplitude]' if no signal needs to be injected.\n");
                inter.warned_quantum_no_fsig = true;
            }
            
            double r, phi, size;
            light_out_t *out = &inter.light_out_list[output_data->detector_index];
            
            frequency_t *f = get_carrier_frequency(out->f[1]);
            
            if(f == NULL){
                output_data->signal = complex_0;
            } else {
                compute_squeezing_factor(out, &r, &phi, &size);
                output_data->signal = z_by_xphr(complex_1, r, phi);
            }
            
        } else if(output_data->detector_type == QD){
            if(inter.fsig == 0 && !inter.warned_quantum_no_fsig){
                warn("Quantum noise computations require a signal frequency to be specified using 'fsig'\n");
                warn("Use 'fsig name frequency [phase] [amplitude]' if no signal needs to be injected.\n");
                inter.warned_quantum_no_fsig = true;
            }
            
            light_out_t *qd = &inter.light_out_list[detector_index];
            
            frequency_t *f = get_carrier_frequency(qd->freq);
            
            if(f == NULL){
                output_data->signal = complex_0;
            } else {
                double quadrature;
                compute_quadrature(qd, &quadrature);
                        
                output_data->signal.re = quadrature;
                output_data->signal.im = 0;
            }
            
        } else if(output_data->detector_type == XD){
            mout = &inter.motion_out_list[detector_index];
            
            int x_rhs_idx, num_motions;
            bool set = false;
            
            switch(mout->component_type){
                case MIRROR:
                    if(!set){
                        mirror_t *m = &inter.mirror_list[mout->component_index];
                        x_rhs_idx = m->x_rhs_idx;
                        num_motions = m->num_motions;
                        mout->motion_index = parse_motion_value(mout, m->mass, m->Ix, m->Iy, m->num_surface_motions); 
                        set = true;
                    }
                case BEAMSPLITTER:      
                    if(!set){
                        beamsplitter_t *bs = &inter.bs_list[mout->component_index];
                        x_rhs_idx = inter.bs_list[mout->component_index].x_rhs_idx;
                        num_motions = bs->num_motions;
                        set = true;
                        // TODO add surface motion parsing when added
                        mout->motion_index = parse_motion_value(mout, bs->mass, bs->Ix, bs->Iy, 0);
                    }
                    
                    // no mechanical motion will be happening if this is null
                    if(mout->motion_index == NOT_FOUND || mout->motion_index >= num_motions || num_motions == 0) {
                        output_data->signal = complex_0;
                    } else {
                        output_data->signal = ((complex_t*)M_ifo_sig.rhs_values)[x_rhs_idx+mout->motion_index];
                        output_data->signal = z_by_x(output_data->signal, init.x_scale);
                    }
                    break;
                default:
                    bug_error("Component type not compatiable with motion detector");
            }
            
        } else if (output_data->detector_type == CP) {
            cpout = inter.cavity_par_out_list[detector_index];

            cavity_t *cav;

            cav = &inter.cavity_list[cpout.cavity];

            if (cpout.x) {
                q = &cav->qx;
            } else {
                q = &cav->qy;
            }

            double n = cav->n;

            switch (cpout.action) {
                case CPW0:
                    output_data->signal.re = w0_size(*q, n);
                    output_data->signal.im = 0.0;
                    break;
                case CPW:
                    output_data->signal.re = w_size(*q, n);
                    output_data->signal.im = 0.0;
                    break;
                case CPZ:
                    output_data->signal.re = z_q(*q);
                    output_data->signal.im = 0.0;
                    break;
                case CPF:
                    output_data->signal.re = cav->finesse;
                    output_data->signal.im = 0.0;
                    break;
                case CPS:
                    if (cpout.x)
                        output_data->signal.re = cav->stability_x;
                    else
                        output_data->signal.re = cav->stability_y;
                    
                    output_data->signal.im = 0.0;
                    break;
                case CPL:
                    output_data->signal.re = cav->loss;
                    output_data->signal.im = 0.0;
                    break;
                case CPOL:
                    output_data->signal.re = cav->length;
                    output_data->signal.im = 0.0;
                    break;
                case CPFSR:
                    output_data->signal.re = cav->FSR;
                    output_data->signal.im = 0.0;
                    break;
                case CPFWHM:
                    output_data->signal.re = cav->FWHM;
                    output_data->signal.im = 0.0;
                    break;
                case CPPOLE:
                    output_data->signal.re = cav->pole;
                    output_data->signal.im = 0.0;
                    break;
                case CPRC:
                    output_data->signal.re = roc(*q);
                    output_data->signal.im = 0.0;
                    break;
                case CPQ:
                    output_data->signal = *q;
                    break;
                default:
                    message("action %d\n", cpout.action);
                    bug_error("No such parameter (detector outputs cp 1)");
                    break;
            }
        } else if (output_data->detector_type == BP) {
            bpout = inter.beampar_out_list[detector_index];

            if (bpout.x) {
                q = output_data->qx;
            } else {
                q = output_data->qy;
            }

            switch (bpout.action) {
                case BPW0:
                    output_data->signal.re =
                            w0_size(*q, *inter.node_list[output_data->node_index].n);
                    output_data->signal.im = 0.0;
                    break;
                case BPW:
                    output_data->signal.re =
                            w_size(*q, *inter.node_list[output_data->node_index].n);
                    output_data->signal.im = 0.0;
                    break;
                case BPG:
                    output_data->signal.re = gouy(*q);
                    output_data->signal.im = 0.0;
                    break;
                case BPZ:
                    output_data->signal.re = z_q(*q);
                    output_data->signal.im = 0.0;
                    break;
                case BPZR:
                    output_data->signal.re = z_r(*q);
                    output_data->signal.im = 0.0;
                    break;
                case BPRC:
                    output_data->signal.re = roc(*q);
                    output_data->signal.im = 0.0;
                    break;
                case BPQ:
                    output_data->signal = *q;
                    break;
                default:
                    message("action %d\n", bpout.action);
                    bug_error("No such parameter (detector outputs 1)");
                    break;
            }
        } else if (output_data->detector_type == PG) {
            bpout = inter.beampar_out_list[detector_index];

            output_data->signal.re = 0;
            output_data->signal.im = 0;

            if (bpout.x) {
                int k;
                for (k = 0; k < bpout.num_spaces; k++) {
                    output_data->signal.re +=
                            inter.space_list[bpout.space_list[k]].gouy_x;
                }
            } else {
                int k;
                for (k = 0; k < bpout.num_spaces; k++) {
                    output_data->signal.re +=
                            inter.space_list[bpout.space_list[k]].gouy_y;
                }
            }
        } else if (output_data->detector_type == FEEDBACK) {
            output_data->signal.re = inter.lock_list[detector_index].value;
            output_data->signal.im = 0.0;
        }/*      else  if (output_data->detector_type==UFUNCTION)
            {
            j=output_data->detector; 
            do_function(j); // to have first value set correctly
            output_data->signal.re=inter.function_list[j].result;         
            output_data->signal.im=0.0;
            } 
    */
        else if (output_data->detector_type != UFUNCTION) {
            fprintf(stderr, "No. %d, name %s, type %d\n",
                    output_index, output_data->name, output_data->detector_type);
            bug_error("No such output (detector outputs 3)");
        }

        if (output_data->detector_type != UFUNCTION) {
            if (inter.debug & 4) {
                message("dp2 z_tmp[%d]=%s\n",
                        output_index, complex_form15(output_data->signal));
            }

            output_data->re = output_data->signal.re;
            output_data->im = output_data->signal.im;
            output_data->abs = sqrt(zabs(output_data->signal));
            output_data->deg = zdeg(output_data->signal);

            if (inter.debug & 2048) {
                message("%d: %s.re: %g\n",
                        output_index, output_data->name, output_data->re);
            }
        }
    }

    if (inter.num_func_cmds) {
        compute_functions();
        for (output_index = 0; output_index < inter.num_outputs; output_index++) {
            output_data = &(inter.output_data_list[output_index]);
            if (output_data->detector_type == UFUNCTION) {
                if (inter.debug & 4) {
                    message("(func) dp2 z_tmp[%d]=%s\n",
                            output_index, complex_form15(output_data->signal));
                }

                detector_index = output_data->detector_index;
                output_data->signal.re = inter.function_list[detector_index].result;
                output_data->signal.im = 0.0;
                output_data->re = output_data->signal.re;
                output_data->im = output_data->signal.im;
                output_data->abs = sqrt(zabs(output_data->signal));
                output_data->deg = zdeg(output_data->signal);
                if (inter.debug & 2048) {
                    message("%d: %s.re: %g\n",
                            output_index, output_data->name, output_data->signal.re);
                }
            }
        }
    }
}

//! Calculate the beam shape

/*!
 *
 *
 * \see Test_beam_shape() TODO: function rewritten, test needs to be adapted
 */
complex_t beam_shape(int detector_index) {
    int tn, tm;
    complex_t at1, at2;
    double phase;

    light_out_t bout = inter.light_out_list[detector_index];
    output_data_t out = inter.output_data_list[bout.output_index];
    complex_t qx = *(out.qx);
    complex_t qy = *(out.qy);
    double nr = *inter.node_list[out.node_index].n;
    double x = w0_size(qx, nr) * bout.x;
    double y = w0_size(qy, nr) * bout.y;

    complex_t z_tmp = complex_0;

    if (bout.freq_is_set) {
        int freq_index;
        for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
            if (eq(f_s[freq_index], bout.freq)) {
                at1 = complex_0;
                int field_index;
                for (field_index = 0; field_index < inter.num_fields; field_index++) {
                    get_tem_modes_from_field_index(&tn, &tm, field_index);
                    if (inter.set_tem_phase_zero & 2) {
                        phase = tn * gouy(qx) + tm * gouy(qy);
                    } else {
                        phase = (tn + 0.5) * gouy(qx) + (tm + 0.5) * gouy(qy);
                    }
                    //          message("x=%f, y=%f",x,y);
                    at1 = z_pl_z(at1,
                            z_by_z(u_nm(tn, tm, qx, qy, x, y, nr),
                            z_by_x(z_by_phr
                            (a_s[field_index][detector_index][freq_index], -1.0 * phase),
                            bout.masking_factor[field_index])));
                }
                z_tmp = z_pl_z(z_tmp, at1);
            }
        }
    } else {
        int i, j;
        for (i = 0; i < inter.num_frequencies; i++) {
            if (t_s[i] != SIGNAL_FIELD && ((NOT bout.freq_is_set) || eq(f_s[i], bout.freq))) {
                for (j = 0; j < inter.num_frequencies; j++) {
                    if (t_s[j] != SIGNAL_FIELD && eq(f_s[i], f_s[j])) {
                        at1 = at2 = complex_0;
                        int field_index;
                        for (field_index = 0; field_index < inter.num_fields; field_index++) {
                            get_tem_modes_from_field_index(&tn, &tm, field_index);
                            if (inter.set_tem_phase_zero & 2) {
                                phase = tn * gouy(qx) + tm * gouy(qy);
                            } else {
                                phase = (tn + 0.5) * gouy(qx) + (tm + 0.5) * gouy(qy);
                            }

                            complex_t u_tmp = u_nm(tn, tm, qx, qy, x, y, nr);
                            at1 = z_pl_z(at1,
                                    z_by_z(u_tmp,
                                    z_by_x(z_by_phr
                                    (a_s[field_index][detector_index][i], -1.0 * phase),
                                    bout.masking_factor[field_index])));
                            at2 = z_pl_z(at2,
                                    z_by_z(u_tmp,
                                    z_by_x(z_by_phr
                                    (a_s[field_index][detector_index][j], -1.0 * phase),
                                    bout.masking_factor[field_index])));
                            //printf("i, j %d, %d - n %d tn %d tm %d: ", i, j, n, tn, tm);
                            //printf("aaa %s %s\n", complex_form(at1), complex_form(at2));
                        }
                        z_tmp = z_pl_z(z_tmp, z_by_zc(at1, at2));
                    }
                }
            }
        }
    }
    return z_tmp;
}


//====================================================================================================================
//! Calculate the beam convolution

/*!
 *
 *
 * 
 */
complex_t beam_convolution(int detector_index) {
    int tn, tm;
    complex_t at1, at2;
    double phase;

    convolution_out_t conv_out = inter.convolution_out_list[detector_index];
    //light_out_t bout = inter.light_out_list[detector_index];
    output_data_t out = inter.output_data_list[conv_out.output_index];
    complex_t qx = *(out.qx);
    complex_t qy = *(out.qy);
    double nr = *inter.node_list[out.node_index].n;
    //  double x = w0_size(qx, nr) * bout.x;
    //double y = w0_size(qy, nr) * bout.y;

    double x = 0;
    double y = 0;

    complex_t model_qx;
    complex_t model_qy;

    // TODO move this into setup
    // adf check the use of model_qx/qy and Gouy phase settings, 240713
    if (conv_out.x_is_set) {
        model_qx = q_w0z(conv_out.w0x, conv_out.z0x, nr);
    } else {
        model_qx = qx;
    }
    if (conv_out.y_is_set) {
        model_qy = q_w0z(conv_out.w0y, conv_out.z0y, nr);
    } else {
        model_qy = qy;
    }

    complex_t z_tmp = complex_0;

    if (conv_out.freq_is_set) {
        int freq_index;
        for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
            if (eq(f_s[freq_index], conv_out.freq)) {
                at1 = complex_0;
                int field_index;
                for (field_index = 0; field_index < inter.num_fields; field_index++) {
                    get_tem_modes_from_field_index(&tn, &tm, field_index);
                    if (inter.set_tem_phase_zero & 2) {
                        phase = tn * gouy(model_qx) + tm * gouy(model_qy);
                    } else {
                        phase = (tn + 0.5) * gouy(model_qx) + (tm + 0.5) * gouy(model_qy);
                    }
                    // message("x=%f, y=%f",x,y);
                    at1 = z_pl_z(at1,
                            z_by_z(u_nm(tn, tm, model_qx, model_qy, x, y, nr),
                            z_by_x(z_by_phr
                            (a_s[field_index][detector_index][freq_index], -1.0 * phase),
                            conv_out.masking_factor[field_index])));
                }
                z_tmp = z_pl_z(z_tmp, at1);
            }
        }
    } else {
        int i, j;
        for (i = 0; i < inter.num_frequencies; i++) {
            if (t_s[i] != SIGNAL_FIELD && ((NOT conv_out.freq_is_set) || eq(f_s[i], conv_out.freq))) {
                for (j = 0; j < inter.num_frequencies; j++) {
                    if (t_s[j] != SIGNAL_FIELD && eq(f_s[i], f_s[j])) {
                        at1 = at2 = complex_0;
                        int field_index;
                        for (field_index = 0; field_index < inter.num_fields; field_index++) {
                            get_tem_modes_from_field_index(&tn, &tm, field_index);
                            if (inter.set_tem_phase_zero & 2) {
                                phase = tn * gouy(model_qx) + tm * gouy(model_qy);
                            } else {
                                phase = (tn + 0.5) * gouy(model_qx) + (tm + 0.5) * gouy(model_qy);
                            }

                            complex_t u_tmp = u_nm(tn, tm, model_qx, model_qy, x, y, nr);
                            at1 = z_pl_z(at1,
                                    z_by_z(u_tmp,
                                    z_by_x(z_by_phr
                                    (a_s[field_index][detector_index][i], -1.0 * phase),
                                    conv_out.masking_factor[field_index])));
                            at2 = z_pl_z(at2,
                                    z_by_z(u_tmp,
                                    z_by_x(z_by_phr
                                    (a_s[field_index][detector_index][j], -1.0 * phase),
                                    conv_out.masking_factor[field_index])));
                            //printf("i, j %d, %d - n %d tn %d tm %d: ", i, j, n, tn, tm);
                            //printf("aaa %s %s\n", complex_form(at1), complex_form(at2));
                        }
                        z_tmp = z_pl_z(z_tmp, z_by_zc(at1, at2));
                    }
                }
            }
        }
    }
    return z_tmp;
}
//====================================================================================================================

//! Calculate the amplitude at detector

/*!
 * \param detector_index detector index
 *
 * \see Test_amplitude_detector()
 */
complex_t amplitude_detector(int detector_index) {
    // sanity check on input
    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    light_out_t *light_output;
    light_output = &(inter.light_out_list[detector_index]);
    output_data_t output_data = inter.output_data_list[light_output->output_index];

    complex_t z_tmp = complex_0;
    double amp = 0.0;

    // make sure that we only compute fields that are below maxtem (or 00 for maxtem off)
    if ((light_output->field_is_set && inter.tem_is_set && (light_output->field>=inter.num_fields))
            || (light_output->field_is_set && !inter.tem_is_set && (light_output->field!=0))) {
        if (inter.debug & 8) {
            message("ad %s: non-exisiting mode, detector ignored\n",light_output->name);            
        }
        return(complex_0);       
    }
    else if (light_output->field_is_set || !inter.tem_is_set) {
        int field_index = light_output->field;
        int freq_index;
        for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
            if (eq(f_s[freq_index], light_output->f[1])) {
                z_tmp = z_pl_z(z_tmp, a_s[field_index][detector_index][freq_index]);
                
                if (inter.debug & 8) {
                    message("ad a[%d][%d]=%s\n", field_index, freq_index, complex_form15(a_s[field_index][detector_index][freq_index]));
                }
            }
        }
        double phase = 0.0;
        if (inter.set_tem_phase_zero & 4) {
            complex_t qx = *(output_data.qx);
            complex_t qy = *(output_data.qy);

            int tn, tm;
            get_tem_modes_from_field_index(&tn, &tm, field_index);

            if (inter.set_tem_phase_zero & 2) {
                phase = tn * gouy(qx) + tm * gouy(qy);
            } else {
                phase = (tn + 0.5) * gouy(qx) + (tm + 0.5) * gouy(qy);
            }
        } else {
            phase = 0.0;
        }

        return (z_by_phr(z_tmp, -1.0 * phase));
    } else {
        // adf 21.04.2012, swapped order of for loops so that same fields in one mode are added as amplitudes!
        int field_index;
        amp=0.0;
        complex_t z_fake=complex_0; // dummy variable to compute a fake phase 
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            z_tmp=complex_0;
            int freq_index;
            for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
                if (eq(f_s[freq_index], light_output->f[1])) {
                    z_tmp = z_pl_z(z_tmp, a_s[field_index][detector_index][freq_index]);           
                    if (inter.debug & 8) {
                        message("ad a[%d][%d]=%s, amp=%g\n", field_index, freq_index,
                                complex_form15(a_s[field_index][detector_index][freq_index]), amp);
                    }
                }
            }
            z_fake = z_pl_z(z_fake,z_tmp);
            amp += zabs(z_tmp);
            if (inter.debug & 8) {
                message("ad a[%d][%d] z_tmp=%s, amp=%g\n", field_index, freq_index,
                        complex_form15(z_tmp), amp);
            }
        }
        return (get_complex_from_real_and_phase(sqrt(amp), zphase(z_fake)));
    }
}

//! Calculate the output of the photodetector with no demodulation (DC output)

/*!
 * For a discussion of how this photodetector works, see the DC detector
 * subsection of the Photodetectors and demodulation section of the user
 * manual.  This calculation performs an extended version of
 * this equation:
 *
 * \f{equation}
 * x = \sum_i \sum_j a_i a_j^* \quad \mathrm{with} 
 * \quad \{ i,j | i,j \in \{ 0, \ldots, N\}
 * \wedge \omega_i = \omega_j \}
 * \f}
 *
 * If the user has defined a photodetector to use, then this equation is
 * used:
 *
 * \f{equation}
 * P_{DC} = \epsilon_0 \sum_n \sum_m \sum_i \sqrt{c_n c_m} 
 * \sum_j B_k a_i a_j^* \quad \mathrm{with} \quad \omega_i = \omega_j
 * \f}
 *
 * where \f$n\f$ and \f$m\f$ represent the field indices and are related to
 * the TEM mode orders \f$n_1\f$, \f$m_1\f$ and \f$n_2\f$, \f$m_2\f$, 
 * \f$c_n\f$ and \f$c_m\f$ are the masking factors for a given field (and
 * therefore TEM mode order),
 * \f$\epsilon_0\f$ is the permittivity of free space, \f$B_k\f$ is the
 * beat factor for the given photodetector and for the given TEM mode
 * indices, and the sums over
 * \f$i\f$ and \f$j\f$ range over the number of frequencies in the system.
 *
 * In the case that the standard photodetector is used, then this is the
 * equation used:
 *
 * \f{equation}
 * P_{DC} = \epsilon_0 \sum_n \sum_i c_n \sum_j a_i a_j^* \quad
 * \mathrm{with} \quad \omega_i = \omega_j
 * \f}
 *
 * \param k detector index
 *
 * \todo if (light_output->detector_type > -1) branch untested
 * \todo k -> detector_index
 *
 * \see Test_photo_detector0()
 */
complex_t photo_detector0(int k) {
    int n1, m1, n2, m2;
    complex_t z_tmp;
    complex_t z_tmp2;
    double pdfactor;

    // sanity checks on input
    // detector index should be in the range: 0 <= k < inter.num_light_outputs
    assert(k >= 0);
    assert(k < inter.num_light_outputs);

    light_out_t *light_output;
    light_output = &(inter.light_out_list[k]);

    // initialise a temporary complex variable.  This will be returned from
    // the routine
    z_tmp2 = complex_0;

    // if the photodetector type is defined in the ini file...
    if (light_output->detector_type > -1) {

        int field_index_outer;
        for (field_index_outer = 0; field_index_outer < inter.num_fields; field_index_outer++) {
            // get the tem mode numbers for the given field index
            get_tem_modes_from_field_index(&n1, &m1, field_index_outer);

            int m;
            for (m = 0; m < inter.num_fields; m++) {
                // get the tem mode numbers for the given field index
                get_tem_modes_from_field_index(&n2, &m2, m);

                // get beat factor for this photodetector, and for the current
                // tem mode numbers
                pdfactor = init.pdt[light_output->detector_type].cross[n1][m1][n2][m2];

                // initialise a temporary complex variable
                z_tmp = complex_0;

                // for every amplitude (that is not a signal sideband; i.e. fsig) 
                // at a given frequency, beat the amplitude with every other 
                // non-signal amplitude at the same frequency
                int freq_index_outer;
                for (freq_index_outer = 0;
                        freq_index_outer < inter.num_frequencies;
                        freq_index_outer++) {

                    if (t_s[freq_index_outer] != SIGNAL_FIELD) {
                        int freq_index_inner;
                        for (freq_index_inner = 0;
                                freq_index_inner < inter.num_frequencies;
                                freq_index_inner++) {

                            // not a signal, and the same frequency
                            if (t_s[freq_index_inner] != SIGNAL_FIELD &&
                                    eq(f_s[freq_index_outer], f_s[freq_index_inner])) {
                                // print debugging output if requested
                                if (inter.debug & 8) {
                                    message("pd0 [%d, %d]=%s\n",
                                            freq_index_outer, freq_index_inner,
                                            complex_form15(z_by_zc(a_s[field_index_outer][k][freq_index_outer],
                                            a_s[m][k][freq_index_inner])));
                                }

                                // if the beat factor is non-zero, beat the complex
                                // amplitudes with one another
                                // if the beat factor is zero, there is no point in doing
                                // the calculation, as the answer would be zero, so z_tmp
                                // should be left at complex_0 which it was initialised to
                                // hence the check
                                if (pdfactor != 0.0) {
                                    // \sum_j (a_{n,i} a_{m,j}^*) pdfactor
                                    z_tmp = z_pl_z(z_tmp,
                                            z_by_x(z_by_zc
                                            (a_s[field_index_outer][k][freq_index_outer],
                                            a_s[m][k][freq_index_inner]),
                                            pdfactor));
                                }
                            }
                        } // j frequency index
                    }
                    // multiply the first sum (that over j) by the masking factors
                    // (c[n]) for the given photodetector (light output)
                    // \sqrt{c_n c_m} \sum_j a_{n,i} a_{m,j}^* pdfactor
                    z_tmp = z_by_x(z_tmp, sqrt(light_output->masking_factor[field_index_outer] * light_output->masking_factor[m]));
                } // i frequency index

                // complete the summation over the fields
                // \sum_i \sqrt{c_n c_m} \sum_j a_{n,i} a_{m,j}^* pdfactor
                z_tmp2 = z_pl_z(z_tmp2, z_tmp);
            } // m tem mode index
        } // n tem mode index
        // E = \sum_n \sum_m \sum_i \sqrt{c_n c_m} \sum_j a_{n,i} a_{m,j}^* pdfactor
    }// otherwise if we are using the pre-defined DC photodetector
    else {
        // loop over all fields
        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {

            // initialise the temporary summation variable
            z_tmp = complex_0;

            // loop over all frequencies
            int freq_index_outer;
            for (freq_index_outer = 0;
                    freq_index_outer < inter.num_frequencies;
                    freq_index_outer++) {

                // if the type of signal is not a signal sideband continue
                if (t_s[freq_index_outer] != SIGNAL_FIELD) {

                    // loop over all frequencies
                    int freq_index_inner;
                    for (freq_index_inner = 0;
                            freq_index_inner < inter.num_frequencies;
                            freq_index_inner++) {

                        // if the current type of signal is not a signal sideband and
                        // the frequencies of the amplitudes to multiply are the same
                        // continue
                        if (t_s[freq_index_inner] != SIGNAL_FIELD &&
                                eq(f_s[freq_index_outer], f_s[freq_index_inner])) {

                            // output debugging information about a_i a_j^*
                            if (inter.debug & 8) {
                                message("pd0 [%d, %d]=%s\n",
                                        freq_index_outer, freq_index_inner,
                                        complex_form15(z_by_zc(a_s[field_index][k][freq_index_outer],
                                        a_s[field_index][k][freq_index_inner])));
                            }

                            // \f$ A_i = \sum_j a_i a_j^* \f$
                            z_tmp = z_pl_z(z_tmp, z_by_zc(a_s[field_index][k][freq_index_outer],
                                    a_s[field_index][k][freq_index_inner]));
                        }
                    } // end inner frequency sum
                }
                // apply the beam masks
                // \f$ A_n = c_n A_i = c_n \sum_j a_i a_j^* \f$
                z_tmp = z_by_x(z_tmp, light_output->masking_factor[field_index]);

            } // end outer frequency sum

            //      printf("n= %d  z_tmp: %s\n", n, complex_form(z_tmp));

            // E = \sum_n \sum_i c_n \sum_j a_i a_j^* 
            z_tmp2 = z_pl_z(z_tmp2, z_tmp);

        } // fields
    }

    // P_{DC} = \epsilon_0 \sum_n \sum_i c_n \sum_j (a_i a_j^*) 
    z_tmp2 = z_by_x(z_tmp2, init.epsilon_0_c); // Watt Ihr Volt !!!!!!!!!!!

    // output debugging information about the photodetector output
    if (inter.debug & 8) {
        message("   sum=%s\n", complex_form15(z_tmp2));
    }

    // ... and return the DC power
    return z_tmp2;
}

//! Calculate the output of a general photodetector

/*!
 * \param detector_index photodetector index
 * \param num_demod number of demodulations
 *
 * \see Test_photo_detector2()
 */
complex_t photo_detector2(int detector_index, int num_demod) {
    complex_t z[MAX_MIX + 1] = {complex_0};
    complex_t z_tmp;

    // sanity checks on input
    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    assert(num_demod >= 0);
    assert(num_demod <= MAX_DEMOD);

    light_out_t *light_output;
    light_output = &(inter.light_out_list[detector_index]);

    double fmix[MAX_MIX + 1][MAX_DEMOD + 1] = {
        { 0},
        { 0}
    };
    int num_mixes = create_mix_table(num_demod, fmix);
    // calculate output for every freq mix
    int mix_index;
    for (mix_index = 1; mix_index <= num_mixes; mix_index++) {
        double f_ref = 0.0;

        // calculate the current mix frequency from the mix table and the
        // current light output's demodulation frequency
        //! \todo could make this loop into get_reference_freq()
        int demod_index;
        for (demod_index = 1; demod_index <= num_demod; demod_index++) {
            f_ref += fmix[mix_index][demod_index] * light_output->f[demod_index];
        }

        z[mix_index] = get_mixed_amplitude(detector_index, mix_index, f_ref, light_output);

        // report on the sum for the given mix frequency if required
        if (inter.debug & 8) {
            message("pd2(5) Sum: z%d %s\n", mix_index, complex_form15(z[mix_index]));
        }
    }

    z_tmp = set_demodulation_phase(num_demod, z, fmix, light_output, num_mixes);

    complex_t demod_signal = get_demodulation_signal(num_demod, detector_index, light_output, z_tmp);

    return demod_signal;
}



//! Rescales output signal in sol; writes array with pointer arithmetic

/*!
 * \param sol the solution output signal
 *
 * \todo not all OTYPEs are tested
 *
 * \see Test_create_data_array()
 */
void create_data_array(complex_t *sol) {
    int num_axes;
    complex_t z, x;
    double y, dout;
    output_data_t *out;

    dout = 0.0; // silence compiler warning;

    y = 0.0;
    if (inter.yaxis[1]) {
        num_axes = 1;
    } else {
        num_axes = 0;
    }

    int j, k;
    for (j = 0; j < inter.num_outputs; j++) {
        out = &(inter.output_data_list[j]);
        z = sol[j];

        for (k = 0; k <= num_axes; k++) {
            if (out->output_type == COMPLEX) {
                // scale output value by user defined factor
                x = z;
                x = z_by_x(x, out->user_defined_scale);

                switch (inter.yaxis[k]) {
                    case OTYPE_RE:
                        dout = x.re;
                        break;
                    case OTYPE_IM:
                        dout = x.im;
                        break;
                    case OTYPE_ABS:
                        dout = zmod(x);
                        break;
                    case OTYPE_DEG:
                        dout = zdeg(x);
                        break;
                    case OTYPE_DEGP:
                        dout = zdegp(x);
                        break;
                    case OTYPE_DEGM:
                        dout = zdegm(x);
                        break;
                    case OTYPE_DB:
                        dout = 0.5 * zdb(x);
                        break;
                    default:
                        bug_error("inter.yaxis[k] 1 ");
                        break;
                }
            } else {
                // scale output value by user defined factor
                y = z.re;
                y = y * out->user_defined_scale;

                switch (inter.yaxis[k]) {
                    case OTYPE_RE:
                    case OTYPE_ABS:
                        dout = y;
                        break;
                    case OTYPE_DB:
                        dout = xdb(y);
                        break;
                    case OTYPE_IM:
                    case OTYPE_DEG:
                    case OTYPE_DEGM:
                        dout = 0.0;
                        break;
                    default:
                        bug_error("inter.yaxis[k] 2 ");
                        break;
                }
            }
            
            if (options.servermode) {
                // output data through socket
                // to be removed here!!!!!
                printf("Sendind data\n");
            } else {
                // write data to temprorary memory array
                // and compute max/min if -max is set.
                //
                // it looks like respt is a pointer to output data, and that dout is
                // the actual value to put into that location, for later writing to
                // file.
                *(vlocal.respt++) = dout;

                if (out->Omax[k] < *(vlocal.respt - 1)) {
                    out->Omax[k] = *(vlocal.respt - 1);
                    if (inter.x1.xtype == FLOG) {
                        out->Omax_x[k] = x_axis_point / inter.x1.op;
                    } else {
                        out->Omax_x[k] = x_axis_point - inter.x1.op;
                    }

                    if (inter.splot) {
                        if (inter.x2.xtype == FLOG) {
                            out->Omax_y[k] = y_axis_point / inter.x2.op;
                        } else {
                            out->Omax_y[k] = y_axis_point - inter.x2.op;
                        }
                    }
                }

                if (out->Omin[k] > *(vlocal.respt - 1)) {
                    out->Omin[k] = *(vlocal.respt - 1);
                    if (inter.x1.xtype == FLOG) {
                        out->Omin_x[k] = x_axis_point / inter.x1.op;
                    } else {
                        out->Omin_x[k] = x_axis_point - inter.x1.op;
                    }

                    if (inter.splot) {
                        if (inter.x2.xtype == FLOG) {
                            out->Omin_y[k] = y_axis_point / inter.x2.op;
                        } else {
                            out->Omin_y[k] = y_axis_point - inter.x2.op;
                        }
                    }
                }
            }
        }
    }
}

//! Evaluate a function

/*!
 * \param function_index function index
 *
 * \see Test_do_function()
 */
void do_function(int function_index) {
    func_command_t *ftmp;
    char *err;

    // sanity check on input
    // function index should be in the range:
    // 0 <= function_index < inter.num_func_cmds
    assert(function_index >= 0);
    assert(function_index < inter.num_func_cmds);

    ftmp = &inter.function_list[function_index];
    ftmp->result = myfval(ftmp->formula, ftmp->tokens, ftmp->input);
    err = (char *) fget_error();

    if ((err != NULL) || isnan(ftmp->result) || isinf(ftmp->result)) {
        fputs("\n*** Error: ", stderr);
        fprintf(stderr, "computation of function %s failed, error code : %s",
                ftmp->name, err);
        putc('\n', stderr);
        my_finish(1);
    }

    if (inter.debug & 2048) {
        message("func %d: %s = %g = %s ( ",
                function_index, ftmp->name + 1, ftmp->result,
                ftmp->fstring);
        int j;
        for (j = 0; j < (int) strlen(ftmp->tokens); j++) {
            message("%c=%g ", *(ftmp->tokens + j), **(ftmp->input + j));
        }
        message(")\n");
    }
}

//! Automatic gain control

/*!
 * \param lock_index lock index
 * \param counter the current iteration number
 *
 * \see Test_autogain()
 */
void autogain(int lock_index, int counter) {
    double input, finput;
    double slope;

    // sanity check on input
    // lock index should be in the range: 0 <= lock_index < inter.num_locks
    assert(lock_index >= 0);
    assert(lock_index < inter.num_locks);

    // counter should be greater than or equal to zero
    assert(counter >= 0);

    if (inter.lock_list[lock_index].once != -1) {

        input = *(inter.lock_list[lock_index].input) /
                inter.lock_list[lock_index].accuracy;
        finput = fabs(input) + 1.0e-9;

        slope = fabs(input -
                ((2 * inter.lock_list[lock_index].last1 -
                inter.lock_list[lock_index].last2) /
                inter.lock_list[lock_index].accuracy));

        if (inter.debug & 2048) {
            message("%4d  lock %2d: break %4d, slope %10.3g "
                    "(input/acc %10.3g, slope/in %10.3g thresholdhigh-low %g - %g)\n",
                    counter, lock_index, inter.breakcount[lock_index], slope, input,
                    slope / finput,
                    init.lockthresholdhigh, init.lockthresholdlow);
        }

        if (slope / finput < init.lockthresholdlow) { // oscillation?
            inter.breakcount[lock_index]--;
        } else if (slope / finput > init.lockthresholdhigh) { // progress too slow?
            inter.breakcount[lock_index]++;
        }
        //              else            // gain OK!

        //  inter.breakcount[lock_index]=0;
        if (inter.breakcount[lock_index] > init.locktest2) {
            inter.breakcount[lock_index] = 0;
            inter.lock_list[lock_index].gain = inter.lock_list[lock_index].gain /
                    init.gainfactor;
            if (init.autogain == 2) {
                message("%4d v lock %d: new gain %g\n",
                        counter, lock_index, inter.lock_list[lock_index].gain);
            }
        } else if ((-1 * inter.breakcount[lock_index]) > init.locktest2) {
            inter.breakcount[lock_index] = 0;
            inter.lock_list[lock_index].gain = inter.lock_list[lock_index].gain *
                    init.gainfactor;
            if (init.autogain == 2) {
                message("%4d ^ lock %d: new gain %g\n",
                        counter, lock_index, inter.lock_list[lock_index].gain);
            }
        }
    }
}

//! Do the locks

/*!
 * \param lock_index lock index
 *
 * \see Test_do_locks()
 */
void do_locks(int lock_index) {
    // sanity check on input
    // lock index should be in the range: 0 <= lock_index < inter.num_locks
    assert(lock_index >= 0);
    assert(lock_index < inter.num_locks);

    if (inter.lock_list[lock_index].once != -1) {
        inter.lock_list[lock_index].last2 = inter.lock_list[lock_index].last1;
        inter.lock_list[lock_index].last1 = *(inter.lock_list[lock_index].input);
        inter.lock_list[lock_index].value +=
                inter.lock_list[lock_index].gain * (*(inter.lock_list[lock_index].input));
    }
}

//! Determine if interferometer isn't locked

/*!
 * \param lock_index lock index
 *
 * \see Test_is_not_locked()
 */
int is_not_locked(int lock_index) {
    double input, finput;

    // sanity check on input
    // lock index should be in the range: 0 <= lock_index < inter.num_locks
    assert(lock_index >= 0);
    assert(lock_index < inter.num_locks);

    // assumed 'locked' for loops stat have stopped (lock* command)
    if (inter.lock_list[lock_index].once == -1)
        return 0;

    input = *(inter.lock_list[lock_index].input) /
            inter.lock_list[lock_index].accuracy;
    finput = fabs(input) + 1.0e-9;

    if (finput > 1.0) {
        return 1;
    }

    return 0;
}

//! Compute the locks for the interferometer

/*! 
 * if mode < 0, computes all locks sequentially
 * if mode >= 0 computes only lock[mode]
 *
 * \param mode the mode in which to run the routine
 *
 * \see Test_compute_locks()
 */
void compute_locks(int mode) {
    int function_index, lock_index;

    lock_index = 0;
    function_index = 0;

    int i;
    for (i = 0; i < inter.num_locks_and_funcs; i++) {
        if (inter.is_locked[i]) {
            if (mode < 0 || mode == lock_index) {
                do_locks(lock_index);
            }
            lock_index++;

            if (inter.debug & 2048) {
                message("commonn %d lock %d : value %g\n", i, lock_index,
                        inter.lock_list[lock_index].value);
            }
        } else {
            do_function(function_index++);
        }
    }
}

//! Compute the functions

/*!
 *
 *
 * \see Test_compute_functions()
 */
void compute_functions(void) {
    int i;
    // along with computing actual functions also compute any fsig's that have
    // a transfer function assigned to it
    for (i=0; i<inter.num_signals; i++){
        signal_t *sig = &inter.signal_list[i];
        
        if(sig->tf != NULL){
            complex_t value = calc_transfer_func(sig->tf, co(0, TWOPI*inter.fsig));
            
            sig->amplitude = zabs(value);
            sig->phase     = zphase(value);
        }
    }
    
    for (i = 0; i < inter.num_func_cmds; i++) {
        do_function(i);
    }
}

//! Process all put commands

/*!
 *
 *
 * \see Test_compute_puts()
 */
void compute_puts(void) {
    double x;
    put_command_t puttmp;

    int i;
    for (i = 0; i < inter.num_put_cmds; i++) {
        puttmp = inter.put_list[i];

        if (puttmp.mode) {
            x = inter.put_list[i].startvalue + (*(inter.put_list[i].source));
        } else {
            x = *(inter.put_list[i].source);
        }

        if ((puttmp.lborder == 1 && x < puttmp.min)
                || (puttmp.uborder == 1 && x > puttmp.max) || (puttmp.lborder == 2
                && x <= puttmp.min)
                || (puttmp.uborder == 2 && x >= puttmp.max)) {
            gerror("put: parameter %s %s out of range\n", puttmp.component_name, puttmp.parameter_name);
        }

        if (inter.debug & 2048) {
            message("put %d : %g\n", i, x);
        }

        *(inter.put_list[i].target) = x;
        
        // if we are updating the fsig, make sure we always update the negative
        // counterpart too
        if(inter.put_list[i].target == &(inter.fsig))
            inter.mfsig = -inter.fsig;
    }
}

//! Tune the parameters

/*!
 * This function changes all the parameters when one moves one step along
 * the axis.  In principle this changes only the "x" parameter, but with all
 * the functions and locks, several others could be affected.  So,
 * tune_parameters() does all this in order to generate a new proper state
 * of the interferometer before the next data point can be computed.
 * The function also calls data_point() which solves the interferometer matrix
 * and fill_detectors to compute the output (detector) signals. 
 *
 * \see Test_tune_parameters()
 */
void tune_parameters(void) {
    int counter, subcounter, gaincounter;
    int notlocked;
    int index1, index2;

    counter = subcounter = 0;
    gaincounter = 0;
    notlocked = 1;

    int i;
    for (i = 0; i < inter.num_locks; i++) {
        inter.breakcount[i] = 0;
    }

    if (inter.lock && init.autostop != 2) {
        while (notlocked && counter < init.locksteps) {
            index1 = index2 = 0;

            while (index1 < inter.num_locks) {
                while (index2 <= index1) {
                    if (!(init.sequential)) {
                        index1 = index2 = inter.num_locks;
                        compute_locks(-1);
                        if (init.autogain && gaincounter > init.locktest1) {
                            gaincounter = 0;
                            int i;
                            for (i = 0; i < inter.num_locks; i++) {
                                autogain(i, counter);
                            }
                        }
                    } else {
                        if (!(init.sequential & 2)) {
                            index2 = index1;
                        }
                        compute_locks(index2);
                        if (init.autogain && gaincounter > init.locktest1) {
                            autogain(index2, counter);
                            if (index2 == inter.num_locks) {
                                gaincounter = 0;
                            }
                        }
                    }
                    gaincounter++;

                    compute_functions();
                    compute_puts();
                    rebuild_all();
                    
                    if(options.sparse_solver == KLU)
                        data_point_old();
                    else if(options.sparse_solver == KLU_FULL)
                        data_point_new();
                    else
                        bug_error("sparse_sovler not handled");

                    fill_detectors();
                    //  compute_functions();

                    index2++;
                }
                index1++;
            }

            // are we locked?
            notlocked = 0;
            int i;
            for (i = 0; i < inter.num_locks; i++) {
                notlocked += is_not_locked(i);
            }

            if (inter.showiterate > 0) {
                if (subcounter >= inter.showiterate) {
                    if (counter == subcounter) {
                        message("\n");
                    }
                    message("steps %d, lock ", counter);
                    int i;
                    for (i = 0; i < inter.num_locks; i++) {
                        message("%d: in=%g out=%g  ", i,
                                *(inter.lock_list[i].input), inter.lock_list[i].value);
                    }
                    message("\n");
                    subcounter = 0;
                }
            }
            counter++;
            subcounter++;
        }

        inter.startlock++;
        if (inter.startlock > 1 && init.sequential & 4) {
            init.sequential = 0;
        }

        if (inter.debug & 2048) {
            message("Locked!\n");
        }

        if (counter >= init.locksteps) {
            message("Stopped lock iteration after %d steps!\n", init.locksteps);
            if (init.autostop) {
                message("Stop locking now.\n");
                init.autostop = 2;
            }
        }

        if (inter.showiterate == -1) {
            //message("\n");
            message("%d Steps to search for initial operating point, given by:\n",
                    counter);
            int i;
            for (i = 0; i < inter.num_locks; i++) {
                message("%d %s: error signal =%11.4g, "
                        "gain = %11.4g, feedback =%19.12g\n",
                        i, inter.lock_list[i].name, *(inter.lock_list[i].input),
                        inter.lock_list[i].gain, inter.lock_list[i].value);
            }
            inter.showiterate = 0;
        }

        int i;
        for (i = 0; i < inter.num_locks; i++) {
            if (inter.lock_list[i].once == 1) {
                inter.lock_list[i].once = -1;
                message("Stop locking of loop %d now.\n", i);
            }
        }
    } else {
        compute_functions();
        compute_puts();
        rebuild_all();
        
        if(options.sparse_solver == KLU)
            data_point_old();
        else if(options.sparse_solver == KLU_FULL)
            data_point_new();
        else
            bug_error("sparse_sovler not handled");

        fill_detectors();
    }
}

//! Rebuild all components in system

/*!
 * For example when the radius of curvature of a mirror is changed, then the
 * ABCD matrix and the coupling coefficients need to be recomputed.  This is
 * what this function does.
 *
 * \see Test_rebuild_all()
 */
void rebuild_all(void) {
    int tid = startTimer("REBUILD_ALL");
    int _tid;
    
    if (inter.rebuild & 1) {
        int i;
        _tid = startTimer("REBUILD_GRATING");
        for (i = 0; i < inter.num_gratings; i++) {
            if (inter.grating_list[i].rebuild) {
                rebuild_grating(&inter.grating_list[i]);
            }
        }
        endTimer(_tid);
    }

    if (inter.rebuild & 2 && inter.tem_is_set) {
        // mirrors
        int i;
        _tid = startTimer("REBUILD_ABCD");
        
        //#pragma omp parallel
        {
            //#pragma omp for private(i) schedule(guided)
            for (i = 0; i < inter.num_mirrors; i++) {
                if (inter.mirror_list[i].rebuild) {
                    if (inter.mirror_list[i].rebuild > 1) {
                        set_ABCD_mirror(i);
                    }
                }
            }

            // beamsplitters
            //#pragma omp for private(i) schedule(guided)
            for (i = 0; i < inter.num_beamsplitters; i++) {
                if (inter.bs_list[i].rebuild) {
                    if (inter.bs_list[i].rebuild > 1) {
                        set_ABCD_beamsplitter(i);
                    }
                }
            }

            // spaces
            //#pragma omp for private(i) schedule(guided)
            for (i = 0; i < inter.num_spaces; i++) {
                if (inter.space_list[i].rebuild) {
                    if (inter.space_list[i].rebuild > 1) {
                        set_ABCD_space(i);
                    }
                }
            }

            // lenses
            //#pragma omp for private(i) schedule(guided)
            for (i = 0; i < inter.num_lenses; i++) {
                if (inter.lens_list[i].rebuild) {
                    if (inter.lens_list[i].rebuild > 1) {
                        set_ABCD_lens(i);
                    }
                }
            }
        }
        endTimer(_tid);
        
        // retrace all
        if (inter.retrace) {
            _tid = startTimer("REBUILD_RETRACE");
            retrace();
            endTimer(_tid);
        }
    }
    
    if (inter.rebuild && inter.tem_is_set) {
        _tid = startTimer("REBUILD_KNM");
        set_k_all_components(true);
        endTimer(_tid);
    }
    
    endTimer(tid);
}

//! Fill modulators with the carrier light.

/*!
 * This is required to compute the sideband amplitudes
 *
 * \see Test_fill_modulators()
 */
void fill_modulators(void) {
    int node1, node2;
    modulator_t *modulator;

    int j;
    for (j = 0; j < inter.num_modulators; j++) {
        modulator = &(inter.modulator_list[j]);

        node1 = 2 * inter.num_fields * modulator->node1_reduced_index;
        node2 = 2 * inter.num_fields * modulator->node2_reduced_index;

        int i;
        for (i = 0; i < inter.num_fields; i++) {
            modulator->a1[i] = klu.crhs[node2 + modulator->node2_amp_dir.to + 2 * i];
            modulator->a2[i] = klu.crhs[node1 + modulator->node1_amp_dir.to + 2 * i];
        }
    }
}

//! Fill signals with the carrier light

/*!
 *
 * \param carrier_freq frequency of light field around which sidebands will be created
 * \todo break up into individual routines
 *
 * \todo MODULATOR and BEAMSPLITTER branches untested
 *
 * \see Test_fill_signals()
 */
void fill_signals(void) {
    int component_index;
    int mat_index1, mat_index2, mat_index3, mat_index4;

    signal_t *sig;
    mirror_t mirror;
    beamsplitter_t bs;
    light_in_t input_field;

#if DEBUG
    FILE *fp_debug;
    fp_debug = fopen("fill_signals_matrix_indices.dat", "w");
#endif

    int j;
    node_t *n1 = NULL, *n2 = NULL, *n3 = NULL, *n4 = NULL;
    
    for (j = 0; j < inter.num_signals; j++) {
        sig = &(inter.signal_list[j]);
        component_index = sig->component_index;

        int n;
        for (n = 0; n < inter.num_fields; n++) {
            switch (sig->component_type) {
                case BEAMSPLITTER:
                    n1 = &inter.node_list[*sig->node1_index];
                    n2 = &inter.node_list[*sig->node2_index];
                    n3 = &inter.node_list[*sig->node3_index];
                    n4 = &inter.node_list[*sig->node4_index];
                    
                    mat_index1 = 2 * inter.num_fields * sig->node1_reduced_index
                            + sig->node1_amp_dir->to + 2 * n;
                    mat_index2 = 2 * inter.num_fields * sig->node2_reduced_index
                            + sig->node2_amp_dir->to + 2 * n;
                    mat_index3 = 2 * inter.num_fields * sig->node3_reduced_index
                            + sig->node3_amp_dir->to + 2 * n;
                    mat_index4 = 2 * inter.num_fields * sig->node4_reduced_index
                            + sig->node4_amp_dir->to + 2 * n;

                    debug_msg_file(fp_debug, "Beam splitter\n");
                    debug_msg_file(fp_debug, "mat_index1 = %d, mat_index2 = %d\n",
                            mat_index1, mat_index2);
                    debug_msg_file(fp_debug, "mat_index3 = %d, mat_index4 = %d\n",
                            mat_index3, mat_index4);

                    // make sure the node indices don't go out of bounds
                    // i.e. the should be in the range 0 <= node_index < inter.num_eqns
                    assert(mat_index1 >= 0);
                    assert(mat_index2 >= 0);
                    assert(mat_index3 >= 0);
                    assert(mat_index4 >= 0);

                    assert(mat_index1 < inter.num_eqns);
                    assert(mat_index2 < inter.num_eqns);
                    assert(mat_index3 < inter.num_eqns);
                    assert(mat_index4 < inter.num_eqns);

                    if(!n1->gnd_node)
                        sig->a1[n] = klu.crhs[mat_index2];
                    else 
                        sig->a1[n] = complex_0;
                    
                    if(!n2->gnd_node)
                        sig->a2[n] = klu.crhs[mat_index1];
                    else
                        sig->a2[n] = complex_0;
                    
                    if(!n3->gnd_node)
                        sig->a3[n] = klu.crhs[mat_index4];
                    else
                        sig->a3[n] = complex_0;
                    
                    if(!n4->gnd_node)
                        sig->a4[n] = klu.crhs[mat_index3];
                    else
                        sig->a4[n] = complex_0;
                    
                    break;
                case MIRROR:
                    n1 = &inter.node_list[*sig->node1_index];
                    n2 = &inter.node_list[*sig->node2_index];
                    
                    // incoming fields
                    mat_index1 = 2 * inter.num_fields * sig->node1_reduced_index
                            + sig->node1_amp_dir->to + 2 * n;
                    mat_index2 = 2 * inter.num_fields * sig->node2_reduced_index
                            + sig->node2_amp_dir->to + 2 * n;
                    // outgoing fields
                    mat_index3 = 2 * inter.num_fields * sig->node1_reduced_index
                            + sig->node1_amp_dir->fro + 2 * n;
                    mat_index4 = 2 * inter.num_fields * sig->node2_reduced_index
                            + sig->node2_amp_dir->fro + 2 * n;

                    debug_msg_file(fp_debug, "Mirror\n");
                    debug_msg_file(fp_debug, "mat_index1 = %d, mat_index2 = %d\n",
                            mat_index1, mat_index2);

                    // make sure the node indices don't go out of bounds
                    // i.e. the should be in the range 0 <= node_index < inter.num_eqns
                    assert(mat_index1 >= 0);
                    assert(mat_index2 >= 0);
                    assert(mat_index3 >= 0);
                    assert(mat_index4 >= 0);
                    assert(mat_index1 < inter.num_eqns);
                    assert(mat_index2 < inter.num_eqns);
                    assert(mat_index3 < inter.num_eqns);
                    assert(mat_index4 < inter.num_eqns);

                    if(!n1->gnd_node){
                        sig->a1[n] = klu.crhs[mat_index1];
                        sig->ao1[n] = klu.crhs[mat_index3];
                    } else {
                        sig->a1[n] = complex_0;
                        sig->ao1[n] = complex_0;
                    }
                    
                    if(!n2->gnd_node){
                        sig->a2[n] = klu.crhs[mat_index2];
                        sig->ao2[n] = klu.crhs[mat_index4];
                    } else {
                        sig->a2[n] = complex_0;
                        sig->ao2[n] = complex_0;
                    }
                    
                    break;
                case SPACE:
                    n1 = &inter.node_list[*sig->node1_index];
                    n2 = &inter.node_list[*sig->node2_index];
                    
                    mat_index1 = 2 * inter.num_fields * sig->node1_reduced_index
                            + sig->node1_amp_dir->to + 2 * n;
                    mat_index2 = 2 * inter.num_fields * sig->node2_reduced_index
                            + sig->node2_amp_dir->to + 2 * n;

                    debug_msg_file(fp_debug, "Space\n");
                    debug_msg_file(fp_debug, "mat_index1 = %d, mat_index2 = %d\n",
                            mat_index1, mat_index2);

                    // make sure the node indices don't go out of bounds
                    // i.e. the should be in the range 0 <= node_index < inter.num_eqns
                    assert(mat_index1 >= 0);
                    assert(mat_index2 >= 0);
                    assert(mat_index1 < inter.num_eqns);
                    assert(mat_index2 < inter.num_eqns);

                    if(!n1->gnd_node)
                        sig->a1[n] = klu.crhs[mat_index2];
                    else
                        sig->a1[n] = complex_0;
                    
                    if(!n2->gnd_node)
                        sig->a2[n] = klu.crhs[mat_index1];
                    else
                        sig->a2[n] = complex_0;
                    
                    break;
                case LIGHT_INPUT:
                    n1 = &inter.node_list[*sig->node1_index];
                    
                    input_field = inter.light_in_list[component_index];
                    if(!n1->gnd_node)
                        sig->a1[n] = z_by_z( z_by_ph( co(sqrt(input_field.I0 / init.epsilon_0_c), 0.0), input_field.phase), input_field.power_coeff_list[n]); //300302
                    else
                        sig->a1[n] = complex_0;
                    
                    sig->a2[n] = complex_0;
                    break;
                case MODULATOR:
                    n1 = &inter.node_list[*sig->node1_index];
                    n2 = &inter.node_list[*sig->node2_index];
                    
                    sig->a1[n] = complex_0;
                    sig->a2[n] = complex_0;
                    break;
                default:
                    bug_error("wrong component type");
            }
        }


        if (sig->type == SIG_X || sig->type == SIG_Y) {
            
            switch (sig->component_type) {
                case MIRROR:
                    mirror = inter.mirror_list[component_index];
                    n1 = &inter.node_list[*sig->node1_index];
                    n2 = &inter.node_list[*sig->node2_index];
                    
                    int i;
                    for (i = 0; i < inter.num_fields; i++) {
                        sig->ao1[i] = complex_0;
                        sig->ao2[i] = complex_0;
                        int n;
                        // TODO: Remove the 1 in the followings !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
                        for (n = 0; n < inter.num_fields; n++) {
                            if(!n1->gnd_node)
                                sig->ao1[i] = z_pl_z(sig->ao1[i], z_by_z(sig->a1[n], z_by_x(*(mirror.a11[n][i]), 1)));
                            else 
                                sig->ao1[i] = complex_0;
                            
                            if(!n2->gnd_node)
                                sig->ao2[i] = z_pl_z(sig->ao2[i], z_by_z(sig->a2[n], z_by_x(*(mirror.a22[n][i]), 1)));
                            else
                                sig->ao2[i] = complex_0;
                        }
                    }
                    break;
                case BEAMSPLITTER:
                    bs = inter.bs_list[component_index];
                    n1 = &inter.node_list[*sig->node1_index];
                    n2 = &inter.node_list[*sig->node2_index];
                    n3 = &inter.node_list[*sig->node3_index];
                    n4 = &inter.node_list[*sig->node4_index];
                    
                    for (i = 0; i < inter.num_fields; i++) {
                        sig->ao1[i] = complex_0;
                        sig->ao2[i] = complex_0;
                        sig->ao3[i] = complex_0;
                        sig->ao4[i] = complex_0;
                        int n;
                        for (n = 0; n < inter.num_fields; n++) {

                            if(!n1->gnd_node)
                                sig->ao1[i] = z_pl_z(sig->ao1[i], z_by_x(z_by_z(sig->a2[n], bs.knm.k21[n][i]), 1));
                            else
                                sig->ao1[i] = complex_0;
                            
                            if(!n2->gnd_node)
                                sig->ao2[i] = z_pl_z(sig->ao2[i], z_by_x(z_by_z(sig->a1[n], bs.knm.k12[n][i]), 1));
                            else
                                sig->ao2[i] = complex_0;
                            
                            if(!n3->gnd_node)
                                sig->ao3[i] = z_pl_z(sig->ao1[i], z_by_x(z_by_z(sig->a4[n], bs.knm.k43[n][i]), 1));
                            else
                                sig->ao3[i] = complex_0;
                            
                            if(!n4->gnd_node)
                                sig->ao4[i] = z_pl_z(sig->ao2[i], z_by_x(z_by_z(sig->a3[n], bs.knm.k34[n][i]), 1));
                            else
                                sig->ao4[i] = complex_0;
                        }
                    }
                    break;
                default:
                    bug_error("sig-xy");
            }
        }
    }

#if DEBUG
    fclose(fp_debug);
#endif
}

//! Write the data file

/*!
 * \param fp output data file
 */
void write_data_header(FILE *ofp) {
    // if option --noheader has not been used
    // write three comment lines into data file:

    if (!options.noheader) {
        // first line: Finesse version and build 
        fprintf(ofp, "%% ");
        print_version(ofp);

        // second line: 2D/3D plot, y2axis set or not
        if (inter.splot) {
            fprintf(ofp, "%% 3D plot");
        } else {
            fprintf(ofp, "%% 2D plot");
        }
        fprintf(ofp, ", y1axis: %s", inter.y1name);
        if (inter.yaxis[1]) {
            fprintf(ofp, ", y2axis: %s", inter.y2name);
        }
        fprintf(ofp, "\n");



        // third line names of the data column names 
        // x1axis:
        fprintf(ofp, "%% %s", inter.x1.xname);

        if (inter.splot) {
            // x2axis:
            fprintf(ofp, ", %s", inter.x2.xname);
        }

        int i;
        for (i = 0; i < inter.num_outputs; i++) {
            fprintf(ofp, ", %s", inter.output_data_list[i].name);
            // if y2axis exists, print output name twice with axis names attached
            if (inter.yaxis[1]) {
                fprintf(ofp, " %s", inter.y1name);
                fprintf(ofp, ", %s", inter.output_data_list[i].name);
                fprintf(ofp, " %s", inter.y2name);
            }
        }
        fprintf(ofp, "\n\n");
    }
}

//! Write the data file

/*!
 *
 *
 * \see Test_write_data_file()write_data_file
 */
void write_data_file(void) {
    const char *stmp = options.out_format;

    // open output data file 
    if (!inter.x3_axis_is_set || inter.video) {
        if ((ofp = fopen(vglobal.output_fname, "w")) == NULL) {
            gerror("cannot open '%s'\n", vglobal.output_fname);
        }
    }

    set_progress_action_text("Saving data");
    set_progress_message_text("");
    int tot_points, curr_points = 0;
    
    if(inter.splot)
        tot_points = inter.x1.xsteps * inter.x2.xsteps;
    else
        tot_points = inter.x1.xsteps;
        
    // write some useful information in the output file header
    write_data_header(ofp);

    // loop across x axis values
    int i;
    for (i = 0; i <= inter.x1.xsteps; i++) {
        if(!inter.x3_axis_is_set)
            print_progress(tot_points, curr_points);
        
        if (inter.splot) {
            // loop across y axis values if appropriate
            int ii;
            for (ii = 0; ii <= inter.x2.xsteps; ii++) {
                vlocal.respt = mem.restab + inter.num_output_cols * (i * (inter.x2.xsteps + 1) + ii);

                if (inter.x3_axis_is_set && !inter.video) {
                    fprintf(ofp, stmp, *inter.x3.xaxis);
                }

                // loop across the number of output data columns and write the data
                int j;
                for (j = 0; j < inter.num_output_cols; j++) {
                    fprintf(ofp, stmp, *(vlocal.respt++));
                }
                putc('\n', ofp);
                curr_points++;
            }
            putc('\n', ofp);
        } else {
            vlocal.respt = mem.restab + i * inter.num_output_cols;
            // loop across the number of output data columns and write the data
            int j;
            for (j = 0; j < inter.num_output_cols; j++) {
                fprintf(ofp, stmp, *(vlocal.respt++));
            }
            putc('\n', ofp);
            
            curr_points++;
        }
    }
    fflush(ofp);

    // close the output data file
    if (!inter.x3_axis_is_set || inter.video) {
        fclose(ofp);
    }
}

//! Compute data points on request in server mode

/*!
 *
 * \see Test_x0_out()
 */
void x0_out(void) {

    // set parameters from command

    if(options.sparse_solver == KLU)
        data_point_old();
    else if(options.sparse_solver == KLU_FULL)
        data_point_new();
    else
        bug_error("sparse_sovler not handled");

    fill_detectors();

    update_fields = true;

    if (inter.num_diff_cmds) {
        compute_derivative(vlocal.zsolution, 0); // compute a derivative of a y value 
    } else {
        compute_result(vlocal.zsolution); // compute the y value 
    }

}

//! Generate results and plot for 1D or 2D simultions

/*!
 *
 * \see Test_x12_out()
 */
void x12_out(void) {
    char command[LINE_LEN] = {0};
    
    compute_data();

    fflush(stdout);
    fflush(stderr);
    
    write_data_file();

    clear_progress();
    
    int i;
    for (i = 0; i < inter.num_light_outputs; i++) {
        // print warning if it was not clear to KAT whether this was 
        // an error signal or a transfer function.
        // I.e. the sidebands from the signal frequency added at 
        // some point with an error signal (same mixing frequency).
        if (inter.debug & 16) {
            message("transfer[%d]=%d\n", i, transfer[i]);
        }

        if (transfer[i] & 1 && transfer[i] & 2) {
            warn("Output %s : some data points represent neither "
                    "a pure transfer function\n   nor a pure error signal! "
                    "Interpret carefully!\n", inter.light_out_list[i].name);
        }
    }

    if (!options.quiet) {
        message("\n");
    }

    if (options.maxmin) {
        max_min();
    }

    if (inter.time) {
        print_time(difftime(time(NULL), now));
    }

    // replace forward slashes by backslashes if on Windows
    char tmp_out_filename[FNAME_LEN];
    strcpy(tmp_out_filename, vglobal.output_fname);

    if (OS == WIN) {
        char *find_slash;
        while ((find_slash = strchr(tmp_out_filename, '/')) != NULL) {
            *find_slash = '\\';
        }
    }

    if (!options.perl1 && !inter.noscripts) {
        if (!options.quiet) {
            message("writing matlab/python/gnuplot batch files...\n");
        }
        // writing all batch files regardless of termnal setting in 
        // the *.kat file. This probably should be adjustable with an
        // option in the kat.ini file?
        open_file_to_write_ascii(vglobal.gnuplot_fname, &gfp);
        gnufile(0, gfp, tmp_out_filename);
        fclose(gfp);

        open_file_to_write_ascii(vglobal.matlab_fname, &mfp);
        matlabfile(mfp, tmp_out_filename);
        fclose(mfp);
        
        open_file_to_write_ascii(vglobal.python_fname,&pfp);
        pythonfile(pfp, tmp_out_filename);
        fclose(pfp);
    }

    if (inter.num_gnuterm_cmds != NO_GNUTERM && !options.perl1) {

        strcpy(command,replace_str(init.gnucommand, "$s", vglobal.gnuplot_fname));        
        message("calling gnuplot...\n");
        
        if (inter.debug & 1)
            message("(using '%s')\n", command);

        system_call(command);

    }


    // creating python file and calling python!
    if (inter.num_pyterm_cmds != NO_PYTERM && !options.perl1) {
         
        strcpy(command,replace_str(init.pycommand, "$s", vglobal.python_fname));

        // replace forward slashes by backslashes if on Windows
        if (OS == WIN) {
            char *find_slash;
            while ((find_slash = strchr(command, '/')) != NULL) {
                *find_slash = '\\';
            }
        }
        message("calling python...\n");
        if (inter.debug & 1) {
            message("(using '%s')\n", command);
        }
        
        system_call(command);
    }


}

//! Function to generate a number of simulation results and plots (x3axis)


/*
 * \see x12_out()
 *
 * \todo currently works only with gnuplot terminal
 *
 * \see Test_x3_out(void)
 */
void x3_out(void) {
    double xz0, xz1, dxz;
    char command[LINE_LEN] = {0};
    char text[LINE_LEN] = {0};
    char gtext[LINE_LEN] = {0};

    xz0 = xz1 = dxz = 0.0;

    if (inter.video) {
        if (inter.num_gnuterm_cmds != NO_GNUTERM) {
            if (inter.num_gnuterm_cmds != 1) {
                gerror("set exactly one gnuterm for video mode.\n");
            }

            if (NOT init.gnuterm[inter.gnuterm[0]].output_is_file) {
                gerror("set gnuterm with file output for video mode.\n");
            }

            strcpy(command, init.gnucommand);
            strcat(command, " ");
            strcat(command, vglobal.gnuplot_fname);
        }
    } else {
        if (inter.num_gnuterm_cmds != NO_GNUTERM) {
            warn("Gnuplot terminals are switched off for x3axis output.\n");
            inter.num_gnuterm_cmds = NO_GNUTERM;
        }

        open_file_to_write_ascii(vglobal.output_fname, &ofp);
    }

    get_xaxis_values(&inter.x3, &dxz, &xz0, &xz1, 0);

    int i;
    double fxz;
    
    set_progress_message_text("");
        
    for (i = 0, fxz = xz0; i <= inter.x3.xsteps; fxz += dxz, i++) {
        
        if (inter.x3.xtype == FLOG) {
            update_xaxis_values(&inter.x3, exp(fxz));
        } else {
            update_xaxis_values(&inter.x3, fxz);
        }

        if (inter.num_gnuterm_cmds != NO_GNUTERM) {
            if (1 + (int) floor(log10(inter.x3.xsteps)) > 3) {
                sprintf(text, "_%.4d.", i);
            } else if (1 + (int) floor(log10(inter.x3.xsteps)) > 2) {
                sprintf(text, "_%.3d.", i);
            } else if (1 + (int) floor(log10(inter.x3.xsteps)) > 1) {
                sprintf(text, "_%.2d.", i);
            } else {
                sprintf(text, "_%d.", i);
            }
            strcpy(gtext, inter.basename);
            strcat(gtext, text);
            strcat(gtext, init.gnuterm[inter.gnuterm[0]].suffix);

            inter.gnutermfn[0] = gtext;
            set_progress_message_text("name : %s - %d of %d...", inter.gnutermfn[0], i, inter.x3.xsteps);
        }
        
        compute_data();

        write_data_file();

        int j;
        for (j = 0; j < inter.num_light_outputs; j++) {
            // print warning if it was not clear to KAT whether this was 
            // an error signal or a transfer function.
            // I.e. the sidebands from the signal frequency added at 
            // some point with an error signal (same mixing frequency).
            if (inter.debug & 16) {
                message("transfer[%d]=%d\n", j, transfer[j]);
            }

            if (transfer[j] & 1 && transfer[j] & 2) {
                warn("Output %s : "
                        "some data points represent neither a pure transfer function\n"
                        "   nor a pure error signal! Interpret carefully!\n",
                        inter.light_out_list[i].name);
            }
        }

        if (options.maxmin) {
            max_min();
            int j;
            for (j = 0; j < mem.num_outputs; j++) {
                inter.output_data_list[j].Omax[0] = -1e20;
                inter.output_data_list[j].Omax[0] = -1e20;
                inter.output_data_list[j].Omin[0] = 1e20;
                inter.output_data_list[j].Omin[0] = 1e20;
            }
        }

        if (inter.num_gnuterm_cmds != NO_GNUTERM) {
            
            open_file_to_write_ascii(vglobal.gnuplot_fname, &gfp);
            gnufile(i, gfp, vglobal.output_fname);
            fclose(gfp);
            
            system_call(command);
        }
    }

    clear_progress();
    
    fflush(stdout);
    if (inter.time) {
        print_time(difftime(time(NULL), now));
    }

    fclose(ofp);
}

//! Store the current solution

/*!
 * \param frequency The frequency for which to store the solution
 * \param frequency_index Index of currently found frequencies
 * \param field_type The type of field currently being stored.
 * CARRIER_FIELD,
 * CARRIER_SIGNAL, MODULATOR_FIELD, or MODULATOR_SIGNAL.
 *
 * \see Test_store_solution()
 */
void store_solution_old(double frequency, int frequency_index, int field_type) {
    // store the values for the given output, field and frequency
    int out_index;
    
    for (out_index = 0; out_index < inter.num_light_outputs; out_index++) {
        output_data_t output =
                inter.output_data_list[inter.light_out_list[out_index].output_index];
        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            // explanation of calculation of index to spsol:
            //  - the term 2*inter.num_fields*node_reduced_index calculates the correct 
            //  starting row corresponding to the node of the light output (detector)
            //  - the term .is_second_beam chooses the other field direction to the
            //    direction, and so will choose a row corresponding to fields
            //    propagating in this "other" direction
            //  - the term 2*field_index chooses rows of the different fields at 
            //  a given node
            //  - the '2' in the first and last terms is due to the fact that we
            //    have for each field a field direction towards and away from
            //    the given component at that node.  The '2' ensures that we are
            //    consistently selecting fields and directions correctly
            a_s[field_index][out_index][frequency_index] =
                    klu.crhs[2 * inter.num_fields * output.node_reduced_index
                    + output.is_second_beam + 2 * field_index];
        }
    }

    // if debugging is requested, output required information
    if (inter.debug & 4) {
        int out_index;
        for (out_index = 0; out_index < inter.num_light_outputs; out_index++) {
            int field_index;
            for (field_index = 0; field_index < inter.num_fields; field_index++) {
                switch (field_type) {
                    case CARRIER_FIELD:
                        message("mode %d carrier no %d freq %s : %s\n",
                                field_index, out_index, freq_form(frequency),
                                complex_form15(a_s[field_index][out_index][frequency_index]));
                        break;
                    case CARRIER_SIGNAL:
                        message("mode %d signal1 no %d freq %s : %s\n",
                                field_index, out_index, freq_form(frequency),
                                complex_form15(a_s[field_index][out_index][frequency_index]));
                        break;
                    case MODULATOR_FIELD:
                        message("mod     no %d mode %d freq %s : %s\n",
                                out_index, field_index, freq_form(frequency),
                                complex_form15(a_s[field_index][out_index][frequency_index]));
                        break;
                    case MODULATOR_SIGNAL:
                        message("signal2 no %d mode %d freq %s : %s\n",
                                out_index, field_index, freq_form(frequency),
                                complex_form15(a_s[field_index][out_index][frequency_index]));
                        break;
                    default:
                        bug_error("Unknown field type %d\n", field_type);
                }
            }
        }
    }
}


void store_carrier_solutions() {

    // store the values for the given output, field and frequency
    int out, f, mode,idx;       
    // for each frequency of the output node requested get all the field 
    // amplitudes and store them
    for (mode = 0; mode < inter.num_fields; mode++) {

        // store the carrier frequency results
        for (f = 0; f < M_ifo_car.num_frequencies; f++){
            idx = M_ifo_car.frequencies[f]->result_index;
            
            for (out = 0; out < inter.num_light_outputs; out++) { 
                int oidx = M_ifo_car.output_rhs_idx[inter.light_out_list[out].output_index] + get_node_rhs_idx(1, f, mode, M_ifo_car.num_frequencies);
                a_s[mode][out][idx] = (((complex_t*)M_ifo_car.rhs_values)[oidx]);
            }
            
            for (out = inter.num_light_outputs; out < inter.num_light_outputs + inter.num_quad_outputs; out++) { 
                int oidx = M_ifo_car.output_rhs_idx[inter.light_out_list[out].output_index] + get_node_rhs_idx(1, f, mode, M_ifo_car.num_frequencies);
                a_s[mode][out][idx] = (((complex_t*)M_ifo_car.rhs_values)[oidx]);
            }
        }
    }
    
}

void store_signal_solutions() {

    // store the values for the given output, field and frequency
    int out, f, mode,idx;
    // for each frequency of the output node requested get all the field 
    // amplitudes and store them
    for (mode = 0; mode < inter.num_fields; mode++) {

        for (f = 0; f < M_ifo_sig.num_frequencies; f++){
            idx = M_ifo_sig.frequencies[f]->result_index;
            
            for (out = 0; out < inter.num_light_outputs; out++) {        
                int oidx = M_ifo_sig.output_rhs_idx[inter.light_out_list[out].output_index] + get_node_rhs_idx(1, f, mode, M_ifo_sig.num_frequencies);
                a_s[mode][out][idx] = (((complex_t*)M_ifo_sig.rhs_values)[oidx]);
                // if value is a conjugate frequency, un-conjugate it...
                if(M_ifo_sig.frequencies[f]->order == -1){
                    a_s[mode][out][idx].im *= -1;
                }
            }
            
            for (out = inter.num_light_outputs; out < inter.num_light_outputs + inter.num_quad_outputs; out++) { 
                int oidx = M_ifo_sig.output_rhs_idx[out] + get_node_rhs_idx(1, f, mode, M_ifo_sig.num_frequencies);
                a_s[mode][out][idx] = (((complex_t*)M_ifo_sig.rhs_values)[oidx]);
                // if value is a conjugate frequency, un-conjugate it...
                if(M_ifo_sig.frequencies[f]->order == -1){
                    a_s[mode][out][idx].im *= -1;
                }
            }
        }
    }
}

//! Calculates the classical power at the given detector

/*!
 * \param detector_index The detector index at which to measure the power
 *
 * \return power The calculated power
 *
 * \see Test_get_classical_power()
 */
double get_classical_power(int detector_index) {
    // field variables
    complex_t current_field;
    complex_t total_field = complex_0;

    // loop over all frequencies
    int freq_index;
    for (freq_index = 0; freq_index < inter.num_frequencies; freq_index++) {
        // loop over all the fields
        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {
            // get the current value of the classical solution
            current_field = a_s[field_index][detector_index][freq_index];

            total_field = z_pl_z(total_field, current_field);
        }
    }

    double power = zabs(total_field);

    return power;
}

//! Creates the frequency mixing table for demodulating photodetector signals

/*!
 * \param num_demod the number of demodulations
 * \param fmix the array of mix frequencies (also returned from the routine)
 *
 * \return num_mixes: the number of mix frequency permutations???
 *
 * \todo fmix -> freq_mix_table???
 *
 * \see Test_create_mix_table()
 */
int create_mix_table(int num_demod, double fmix[MAX_MIX + 1][MAX_DEMOD + 1]) {
    // sanity check inputs
    assert(num_demod >= 0);
    assert(num_demod < MAX_DEMOD);

    // initialise the mix frequency variables
    //  why not just go fmix_tmp = { 1.0 }; here???
    //  Is the zeroth element of the fmix_tmp and fmix arrays ever used???
    double fmix_tmp[MAX_DEMOD + 1] = {0};
    int demod_index;
    for (demod_index = 1; demod_index <= num_demod; demod_index++) {
        fmix_tmp[demod_index] = 1.0;
        fmix[1][demod_index] = 1.0;
    }

    if (inter.debug & 8) {
        message("(1)\n");
    }

    int num_mixes = 1;
    // generate the possible mix frequency permutations and store in mix table
    int demod_index_outer;
    for (demod_index_outer = 1;
            demod_index_outer < num_demod;
            demod_index_outer++) {

        // mix-frequency permutations
        int mix_permutation_index;
        for (mix_permutation_index = demod_index_outer;
                mix_permutation_index >= 1;
                mix_permutation_index--) {
            num_mixes++;

            // if num_mixes is larger than the maximum number of mixes allowed, barf
            if (num_mixes > MAX_MIX) {
                bug_error("maxmix");
            }

            // flip the sign of all previously defined mix frequency permutations
            fmix_tmp[mix_permutation_index] *= -1.0;

            // store the current set of permutations into fmix
            int demod_index_inner;
            for (demod_index_inner = 1;
                    demod_index_inner <= num_demod;
                    demod_index_inner++) {
                fmix[num_mixes][demod_index_inner] = fmix_tmp[demod_index_inner];
            }
        }
    }

    // output debugging information of the mix table if so requested
    if (inter.debug & 8) {
        message("mix table for %d demodulations:\n", num_demod);
        int mix_index;
        for (mix_index = 1; mix_index <= num_mixes; mix_index++) {
            message("%d : ", mix_index);
            int demod_index;
            for (demod_index = 1; demod_index <= num_demod; demod_index++) {
                message(" %s", double_form(fmix[mix_index][demod_index]));
            }
            message("\n");
        }
    }

    return num_mixes;
}

//! Calculates the internal sum of the photodetector signal calculation

/*!
 * \param mix_index index of the current mix frequency
 * \param field_index_outer index for outer field loop
 * \param field_index_inner index for inner field loop
 * \param detector_index index of the photodetector
 * \param f_ref the reference frequency
 * \param pdfactor photodetector amplication factor
 * \param light_output light output (i.e. detector) structure
 *
 * \return the amplitude sum
 *
 * \see Test_get_amplitude_sum()
 */
complex_t get_amplitude_sum(
        int mix_index,
        int field_index_outer,
        int field_index_inner,
        int detector_index,
        double f_ref,
        double pdfactor,
        light_out_t *light_output) {
    // sanity check inputs
    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    complex_t amplitude_sum = complex_0;

    int freq_index_outer;
    
    //printf("\n");
    
    for (freq_index_outer = 0; freq_index_outer < inter.num_frequencies; freq_index_outer++) {
        int freq_index_inner;
        for (freq_index_inner = 0; freq_index_inner < inter.num_frequencies; freq_index_inner++) {
            // if the difference between the signal frequencies is equal
            // to the current demodulation frequency, and either of the 
            // types of signal is NOT a gravitational wave signal 
            // (i.e. SIGNAL_FIELD) then continue to calculate some kind of
            // transfer parameter and the internal sum contributing to the
            // photodetector output signal
            double factor;
            if (teq(f_s[freq_index_outer], f_s[freq_index_inner], f_ref) && (t_s[freq_index_outer] != SIGNAL_FIELD || t_s[freq_index_inner] != SIGNAL_FIELD)) {
                
                if (t_s[freq_index_outer] == SIGNAL_FIELD || t_s[freq_index_inner] == SIGNAL_FIELD) {

                    // we have a signal; beef up its contribution to the sum
                    factor = 1.0 / EPSILON;
                    if (!ceq(a_s[field_index_outer][detector_index][freq_index_outer], complex_0) &&
                            !ceq(a_s[field_index_inner][detector_index][freq_index_inner], complex_0)) {

                        // output debugging information (about what exactly???)
                        if (inter.debug & 16) {
                            if ((transfer[detector_index] & 1 && !(transfer[detector_index] & 2))) {
                                message("err->trans (out %d) %.15g\n", detector_index, f_ref);
                            }
                        }

                        // The demodulated signal contains a transfer function
                        transfer[detector_index] = transfer[detector_index] | 2;
                    }
                } else {
                    factor = 1;
                    if (!ceq(a_s[field_index_outer][detector_index][freq_index_outer], complex_0) &&
                            !ceq(a_s[field_index_inner][detector_index][freq_index_inner], complex_0)) {

                        // output debugging information (about what exactly???)
                        if (inter.debug & 16) {
                            if (!(transfer[detector_index] & 1) &&
                                    transfer[detector_index] & 2) {
                                message("trans->err (out %d) %.15g\n", detector_index, f_ref);
                            }
                        }

                        transfer[detector_index] = transfer[detector_index] | 1;
                    }
                }

                // if the beat factor is nonzero, calculate the current term
                // in the sum.  If the beat factor is zero, then the current
                // term is zero, and so no point in running the calculation.
                if (pdfactor != 0.0) {
                    
                    complex_t tmp =   z_by_x(z_by_zc
                            (a_s[field_index_outer][detector_index][freq_index_outer],
                            a_s[field_index_inner][detector_index][freq_index_inner]),
                            factor * pdfactor);
                    amplitude_sum = z_pl_z(amplitude_sum, tmp);
                    //printf("%.15g %.15g %g %s %s %s %g %g\n",f_s[freq_index_outer],f_s[freq_index_inner], f_ref, complex_form15(tmp), complex_form15(a_s[field_index_outer][detector_index][freq_index_outer]),complex_form(a_s[field_index_inner][detector_index][freq_index_inner]), factor, pdfactor);
                }

                // apply the beam masks
                amplitude_sum = z_by_x(amplitude_sum,
                        sqrt(light_output->masking_factor[field_index_outer] * light_output->masking_factor[field_index_inner]));

                // output debugging information about the photodetector and
                // the amplitude and phase of the two amplitudes if asked by
                // the user
                if (inter.debug & 8) {
                    message("pd2(1) z%d[%d, %d] (f=%s=%s - %s) : %s\n",
                            mix_index, freq_index_outer, freq_index_inner,
                            freq_form(f_ref), freq_form(f_s[freq_index_outer]),
                            freq_form(f_s[freq_index_inner]),
                            complex_form15(z_by_zc(a_s[field_index_outer][detector_index][freq_index_outer],
                                                          a_s[field_index_inner][detector_index][freq_index_inner])));
                    message("factor %.10g, pdfactor %.10g\n",factor, pdfactor);
                    message("a1: amp=%s ph=%s\n",
                            xdouble_form(zmod(a_s[field_index_outer][detector_index][freq_index_outer])),
                            xdouble_form(zdeg(a_s[field_index_outer][detector_index][freq_index_outer])));
                    message("a2: amp=%s ph=%s\n",
                            xdouble_form(zmod(a_s[field_index_outer][detector_index][freq_index_inner])),
                            xdouble_form(zdeg(a_s[field_index_outer][detector_index][freq_index_inner])));
                }
            }
        } // inner frequency loop
    } // outer frequency loop

    return amplitude_sum;
}

//! Calculates the demodulation phase and applies it to the field amplitudes

/*!
 * \param num_demod the number of demodulations
 * \param z complex amplitude
 * \param fmix the demodulation mix table
 * \param light_output the output light (i.e. detector) structure
 * \param num_mixes number of mixes being used in demodulation
 *
 * \todo passing the arrays z, fmix between the routines works, but
 * there should be a cleaner and safer way to do this to make sure that we
 * don't overwrite fmix or z unintentionally (for example).
 *
 * \todo fmix -> freq_mix_table???
 *
 * \return the resultant amplitude
 *
 * \see Test_set_demodulation_phase()
 */
complex_t set_demodulation_phase(
        int num_demod,
        complex_t z[MAX_MIX + 1],
        double fmix[MAX_MIX + 1][MAX_DEMOD + 1],
        light_out_t *light_output,
        int num_mixes) {
    // sanity check inputs
    assert(num_demod >= 0);
    assert(num_demod < MAX_DEMOD);

    // for each demodulation, set the demodulation phase
    double phi[MAX_MIX + 1] = {0.0};

    int demod_index;
    for (demod_index = 1; demod_index < num_demod; demod_index++) {
        complex_t z1 = complex_0;
        complex_t z2 = complex_0;
        int mix_index;
        switch (light_output->demod_phase_mode[demod_index]) {
                // use the default phase if not user-defined
            case MAX_PHASE:
                // add up the amplitude calculated above, depending upon the 
                // sign of the mixing frequency
                for (mix_index = 1; mix_index <= num_mixes; mix_index++) {
                    if (fmix[mix_index][demod_index] < 0.0) {
                        z1 = z_pl_z(z1, z[mix_index]);
                    } else {
                        z2 = z_pl_z(z2, z[mix_index]);
                    }
                }
                // use the extremal values of amplitude to calculate the maximum phase
                double phimax = 0.5 * (zdeg(z2) - zdeg(z1));
                // store max phase values for later reference
                light_output->phi[demod_index] = phimax;
                
                // calculate the phase to apply to the amplitude later
                for (mix_index = 1; mix_index <= num_mixes; mix_index++) {
                    phi[mix_index] += -phimax * fmix[mix_index][demod_index];
                    
                    if (inter.debug & 8) {
                        message("phimax(%d) :% g\n", mix_index, phi[mix_index]);
                    }
                }
                break;
                // the phase is user-defined
            case USER_PHASE:
                // calculate the phase to apply to the amplitude later
                for (mix_index = 1; mix_index <= num_mixes; mix_index++) {
                    phi[mix_index] += -light_output->phi[demod_index] * fmix[mix_index][demod_index];
                    if (inter.debug & 8) {
                        message("phi(%d) :% g\n", mix_index, phi[mix_index]);
                    }
                }
                break;
                // barf otherwise
            default:
                bug_error("light_output->phase i");
                break;
        }
    }

    // apply the phase to the amplitude
    complex_t z_tmp = complex_0; // initialise z_tmp
    int mix_index;
    for (mix_index = 1; mix_index <= num_mixes; mix_index++) {
        z_tmp = z_pl_z(z_tmp, z_by_ph(z[mix_index], phi[mix_index]));
    }

    if (inter.debug & 8) {
        message("pd2(7) z_tmp(p 1->i) %s\n", complex_form15(z_tmp));
    }

    return z_tmp;

}

//! Calculate and return the demodulation signal

/*!
 * \param num_demod the number of demodulations
 * \param detector_index the current detector
 * \param light_output the light output structure
 * \param demod_amplitude the demodulated amplitude
 *
 * \return the demodulation signal in Watts 
 *
 * \see Test_get_demodulation_signal()
 */
complex_t get_demodulation_signal(
        int num_demod,
        int detector_index,
        light_out_t *light_output,
        complex_t demod_amplitude) {
    // sanity check on inputs
    assert(num_demod >= 0);
    assert(num_demod < MAX_DEMOD);

    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    switch (light_output->demod_phase_mode[num_demod]) {
        case MAX_PHASE:
            demod_amplitude.re = zmod(demod_amplitude);
            demod_amplitude.im = 0;
            break;
        case USER_PHASE:
            demod_amplitude.re =
                    (demod_amplitude.re * cos(light_output->phi[num_demod] * RAD) +
                    demod_amplitude.im * sin(light_output->phi[num_demod] * RAD));
            demod_amplitude.im = 0;
            break;
        default:
            break;
    }

    // 0.5 for each demodulation 050505
    demod_amplitude = z_by_x(demod_amplitude, npow(0.5, num_demod - 1));

    if (transfer[detector_index] == 2) {
        if (inter.debug & 8) {
            message("pd2(8.5) compensating signal mixer.\n");
        }
        // compensate 0.5 from mixer at fsig  050505
        demod_amplitude = z_by_x(demod_amplitude, 2.0);
    }

    if (inter.debug & 8) {
        message("pd2(8) z_tmp(p num_demod) %s\n", complex_form15(demod_amplitude));
    }

    // calculate the power from the amplitude
    demod_amplitude = z_by_x(demod_amplitude, init.epsilon_0_c); // Watt Ihr Volt !!!!!!!!!!!!

    return demod_amplitude;
}

//! Calculates the field amplitude for the current mixing frequency

/*!
 * \param detector_index index of the current detector
 * \param mix_index the current mix index
 * \param f_ref the current value of the reference frequency
 * \param light_output the light output structure
 *
 * \todo much simplification can go on here.  for instance: do we really
 * need to pass mix_index and the detector_index around???
 *
 * \return the mixed amplitude
 *
 * \see Test_get_mixed_amplitude()
 */
complex_t get_mixed_amplitude(
        int detector_index,
        int mix_index,
        double f_ref,
        light_out_t *light_output) {
    // initialise a temporary summation variable
    complex_t z = complex_0;

    // if we have a user-defined photodectector, load its beat and masking
    // definition and calculate the equations appropriately (I can be more
    // clear in this explanation, although what happens below does follow
    // closely to what happens in photo_detector0() reasonably closely)
    int n1, m1, n2, m2; // tem mode indices
    if (light_output->detector_type > -1) {
        int field_index_outer;
        for (field_index_outer = 0;
                field_index_outer < inter.num_fields;
                field_index_outer++) {
            // get the TEM mode indices for the given field index
            get_tem_modes_from_field_index(&n1, &m1, field_index_outer);

            int field_index_inner;
            for (field_index_inner = 0;
                    field_index_inner < inter.num_fields;
                    field_index_inner++) {
                // get the TEM mode indices for the given field index
                get_tem_modes_from_field_index(&n2, &m2, field_index_inner);

                // determine the photodetector beat factor
                double pdfactor = init.pdt[light_output->detector_type].cross[n1][m1][n2][m2];

                // initialise the temporary variable for summation
                complex_t amplitude_sum = get_amplitude_sum(mix_index, field_index_outer, field_index_inner, detector_index, f_ref, pdfactor, light_output);

                // append to the outer sum
                z = z_pl_z(z, amplitude_sum);
                //printf("Amplitude sum: %s\n", complex_form(amplitude_sum));

            } // inner field index loop
        } // outer field index loop
    }// otherwise, use the predefined photodetector to calculate the output
    else {
        int field_index;
        for (field_index = 0; field_index < inter.num_fields; field_index++) {

            // initialise the summation variable
            complex_t amplitude_sum = get_amplitude_sum(mix_index, field_index, field_index, detector_index, f_ref, 1.0, light_output);

            // complete the summation, and multiply by the relevant beam
            // masking parameter
            //masking 220402
            z = z_pl_z(z, z_by_x(amplitude_sum, light_output->masking_factor[field_index]));
        }
    }

    // icm bug 0805 ##############
    // use -sin as second quadrature because that's more common
    // i.e. use first quadrature PLUS 90 degree for second quad

    //     z[m]=cconj(z[m]);

    return z;
}

//! Calculate and return the quantum demodulation signal

/*!
 * \param num_demod the number of demodulations
 * \param detector_index the current detector
 * \param light_output the light output structure
 * \param demod_amplitude the demodulated amplitude
 *
 * \todo Check that this works for *both* qshot and qnoised!!!
 * Am calling just ...qshot_demodulation... temporarily
 *
 * \return the demodulation signal in Watts (huh???) (correct units ???)
 *
 * \todo is this used anymore by qshot???  It is currently untested
 *
 * \see Test_get_qshot_noise_spectral_density()
 */
complex_t get_qshot_noise_spectral_density(
        int num_demod,
        int detector_index,
        light_out_t *light_output,
        complex_t demod_amplitude) {
    // sanity check on inputs
    assert(num_demod >= 0);
    assert(num_demod < MAX_DEMOD);

    assert(detector_index >= 0);
    assert(detector_index < inter.num_light_outputs);

    complex_t noise_spectral_density;

    switch (light_output->demod_phase_mode[num_demod]) {
        case MAX_PHASE:
            noise_spectral_density.re = zmod(demod_amplitude);
            noise_spectral_density.im = 0;
            break;
        case USER_PHASE:
            noise_spectral_density.re =
                    (demod_amplitude.re * demod_amplitude.re) +
                    (demod_amplitude.im * demod_amplitude.im);
            noise_spectral_density.im = 0;
            break;
        default:
            break;
    }

    // 0.5 for each demodulation 050505
    noise_spectral_density = z_by_x(noise_spectral_density, npow(0.5, num_demod - 1));

    if (transfer[detector_index] == 2) {
        if (inter.debug & 8) {
            message("pd2(8.5) compensating signal mixer.\n");
        }
        // compensate 0.5 from mixer at fsig  050505
        noise_spectral_density = z_by_x(noise_spectral_density, 2.0);
    }

    if (inter.debug & 8) {
        message("pd2(8) z_tmp(p num_demod) %s\n", complex_form15(noise_spectral_density));
    }

    // calculate the power from the amplitude
    noise_spectral_density = z_by_x(noise_spectral_density, init.epsilon_0_c); // Watt Ihr Volt !!!!!!!!!!!!

    return noise_spectral_density;
}

/*
 * Local variables:
 * c-indentation-style: bsd
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * End:
 *
 * vim: tabstop=2 expandtab shiftwidth=2:
 */
